//==========================================================================
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// Permission to use, copy, modify and distribute any part of this Makefile
// for educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above 
// copyright notice, this paragraph and the following paragraphs appear 
// in all copies.
//
// Those desiring to incorporate this into commercial products or use for
// commercial purposes should contact the
//
//      Technology Transfer Office, University of California, San Diego,
//      9500 Gilman Drive, La Jolla, CA 92093-0910
//      Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS PLOTPATH, EVEN 
// IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
//
// THE PLOTPATH, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF 
// CALIFORNIA PLOTPATH NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF 
// ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
// PURPOSE, OR THAT THE USE OF THE SKITTERTRACES WILL NOT INFRINGE ANY 
// PATENT, TRADEMARK OR OTHER RIGHTS.
//
// PlotPaths is developed by John Gallagher Bradley L. Huffaker at the
// University of California, San Diego under the Cooperative Association
// for Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by NSF "Internet Atlas" grant ANI-I-9996248, NSF "CAIDA Core"
// NCR-9711092, and by CAIDA memebers.
//==========================================================================

package org.caida.otter;

import java.util.Hashtable;
import java.util.Enumeration;

/**
 * This is a hash of hashes.
 *
 * At the top level a category defines a top level name category.
 *    Under this top level, there may be several key->value pairs.
 *    The following is a schematic of how data is represented.
 *
 *
 *  <pre>
 *  [Set of categories]-  category 1 			(i.e. inbound path info)
 *                     |    |------> attribute1:value1 	(i.e. from host 1: 70%)
 *                     |    |------> attribute2:value2	(i.e. from host 2: 90%)
 *                     |
 *                     -  category 2		(i.e. outbound path info)
 *                     |    |------> attribute1:value1 	(i.e. to host 1: 70%)
 *                     |    |------> attribute2:value2	(i.e. to host 2: 90%)
 *                        etc.
 *  </pre>
 */
public class AttributeSet {

    /**
     * Create an empty attribute set.
     */
    public AttributeSet() {
    }


    /**
     * Add a category to the set.
     *
     * @param category An object that will represent a category in the 
     *   attribute set.
     */
    public Object addCategory(Object category) {

	if(category == null) {
	    throw new IllegalArgumentException("don't pass null");
	}

	return categoryHash.put(category, new Hashtable());
    }

    /**
     * Add an attribute to inside a category
     *
     * @param category The category to insert the new object in.
     * @param attribute The attribute we're setting.
     * @param value The value we're assigning to that attribute.
     * @return The previous value of the attribute, null otherwise.
     */
    public Object addAttribute(Object category, Object attribute, 
				Object value) {

	if(category == null || attribute == null || value == null) {
	    throw new IllegalArgumentException("don't pass null");
	}

	Hashtable attribHash = (Hashtable)categoryHash.get(category);
	if(attribHash == null) {
	    attribHash = new Hashtable();
	    categoryHash.put(category, attribHash);
	}

	return attribHash.put(attribute, value);
    }

    /**
     * Get an attribute from inside a category
     *
     * @param category The category to look for the attribute in.
     * @param attribute The value of the attribute to retrieve.
     * @return The value associated with that attribute, null otherwise.
     */
    public Object getAttribute(Object category, Object attribute) {

	if(category == null || attribute == null) { 
	    throw new IllegalArgumentException("don't pass null");
	}

	Hashtable attribHash = (Hashtable)categoryHash.get(category);
	if(attribHash == null) {
	    return null;
	}

	return attribHash.get(attribute);
    }

    /**
     * Remove a category from the set of categories.
     *
     * @param category The category to remove from the set.
     * @return The category that is being removed, null if it didn't exist.
     */
    public Object deleteCategory(Object category) {
	if(category == null) {
	    throw new IllegalArgumentException("don't pass null");
	}
	return categoryHash.remove(category);
    }

    /**
     * Remove an attribute from inside a category
     *
     * @param category The category to look for the attribute in.
     * @param attribute The attribute to remove from the set.
     * @return The value of the attribute being removed.
     */
    public Object deleteAttribute(Object category, Object attribute) {
	if(category == null || attribute == null) {
	    throw new IllegalArgumentException("don't pass null");
	}

	Hashtable attribHash = (Hashtable)categoryHash.get(category);
	if(attribHash == null) {
	    return null;
	}

	return attribHash.remove(attribute);
    }

    /**
     * Replace an attribute inside a category
     *
     * @param category The category to look for the attribute in.
     * @param attribute The attribute we want to change the value of.
     * @param value The new value to assign to the attribute.
     * @return The old value of the attribute.
     */
    public Object changeAttribute(Object category, Object attribute, 
				Object value) {

	if(category == null || attribute == null || value == null) {
	    throw new IllegalArgumentException("don't pass null");
	}
	return addAttribute(category, attribute, value); 
    }
    
    /**
     * Get a list of objects that represent all of the categories
     *    contained in this node's category set.
     *
     * @return an array of objects that correspond to this set's
     *    categories.
     */
    public Object[] enumerateCategories() {

	Object[] valueArray = new Object[categoryHash.size()];

	Enumeration enumerate = categoryHash.keys();
	for (int i=0; enumerate.hasMoreElements(); i++) {
	    valueArray[i] = enumerate.nextElement();
	}

	return valueArray;
    }

    /** 
     * Get a list of attributes in a given category
     *
     * @param category The category to list the attributes in.
     * @return an array of objects representing the attributes in
     *      the specified category.
     */
    public Object[] enumerateAttributes(Object category) {

	if(category == null) {
	    throw new IllegalArgumentException("Null is not a valid category");
	}
    
	Hashtable attribHash = (Hashtable)categoryHash.get(category);
	if(attribHash == null) {
	    return null;
	}

	Object[] valueArray = new Object[attribHash.size()];

	Enumeration enumerate = attribHash.keys();
	for (int i=0; enumerate.hasMoreElements(); i++) {
	    valueArray[i] = enumerate.nextElement();
	}
	
	return valueArray;
    }

    private Hashtable categoryHash = new Hashtable();
}
