//==========================================================================
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// Permission to use, copy, modify and distribute any part of this Makefile
// for educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above 
// copyright notice, this paragraph and the following paragraphs appear 
// in all copies.
//
// Those desiring to incorporate this into commercial products or use for
// commercial purposes should contact the
//
//      Technology Transfer Office, University of California, San Diego,
//      9500 Gilman Drive, La Jolla, CA 92093-0910
//      Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS PLOTPATH, EVEN 
// IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
//
// THE PLOTPATH, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF 
// CALIFORNIA PLOTPATH NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF 
// ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
// PURPOSE, OR THAT THE USE OF THE SKITTERTRACES WILL NOT INFRINGE ANY 
// PATENT, TRADEMARK OR OTHER RIGHTS.
//
// PlotPaths is developed by John Gallagher Bradley L. Huffaker at the
// University of California, San Diego under the Cooperative Association
// for Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by NSF "Internet Atlas" grant ANI-I-9996248, NSF "CAIDA Core"
// NCR-9711092, and by CAIDA memebers.
//==========================================================================
package org.caida.otter;

import java.util.Vector;

/**
 * A column object.  A container for rows, which are in turn containers for
 * coordinates.  
 */
public class Column {

    /**
     * Create an empty column.
     */
    public Column() {
    }

    /**
     * Create a column with a name.
     *
     * @param colName The name of the new column.
     */
    public Column(String colName) {
	this.colName = colName;
    }

    /**
     * Create a column with an x value.
     *
     * @param x The initial x value of the new column.
     */
    public Column(int x) {
	colXVal = x;
    }

    /**
     * Create a new column and insert a node as the first member.
     *
     * @param newNode The first node to put into the new column.
     */
    public Column(Node newNode) {

	colXVal = newNode.x;

	Row newRow = new Row(newNode);
	rowList.addElement(newRow);
    }

    /**
     * Create a new column and insert a label as the first member
     *
     * @param newLabel The first label to put into the new column.
     */
    public Column(Label newLabel) {

	colXVal = newLabel.x;

	Row newRow = new Row(newLabel);
	rowList.addElement(newRow);
    }

    /**
     * Return an array of the coords in the column
     */
    public Coordinate[] getCoords() {
	// messy but gets the job done.
	//    the better way is to ask all rows how many nodes
	//    they have, then join their arrays into one big one.

	Vector coordVect = getCoordVect();
	Coordinate[] coordList = new Coordinate[coordVect.size()];
	for(int i=0; i < coordList.length; ++i) {
	    coordList[i] = (Coordinate)coordVect.elementAt(i);
	}
	return coordList;
    }

    /**
     * Return the coordinate vector.  
     *
     * @return A Vector of the coordinates stored in this Column.
     */
    public Vector getCoordVect() {

	Vector coordVect = new Vector();
	Row thisRow;
	Vector thisRowCoordList;

	for(int i=0; i < rowList.size(); i++) {

	    thisRow = (Row)rowList.elementAt(i);
	    thisRowCoordList = thisRow.getCoordVect();
	    
	    for(int j=0; j < thisRowCoordList.size(); j++) {
		coordVect.addElement(thisRowCoordList.elementAt(j));
	    }
	}
	return coordVect;
    }
    /**
     * Return the node vector.  
     *
     * @return A Vector of the nodes stored in this column.
     */
    public Vector getNodeVect() {

	Vector nodeVect = new Vector();
	Row thisRow;
	Vector thisRowCoordList;

	for(int i=0; i < rowList.size(); i++) {

	    thisRow = (Row)rowList.elementAt(i);
	    thisRowCoordList = thisRow.getCoordVect();
	    
	    for(int j=0; j < thisRowCoordList.size(); j++) {

		if(thisRowCoordList.elementAt(j) instanceof Node) {
		    nodeVect.addElement(thisRowCoordList.elementAt(j));
		}

	    }
	}
	return nodeVect;
    }

    /**
     * Return the node array.  
     *
     * @return An array of the nodes stored in this column.
     */
    public Node[] getNodes() {

	Vector nodeVect = new Vector();
	Row thisRow;
	Vector thisRowCoordList;

	for(int i=0; i < rowList.size(); i++) {

	    thisRow = (Row)rowList.elementAt(i);
	    thisRowCoordList = thisRow.getCoordVect();
	    
	    for(int j=0; j < thisRowCoordList.size(); j++) {

		if(thisRowCoordList.elementAt(j) instanceof Node) {
		    nodeVect.addElement(thisRowCoordList.elementAt(j));
		}

	    }
	}

	Node[] nodeArray = new Node[nodeVect.size()];
	for(int i=0; i < nodeVect.size(); i++) {
	    nodeArray[i] = (Node)nodeVect.elementAt(i);
	}
	return nodeArray;
    }

    /**
     * Remove a coordinate from this column 
     *
     * @param coord The coordinate to remove from the column.
     */
    public void removeCoord(Coordinate coord) {

	Row thisRow;
	int rowNum;
	for(int i=0; i < rowList.size(); i++) {
	    rowNum = i;
	    thisRow = (Row)rowList.elementAt(i);
	    if(thisRow.containsCoord(coord)) {
		thisRow.removeCoord(coord);
		break;
	    }
	}

    }


    /**
     * Return the label vector.  
     *
     * @return A vector of the labels found in this column.
     */
    public Vector getLabelVect() {

	Vector labelVect = new Vector();
	Row thisRow;
	Vector thisRowCoordList;

	for(int i=0; i < rowList.size(); i++) {

	    thisRow = (Row)rowList.elementAt(i);
	    thisRowCoordList = thisRow.getCoordVect();
	    
	    for(int j=0; j < thisRowCoordList.size(); j++) {

		if(thisRowCoordList.elementAt(j) instanceof Label) {
		    labelVect.addElement(thisRowCoordList.elementAt(j));
		}

	    }
	}
	return labelVect;
    }

    /**
     * Set the X val of everything in column
     *
     * @param new_x The X value to set for this column.
     */
    public void setColumnX(int new_x) {

	Row thisRow;
	colXVal = new_x;

	for(int i=0; i < rowList.size(); i++) {
	    thisRow = (Row)rowList.elementAt(i);
	    thisRow.setRowX(new_x);
	}
    }

    /**
     * Make sure the column contents contain the right X val.
     */
    public void finalizeColumn() {

	Row thisRow;

	for(int i=0; i < rowList.size(); i++) {
	    thisRow = (Row)rowList.elementAt(i);
	    thisRow.setRowX(colXVal);
	}
    }

    /**
     * Shift all the values in a column by the offset.
     *
     * @param offset The offset to shift the X value of this column by.
     @ @return the new column X value.
     */
    public int shiftColumnX(int offset) {

	Row thisRow;
	colXVal += offset;

	for(int i=0; i < rowList.size(); i++) {
	    thisRow = (Row)rowList.elementAt(i);
	    thisRow.shiftRowX(offset);
	}

	return colXVal;
    }


    /**
     * Add a coordinate.
     *
     * @param newCoord The coordinate to add to this column
     */
    public void addCoordinate(Coordinate newCoord) {
	
	newCoord.x = colXVal;

	while(newCoord.rowNum >= rowList.size()) {
	    rowList.addElement(new Row());
	}

	Row insertRow = (Row)rowList.elementAt(newCoord.rowNum);
	insertRow.addCoordinate(newCoord);
    }

    /**
     * Add a node to the column.
     *
     * @param newNode The node to add to the column.
     */
    public void addNode(Node newNode) {
	
	addCoordinate(newNode);

    }

    /**
     * Add a label to the column.
     *
     * @param newLabel The label to add to the column.
     */
    public void addLabel(Label newLabel) {

	addCoordinate(newLabel);

    }
    
    /**
     * Get the x coordinate of the column.
     *
     * @return The X value of this column.
     */
    public int getXLoc() {
	return colXVal;
    }

    /**
     * Generic insert Routine used by addLabel() and addNode()
     *
     * @param coord The coordinate to add to this column
     * @param rowList A Vector of the rows contained by this column.
     *
     * @return The count of nodes in this column.
     */
    private void addCoordinateToRow(Coordinate coord, Vector rowList) {

	boolean coordAdded = false;

	if(rowList.size() == 0) {
	    Row newRow = new Row(coord);
	    rowList.addElement(newRow);
	    return;
	}

	Row thisRow;
	for(int i=0; i < rowList.size(); i++) {
	    thisRow = (Row)rowList.elementAt(i);
	    if(coord.y == thisRow.getYLoc()) {
		thisRow.addCoordinate(coord);
		coordAdded = true;
		break;
	    } else if(coord.y < thisRow.getYLoc()) {
		Row newRow = new Row(coord);
		rowList.insertElementAt(newRow, i);
		coordAdded = true;
		break;
	    } else if(i == rowList.size()-1) {
		Row newRow = new Row(coord);
		rowList.addElement(newRow);
		coordAdded = true;
		break;
	    }
	}

	if(!coordAdded) {
	    System.out.println("[ERROR] coord not added to row");
	}
    }

    /**
     * Return the number of nodes in the column.
     *
     * @return The count of nodes in this column.
     */
    public int getNodeCount() {

	int count = 0;

	Row thisRow;
	for(int i=0; i < rowList.size(); i++) {
	    thisRow = (Row)rowList.elementAt(i);
	    count += thisRow.getNodeCount();
	}
	return count;
    }

    /**
     * Return the number of labels in the column.
     *
     * @return the count of labels found in this column.
     */
    public int getLabelCount() {

	int count = 0;

	Row thisRow;
	for(int i=0; i < rowList.size(); i++) {
	    thisRow = (Row)rowList.elementAt(i);
	    count += thisRow.getLabelCount();
	}
	return count;
    }

    /**
     * Count the number of outbound links from the column.
     *
     * @return the count of outbound links from this column.
     */
    public int countOutboundLinks() {

	int outCount = 0;
	Node[] neighbors;
	Node nodeNeighbor;
	Vector nodeVec = getNodeVect();

	for(int i=0; i < nodeVec.size(); i++) {
	    neighbors = ((Node)nodeVec.elementAt(i)).getNeighborNodes();
	    for(int j=0; j < neighbors.length; j++) {
		nodeNeighbor = neighbors[j];
		if(!nodeNeighbor.colName.equals(colName)) {
		    outCount++;
		}
	    }
	}

	return outCount;
    }

    /**
     * Calculate the bounding boxes containing all the coordinates in the column
     *    and save the results in the boundingBoxes Vector.  Each call to this
     *    method overwrites previous calculations of the bounding boxes.  This
     *    version works on columns before they have been spaced out into their
     *    final X ( and rows to Y) values.  That's why the column number is 
     *    taken as a parmeter.
     *
     * @param colNum The column number this column is in the ColumnSet.
     */
    public void calculateRowColBoundingBoxes(int colNum) {

	// reset the bounding boxes list
	boundingBoxes.removeAllElements();

	if(rowList.size() == 0) {

	    // no work to be done

	} else if(rowList.size() == 1) {

	    Row thisRow = (Row)rowList.elementAt(0);
	    boundingBoxes.addElement(new BoundingBox(colNum-BOUNDING_BOX_MARGIN,
				0-BOUNDING_BOX_MARGIN,
				colNum+BOUNDING_BOX_MARGIN,
				0+BOUNDING_BOX_MARGIN));
	} else {
	    
	    Row thisRow;
    
	    // initialize vars to make compiler happy
	    int thisRowNum	= 0;
	    int lastRowNum	= 0;
	    int boxStart 	= 0;
	    int boxEnd 	= 0;

	    boolean startingNewBox = true;

	    // loop over column nodes making groups 
	    //System.out.println("Starting node group");
	    for(thisRowNum=0; thisRowNum < rowList.size(); thisRowNum++) {

		thisRow = (Row)rowList.elementAt(thisRowNum);

		if(startingNewBox && !thisRow.isEmpty()) {
		    boxStart = thisRowNum;
		    startingNewBox = false;
		}

		if(thisRowNum+1 == rowList.size()) {
		    if(thisRowNum - lastRowNum < BOUNDING_BOX_DENSITY) {

			if(thisRow.isEmpty()) {
			    boxEnd = lastRowNum;
			} else {
			    boxEnd = thisRowNum;
			}

			boundingBoxes.addElement(new BoundingBox(
				colNum-BOUNDING_BOX_MARGIN, 
				boxStart-BOUNDING_BOX_MARGIN,
				colNum+BOUNDING_BOX_MARGIN,
				boxEnd+BOUNDING_BOX_MARGIN));

		    } else {

			boxEnd = lastRowNum;
			boundingBoxes.addElement(new BoundingBox(
				colNum-BOUNDING_BOX_MARGIN, 
				boxStart-BOUNDING_BOX_MARGIN,
				colNum+BOUNDING_BOX_MARGIN,
				boxEnd+BOUNDING_BOX_MARGIN));
    
			boundingBoxes.addElement(new BoundingBox(
				colNum-BOUNDING_BOX_MARGIN,
				thisRowNum-BOUNDING_BOX_MARGIN,
				colNum+BOUNDING_BOX_MARGIN,
				thisRowNum+BOUNDING_BOX_MARGIN));
		    }
		} else {
		    if(thisRowNum - lastRowNum < BOUNDING_BOX_DENSITY) {

			// wait for the end of the bounding box

		    } else if(!startingNewBox){

			boxEnd = lastRowNum;
			boundingBoxes.addElement(new BoundingBox(
				colNum-BOUNDING_BOX_MARGIN, 
				boxStart-BOUNDING_BOX_MARGIN,
				colNum+BOUNDING_BOX_MARGIN,
				boxEnd+BOUNDING_BOX_MARGIN));
			
			startingNewBox = true;
		    }
		}
    
		if(!thisRow.isEmpty()) {
		    lastRowNum = thisRowNum;
		}
	    }
	}
    }

    /**
     * Calculate the bounding boxes containing all the coordinates in the column
     *    and save the results in the boundingBoxes Vector.  Each call to this
     *    method overwrites previous calculations of the bounding boxes. This 
     *    verison is useful after the X (and row Y) values have been set, 
     *    giving a more accurate representation of the bounding boxes.
     *
     */
    public void calculateXYBoundingBoxes() {

	// reset the bounding boxes list
	boundingBoxes.removeAllElements();

	if(rowList.size() == 0) {

	    // no work to be done

	} else if(rowList.size() == 1) {

	    Row thisRow = (Row)rowList.elementAt(0);
	    boundingBoxes.addElement(new BoundingBox(
				colXVal-BOUNDING_BOX_MARGIN,
				thisRow.getYLoc()-BOUNDING_BOX_MARGIN,
				colXVal+BOUNDING_BOX_MARGIN,
				thisRow.getYLoc()+BOUNDING_BOX_MARGIN));
	} else {
	    
	    int lasty;
	    int thisy;
	    int boxStart;
	    int boxEnd;
	    boolean startingNewBox = true;
    
	    // initialize vars to make compiler happy
	    thisy = ((Row)rowList.elementAt(0)).getYLoc();
	    lasty = thisy;
	    boxStart = thisy;
	    boxEnd = thisy;

	    // loop over column nodes making groups 
	    //System.out.println("Starting node group");
	    for(int i=0; i < rowList.size(); i++) {

		if(startingNewBox) {
		    boxStart = thisy;
		    startingNewBox = false;
		}

		thisy = ((Row)rowList.elementAt(i)).getYLoc();
		//System.out.println("Looking at node y : "+thisy);

    
    
		if(i+1 == rowList.size()) {
		    if(thisy - lasty < BOUNDING_BOX_DENSITY) {
			boxEnd = thisy;
			boundingBoxes.addElement(new BoundingBox(
				colXVal-BOUNDING_BOX_MARGIN, 
				boxStart-BOUNDING_BOX_MARGIN,
				colXVal+BOUNDING_BOX_MARGIN,
				boxEnd+BOUNDING_BOX_MARGIN));

		    } else {

			boxEnd = lasty;
			boundingBoxes.addElement(new BoundingBox(
				colXVal-BOUNDING_BOX_MARGIN, 
				boxStart-BOUNDING_BOX_MARGIN,
				colXVal+BOUNDING_BOX_MARGIN,
				boxEnd+BOUNDING_BOX_MARGIN));
    
			boundingBoxes.addElement(new BoundingBox(
				colXVal-BOUNDING_BOX_MARGIN,
				thisy-BOUNDING_BOX_MARGIN,
				colXVal+BOUNDING_BOX_MARGIN,
				thisy+BOUNDING_BOX_MARGIN));
		    }
		} else {
		    if(thisy - lasty < BOUNDING_BOX_DENSITY) {

			// wait for the end of the bounding box

		    } else {

			boxEnd = lasty;
			boundingBoxes.addElement(new BoundingBox(
				colXVal-BOUNDING_BOX_MARGIN, 
				boxStart-BOUNDING_BOX_MARGIN,
				colXVal+BOUNDING_BOX_MARGIN,
				boxEnd+BOUNDING_BOX_MARGIN));
			
			startingNewBox = true;
		    }
		}
    
		lasty = thisy;
	    }
	}
    }

    /** 
     * Overloaded version of doesLineCrossBoundingBox to accept Nodes
     *
     * @param n1 The node representing the "from" coordinate.
     * @param n2 The node representing the "to" coordinate.
     * @return true if the line intersects this column's bounding box, 
     *           false otherwise.
     */
    public boolean doesLineCrossBoundingBox(Node n1, Node n2) {
	return doesLineCrossBoundingBox(n1.x, n1.y, n2.x, n2.y);
    }
    
    /**
     * Test a line against the bounding boxes of the column.  The parameters
     *    represent a line segment from (x1,y1) to (x2,y2).
     *
     * @return true if the line intersects this column's bounding box, 
     *           false otherwise.
     */
    public boolean doesLineCrossBoundingBox(int x1, int y1, int x2, int y2) {

	// query each bounding box if it overlaps with the given line

	boolean result = false;
	for(int i=0; i < boundingBoxes.size(); i++) {
	    if(((BoundingBox)boundingBoxes.elementAt(i)).
			doesLineCrossBoundingBox(x1, y1, x2, y2)) {
		result = true;
		break;
	    }
	}
	return result;
    }

    /**
     * Explicitly synchronize the x and y values of nodes w/ the
     *   row and column they belong to.  Update this columns X value
     *   as a side affect.
     *
     * @param rowStart The offset for the first row in the column.
     * @param colPos The x value of the column.
     * @param rowHeight The spacing in between rows.
     */
    public void setXandYvalues(int rowStart, int colPos, int rowHeight) {

	Row tmpRow;
	colXVal = colPos;

	for(int i=0; i < rowList.size(); i++) {
	    tmpRow = (Row)rowList.elementAt(i);
	    tmpRow.setXandYvalues(colPos, (rowStart + (rowHeight*i)));
	}
    }

    /**
     * Explicitly synchronize the x and y values of nodes w/ the
     *   row and column they belong to.  Use this column's x value
     *   as the default X value.
     *
     * @param rowStart The offset for the first row.
     * @param rowHeight The spacing in between rows.
     */
    public void setXandYvalues(int rowStart, int rowHeight) {

	Row tmpRow;

	for(int i=0; i < rowList.size(); i++) {
	    tmpRow = (Row)rowList.elementAt(i);
	    tmpRow.setXandYvalues(colXVal, (rowStart + (rowHeight*i)));
	}
    }

    /**
     * Set a column's name.
     *
     * @param colName The column's new name.
     */
    public void setName(String colName) {
	this.colName = colName;
    }

    /**
     * Get a column's name
     *
     * @return The column's name.
     */
    public String getName() {
	return colName;
    }

    /**
     * Check if the column contains any coordinates.
     *
     * @return true if the column is empty, false otherwise.
     */
    public boolean isEmpty() {

	boolean isEmpty = true;

	for(int i=0; i < rowList.size(); i++) {
	    if(!((Row)rowList.elementAt(i)).isEmpty()) {
		isEmpty = false;
		break;
	    }
	}
	return isEmpty;
    }

    /**
     * Set a column's index number.
     *
     * @param index The column's new index number.
     */
    public void setColIdx(int index) {

	Row tmpRow;

	for(int i=0; i < rowList.size(); i++) {
	    tmpRow = (Row)rowList.elementAt(i);
	    tmpRow.setColIdx(index);
	}
    }

    /**
     * Request all rows to sort their nodes by parent mean.
     *
     * @param minNodeSpacing The minimun spacing between nodes, so the amount of
     *          space individual nodes are seperated by when spreading a row.
     */
    public void sortNodesByParentMean(int minNodeSpacing, 
			    int maxChildSearchDepth) {

	Row thisRow;

	for(int i=0; i < rowList.size(); i++) {
	    thisRow = (Row)rowList.elementAt(i);
	    thisRow.sortNodesByParentMean(  minNodeSpacing,
					    maxChildSearchDepth);
	}
    }

    /**
     * Insert hidden nodes within the column for links that
     *   are purely vertical.
     *
     * @param colNum Our column number
     */
    public void insertHiddenNodes(int colNum, DisplayObjects data) {

	// only add links when child is above us
	//   (so we account for each pair exactly once)

	Row tmpRow;
	Vector nodeList;
	Node tmpNode;
	Node tmpNeighborNode;
	Node tmpHiddenNode;
	Node[] neighborNodes;

	for(int i=0; i < rowList.size(); i++) {
	    tmpRow = (Row)rowList.elementAt(i);
	    nodeList = tmpRow.getNodeVect();
	    for(int j=0; j < nodeList.size(); j++) {
		tmpNode = (Node)nodeList.elementAt(j);

		//
		// now check each child to see if we have
		// any vertical links in the column
		//

		neighborNodes = tmpNode.getNeighborNodes();

		for(int k=0; k < neighborNodes.length; k++) {
		    tmpNeighborNode = neighborNodes[k];
		    if(tmpNeighborNode.getColumnName() == colName &&
			    tmpNeighborNode.rowNum > tmpNode.rowNum) {

			// add some hidden nodes

			Node lastNeighbor = tmpNode;
			Node nextNeighbor = null;

			for(int m=tmpNode.rowNum+1; m < tmpNeighborNode.rowNum;
				     m++){

			    // the new hidden node
			    tmpHiddenNode = data.addNode(   
					data.getNodeCount(),
					colNum,
					m,	    // row num
					"hidden node"+data.getNodeCount(),
					colName);
			    tmpHiddenNode.setIsHidden(true);

			    lastNeighbor.addNeighborNode(tmpHiddenNode);
			    tmpHiddenNode.addNeighborNode(lastNeighbor);

			    if(m < tmpNeighborNode.rowNum-1) {

			    } else {
				tmpHiddenNode.addNeighborNode(tmpNeighborNode);
				tmpNeighborNode.addNeighborNode(tmpHiddenNode);
			    }

			    addCoordinate(tmpHiddenNode);
			    lastNeighbor = tmpHiddenNode;

			    /*
			    tmpHiddenNode = data.addNode(   
					data.getNodeCount(),
					colNum,
					m,	    // row num
					"hidden node"+data.getNodeCount(),
					colName);

			    // see who our next node will be 
			    //	    (boundary conditions)
			    if(m == tmpNeighborNode.rowNum-1) {
				nextNeighbor = tmpNeighborNode;
			    } else if(m < tmpNeighborNode.rowNum-1) {
				nextNeighbor = tmpHiddenNode;
			    }


			    addCoordinate(tmpHiddenNode);

			    lastNeighbor.addNeighborNode(tmpHiddenNode);
			    tmpHiddenNode.addNeighborNode(nextNeighbor);
			    tmpHiddenNode.addNeighborNode(lastNeighbor);
			    tmpHiddenNode.setIsHidden(true);


			    // boundary conditions
			    if(m == tmpNode.rowNum+1) {
				tmpNode.addNeighborNode(tmpHiddenNode);
			    } 
			    if(m == tmpNeighborNode.rowNum-1) {
				tmpNeighborNode.addNeighborNode(tmpHiddenNode);
			    }

			    lastNeighbor = tmpHiddenNode;
			    */
			}
		    }
		}
	    }
	}
    }

    /**
     * Purge the column contents to the screen
     */
    public void debugColContents() {

	Row thisRow;
	Vector thisCoordVect;
	Coordinate thisCoord;

	System.out.println("*********************************");
	System.out.println("     Column "+colXVal);
	System.out.println("Row            Row Contents");
	System.out.println("---            ---------------------------");
	for(int i=0; i < rowList.size(); i++) {
	    thisRow = (Row)rowList.elementAt(i);
	    thisCoordVect = thisRow.getCoordVect();
	    if(i < 10) {
		System.out.print(i+"         ");
	    } else {
		System.out.print(i+"        ");
	    }
	    for(int j=0; j < thisCoordVect.size(); j++) {
		thisCoord = (Coordinate)thisCoordVect.elementAt(j);
		System.out.print("("+thisCoord.x+", "+thisCoord.y+") ");
	    }
	    System.out.println();
	}
	System.out.println("*********************************\n");
    }

    public void debug() {
	System.out.println("============= begin Column ============");
	System.out.println("[Debug] Column");
	System.out.println("colXVal: "+colXVal);
	System.out.println("colName: "+colName);
	for(int i=0; i < rowList.size(); i++) {
	    ((Row)rowList.elementAt(i)).debug();
	}
	for(int i=0; i < boundingBoxes.size(); i++) {
	    ((BoundingBox)boundingBoxes.elementAt(i)).debug();
	}
	System.out.println("============= end Column ==============");
    }


    private Vector rowList			= new Vector();
    private Vector boundingBoxes 		= new Vector();
    private String colName = null;
    private int colXVal;
    //public static final int BOUNDING_BOX_MARGIN 	= 5;
    public static final int BOUNDING_BOX_MARGIN 	= 0;
    public static final int BOUNDING_BOX_DENSITY 	= 2;
}
