//==========================================================================
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// Permission to use, copy, modify and distribute any part of this Makefile
// for educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above 
// copyright notice, this paragraph and the following paragraphs appear 
// in all copies.
//
// Those desiring to incorporate this into commercial products or use for
// commercial purposes should contact the
//
//      Technology Transfer Office, University of California, San Diego,
//      9500 Gilman Drive, La Jolla, CA 92093-0910
//      Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS PLOTPATH, EVEN 
// IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
//
// THE PLOTPATH, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF 
// CALIFORNIA PLOTPATH NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF 
// ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
// PURPOSE, OR THAT THE USE OF THE SKITTERTRACES WILL NOT INFRINGE ANY 
// PATENT, TRADEMARK OR OTHER RIGHTS.
//
// PlotPaths is developed by John Gallagher Bradley L. Huffaker at the
// University of California, San Diego under the Cooperative Association
// for Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by NSF "Internet Atlas" grant ANI-I-9996248, NSF "CAIDA Core"
// NCR-9711092, and by CAIDA memebers.
//==========================================================================
package org.caida.otter;

import java.util.Vector;
import java.util.Comparator;
import java.util.Enumeration;

//
// Space columns by node count () uses a default column width calculated 
// empirically.
//
// This is icky when I should use the columnWidth convention as in other 
// routines.
//

/** 
 * Encapsulates a DisplayObjects object, it keeps track of elements in
 *   terms of columns as well.  Columns may be treated as a single unit
 *   when managing the location of nodes on the display.
 */
public class ColumnSet {

    /**
     * Create an empty ColumnSet.
     */
    public ColumnSet(int columnWidth, int rowHeight) {
	this.columnWidth = columnWidth;
	this.rowHeight = rowHeight;
    }

    /** 
     * Initialize a new ColumnSet object with a pre-made
     *   DisplayObjects object.  Values are copied out of
     *   the argument, such that the encapsulated data
     *   has its own copy.
     *
     * @param columnWidth The space between columns.
     * @param rowHeight The space between rows.
     * @param displayData The DisplayObjects object to sort into rows and 
     *    columns.
     */
    public ColumnSet(int columnWidth, int rowHeight, 
		    DisplayObjects displayData) {

	this.columnWidth = columnWidth;
	this.rowHeight = rowHeight;

	addDisplayObjects(displayData);
	//setDisplayObjects(displayData);

	setXandYvalues();
    }

    /** 
     * Space out overlapping nodes by looking at their parent
     *   and sorting nodes by the mean of the parent.
     *
     * @param minNodeSpacing The minimum space between nodes when spreading a 
     *     row.
     */
    public void spaceNodesByParentMean(int minNodeSpacing, 
			    int maxChildSearchDepth) {

    
	// traverse from one column to the next.
	Column thisCol;

	for(int i=0; i < columnList.size(); i++) {
	    thisCol = (Column)columnList.elementAt(i);
	    thisCol.sortNodesByParentMean(  minNodeSpacing,
					    maxChildSearchDepth);
	}
    }

    /**
     * Select the next column to be placed by the column placing algorithm
     *
     * @param placedColumnSet array of bools corresponding to columnList
     *          indicating whether or not that column has been place.
     * @return the index of the column with the most outbound links
     *          (links to columns other than itself)
     */
    private int selectNextColumnToPlace(boolean[] placedColumnSet)
    {
	int largestOutboundCount = -1;
	int thisOutboundCount;
	int placingColumnIdx = 0;
	for(int i=0; i < columnList.size(); i++) {
	    if(!placedColumnSet[i]) {

		thisOutboundCount = ((Column)columnList.elementAt(i)).
					countOutboundLinks();

		if(largestOutboundCount < thisOutboundCount) {

		    largestOutboundCount = thisOutboundCount;
		    placingColumnIdx = i;

		}
	    }
	}
	return placingColumnIdx;
    }

    /**
     * Given an index of a column to place from the columnList, decide where
     *    to place it into the placed set, and do it.
     *
     * @param placingColumnIdx the index of the column to place from columnList
     * @param placedColumnSet array of booleans showing which columns from
     *             columnList have already been placed.
     * @param placedColumnOrder the new order of the columns after being placed
     *             into the placed set.
     * @param placedColumnIndex vector of indicies mapping the new placement
     *             to the old placement.  The first value in the vector 
     *             represents the index into the current columnList where the
     *             new first element will be found.
     */
    private void placeNextColumn(   int placingColumnIdx,
				    boolean[] placedColumnSet,
				    Vector placedColumnOrder, 
				    Vector placedColumnIndex) {

	int bestPosition    = 0;  // index of best column placement so far
	int bestScore       = -1; // heuristic score 
	int bestScoreDist   = -1; // X distance used to break score
	int thisScore;
	int thisDist;
	Node[] nodeIdxList;
	Node[] nodeList; 
	Node node1, node2;
	Column biggestNeighbor;
	Column placingColumn;
	Column tmpCol;
	placingColumn = (Column)columnList.elementAt(placingColumnIdx);


	for(    int testingPosition = 0; 
		testingPosition <= placedColumnOrder.size(); 
		testingPosition++) {


	    // set X coord of each column so that bounding 
	    // box test has correct coords
	    for(int i=0; i < testingPosition; i++) {
		((Column)placedColumnOrder.elementAt(i)).
			setColumnX(columnWidth * i);
		((Column)placedColumnOrder.elementAt(i)).
			calculateRowColBoundingBoxes(i);
	    }
	    placingColumn.setColumnX(columnWidth * testingPosition);
	    for(int i=testingPosition; i < placedColumnOrder.size(); i++) {
		((Column)placedColumnOrder.elementAt(i)).
			setColumnX(columnWidth * (i+1));
		((Column)placedColumnOrder.elementAt(i)).
			calculateRowColBoundingBoxes(i+1);
	    }



	    thisScore = 0;
	    thisDist = -1;

	    // to extract the scoring into another method (or class)
	    // I'm thinking it would need to return some sort of a 
	    // score object, b/c right now we really use two scores...
	    // and an object would let us use even more parameters
	    // when placing the column.

	    for(int i=0; i < placedColumnOrder.size(); i++) {
		tmpCol = (Column)placedColumnOrder.elementAt(i);

		// calculate all the lines from this column 
		// that cross other bounding boxes

		nodeList = placingColumn.getNodes();
		for(int j=0; j < nodeList.length; j++) {

		    node1 = nodeList[j];
		    nodeIdxList = node1.getNeighborNodes();

		    for(int k=0; k < nodeIdxList.length; k++) {
			node2 = nodeIdxList[k];


			if((Column)columnList.elementAt(node2.getColNum()) 
				    != tmpCol && 
				placedColumnSet[node2.getColNum()] 
				    == true) {

			    if(tmpCol.doesLineCrossBoundingBox(
				    node1, node2)) {
				thisScore++;
			    }
			    if(node1.x > node2.x) {
				thisDist += node1.x - node2.x;
			    } else {
				thisDist += node2.x - node1.x;
			    }
			}
		    }
		}

		
		// calculate all the other lines that cross 
		// this column's bounding boxes
		nodeList = tmpCol.getNodes();
		for(int j=0; j < nodeList.length; j++) {

		    node1 = nodeList[j];
		    nodeIdxList = node1.getNeighborNodes();

		    for(int k=0; k < nodeIdxList.length; k++) {

			node2 = nodeIdxList[k];

			if((Column)columnList.elementAt(node2.getColNum()) 
				    != placingColumn && 
				placedColumnSet[node2.getColNum()] 
				    == true) {

			    if(placingColumn.doesLineCrossBoundingBox(
				    node1, node2)) {
				thisScore++;
			    }
			    if(node1.x > node2.x) {
				thisDist += node1.x - node2.x;
			    } else {
				thisDist += node2.x - node1.x;
			    }
			}
		    }
		}
	    }


	    // see if this is the best position so far
	    if((thisScore < bestScore || bestScore == -1) && 
			    thisDist != -1) {

		bestScore = thisScore;
		bestPosition = testingPosition;
		bestScoreDist = thisDist;

	    } else if (thisScore == bestScore && 

			    thisDist < bestScoreDist) { 
		bestPosition = testingPosition;
		bestScoreDist = thisDist;
	    }
	}


	// actually place the column in the position we chose
	placedColumnOrder.insertElementAt(
		placingColumn, bestPosition);

	placedColumnIndex.insertElementAt(
		new Integer(placingColumnIdx), bestPosition);

	placedColumnSet[placingColumnIdx] = true; 
    }

    /**
     * Order columns trying to minimize the number of overlapping links.
     *   This is the second try at optimizing the link crossing.  Before
     *   any link crossing a column was considered a problem. Now scoring
     *   will only count links that overlap with a bounding box of the column
     *   Links may now cross a column with no penalty provided they do not
     *   enter one of the bounding boxes that specifies a colunmn
     */
    public void orderColumnsByBoundingBoxes() {

	if(columnList.size() <= 1) {
	    return;
	}

	/////////////////////////
	// Calculate Meta-data //
	/////////////////////////

	// create bounding boxes in each of the columns

	Column tmpCol;
	for(int i=0; i < columnList.size(); i++) {
	    tmpCol = (Column)columnList.elementAt(i);
	    tmpCol.calculateRowColBoundingBoxes(i);
	    //tmpCol.debugBoundingBoxes();
	}

	// keep track of who has been placed already
	boolean[] placedColumnSet = new boolean[columnList.size()];
	for(int i=0; i < placedColumnSet.length; i++) {
	    placedColumnSet[i] = false;
	}
	// keep Vector of Integers to represent placed order
	Vector placedColumnOrder = new Vector();
	Vector placedColumnIndex = new Vector();


	/////////////////////////////////
	// Place columns one at a time //
	/////////////////////////////////

	boolean stillPlacingColumns = true;
	int placingColumnIdx = 0;
	while(stillPlacingColumns) {


	    /////////////////////////////////////////////////////
	    //  Select the column with the most outbound links //
	    /////////////////////////////////////////////////////
	    placingColumnIdx = selectNextColumnToPlace(placedColumnSet);


	    ///////////////////////////////////////////
	    // try all positions to place new column //
	    ///////////////////////////////////////////
	    placeNextColumn(	placingColumnIdx, 
				placedColumnSet,
				placedColumnOrder,
				placedColumnIndex);

	    // see if all the columns have been placed 
	    stillPlacingColumns = false;
	    for(int i=0; i < placedColumnSet.length; i++) {
		if(!placedColumnSet[i]) {
		    stillPlacingColumns = true;
		    break;
		}
	    }
	}



	//////////////////////////////////////////
	// Write new column order to class var  //
	//////////////////////////////////////////

	int colNum;

	Vector resultVec = new Vector();
	for(int i=0; i < placedColumnOrder.size(); i++) {
	    ((Column)placedColumnOrder.elementAt(i)).setColumnX(
		    columnWidth * i);
	}
	columnList = placedColumnOrder;

	setXandYvalues(0, columnWidth, rowHeight);
    }


    /**
     * helper method to find distance between columns.
     *
     * @param col1 The first column to compare.
     * @param col2 The second column used in the comparison.
     * @return The distance between the two specified columns.
     */
    public int findDistanceBetweenColumns(Column col1, Column col2) {

	int tmpDist = col1.getXLoc() - col2.getXLoc();
	if(tmpDist < 0 ) {
	    tmpDist *= -1;
	}
	return tmpDist;
    }


    /**
     * Count the lines leaving col1, that cross bounding boxes belonging 
     * to col2.
     *
     * @param col1 Lines are used from this column.
     * @param col2 BoundingBoxes are used from this column.
     * @return The number of intersections that lines from col1 have with 
     *      bounding boxes belonging to col2.
     */
    private int countLinesCrossingBoundingBox(Column col1, Column col2) {
    
	int intersectCount = 0;

	Node[] nodeList = col1.getNodes();
	Node startNode, endNode;
	Node[] neighborList;
	
	for(int i=0; i < nodeList.length; i++) {
	    startNode = nodeList[i];
	    neighborList = startNode.getNeighborNodes();
	    for(int j=0; j < neighborList.length; j++) {
		endNode = neighborList[j];
		if(col2.doesLineCrossBoundingBox(startNode, endNode)) {
		    intersectCount++;
		}
	    }
	}
	return intersectCount;
    }

    /**
     * Count the lines between col1 and col2, that cross bounding 
     *    boxes belonging to col3.
     *
     * @param col1 used to find links going to col2.
     * @param col2 used to find links going to col1.
     * @param col3 used for BoundingBoxes.
     * @return The number of intersections that lines between col1 and col2
     *    have with bounding boxes belonging to col3.
     */
    private int countLinesCrossingBoundingBox(
			    Column col1, Column col2, Column col3) {

	int intersectCount = 0;

	Node[] nodeList = col1.getNodes();
	Node startNode, endNode;
	Node[] neighborList;
	
	for(int i=0; i < nodeList.length; i++) {
	    startNode = nodeList[i];
	    neighborList = startNode.getNeighborNodes();
	    for(int j=0; j < neighborList.length; j++) {
		endNode = neighborList[j];
		if(endNode.x == col2.getXLoc()) {
		    if(col3.doesLineCrossBoundingBox(startNode, endNode)) {
			intersectCount++;
		    }
		}
	    }
	}
	return intersectCount;
    }

    /**
     * Swap positions of two columns.
     */
    private void swapColumns(int col1num, int col2num) {

	int x1, x2;
	Column col1, col2;

	col1 = (Column)columnList.elementAt(col1num);
	col2 = (Column)columnList.elementAt(col2num);

	x1 = col1.getXLoc();
	x2 = col2.getXLoc();

	// swap internal values
	col1.setColumnX(x2);
	col2.setColumnX(x1);

	// swap order in class Vector columnList

	columnList.removeElementAt(col2num);
	columnList.insertElementAt(col1, col2num);
	columnList.removeElementAt(col1num);
	columnList.insertElementAt(col2, col1num);

    }

    /**
     * Space each column in the DisplayObjects evenly.
     *
     * @param min_x The starting X value used.
     * @param max_x The ending X value used.
     */
    public void spaceColumnsEvenly(int min_x, int max_x) {

	// find out how many columns we have.
	int columnCount = columnList.size();

	// find out how many pixels each column gets.
	// (space between first and last column, these two will not move)

	int pixelPerColumn = (max_x - min_x) / columnCount;

	for(int i = 0; i < columnList.size(); i++) {
	    ((Column)columnList.elementAt(i)).
			setColumnX(min_x + (i*pixelPerColumn));
	}
    }


    /**
     * Space each column in the DisplayObjects evenly.
     *
     * @param colSpacing The space to put between neighboring columns.
     */
    public void spaceColumnsEvenly(int colSpacing) {

	for(int i = 0; i < columnList.size(); i++) {
	    ((Column)columnList.elementAt(i)).setColumnX(i*colSpacing);
	}
    }

    /**
     * Space each column in the DisplayObjects evenly using the default
     *   column spacing.
     */
    public void spaceColumnsEvenly() {
	spaceColumnsEvenly(columnWidth);
    }

    /**
     * Space columns by node count with default column width
     *
     * @param minNodeSpacing The minimum space between nodes when they are 
     *         spread along a row.  This is needed to estimate the width 
     *         of each column, and then its new width considering its size 
     *         in nodes.
     */
    public void spaceColumnsByNodeCount(int minNodeSpacing) {
	spaceColumnsByNodeCount(minNodeSpacing, columnWidth);
    }

    /**
     * Update X coords of nodes such that each column
     *   receives space on account of how many nodes
     *   are contained on one level.  The X coords will
     *   be distributed among the same range as before,
     *   but the location of nodes within that range
     *   will change.
     *
     * Columns are also spead out to accomidate the new
     *   width of the column - especially if the new
     *   column is a lot fatter.
     *
     * @param minNodeSpacing The minimum space between nodes when they are 
     *         spread along a row.  This is needed to estimate the width of 
     *         each column, and then its new width considering its size in 
     *         nodes.
     * @param minColSpacing The minimum column Spacing regardless of node count.
     */
    public void spaceColumnsByNodeCount(int minNodeSpacing, int minColSpacing) {
	
	int colOffset = 0;
	int thisColOffset;

	// Assign each column an initial coordinate by giving all of them
	// equal space
	spaceColumnsEvenly(minColSpacing);

	if(columnList.size() < 2) {
	    // no work to do
	    return;
	}

	// Iterate over each column, spreading columns if 
	//     they exceed the threshold
	int maxOccupancy;
	int occupancy;
	int thisY;
	int lastY;
	Vector thisColumn;
	Column realThisCol;
	for(int i=0; i < columnList.size(); i++) {
	    maxOccupancy = 0;
	    thisColumn = (Vector)((Column)columnList.elementAt(i)).
		    getCoordVect();
	    realThisCol = (Column)columnList.elementAt(i);
	    thisColOffset = 0;
	    

	    // find the maximum occupancy of a row in each column
	    lastY = realThisCol.getNodes()[0].y;
	    occupancy = 1;
	    for(int j=1; j < thisColumn.size(); j++) {
		thisY = ((Coordinate)thisColumn.elementAt(j)).y;
		if(thisY == lastY) {
		    occupancy++;
		} else {
		    if(occupancy > maxOccupancy) {
			maxOccupancy = occupancy;
		    }
		    occupancy = 0;
		    lastY = thisY;
		}
	    }
	    
	    // see if we need to make the column bigger
	    if((maxOccupancy * minNodeSpacing) + minColSpacing > 
		    minColSpacing) {

		thisColOffset = maxOccupancy * minNodeSpacing - 
		    minColSpacing + minColSpacing;
		
		realThisCol.shiftColumnX((thisColOffset / 2) + colOffset);

	    } else {

		realThisCol.shiftColumnX(colOffset);

	    }

	    colOffset += thisColOffset;
	}
    }

    
    /**
     * Add a node to the encapsulated DisplayObjects object,  
     *   and include the node in it's appropriate column and row.     
     *
     * @param node The new node.
     */
    public void addNode(Node node) {
	addCoordinateToColumn(node.getColumnName(), node);
    }

    /**
     * Insert a coordinate into the row and column data structure.
     *   Don't touch the DisplayObjects object though.
     *
     * @param node The new node.
    public void updateNode(Node node) {
	addCoordinateToColumn(node.colName, node);
    }
     */

    /**
     * Disperse an entire column into the rest of the graph.  Here
     *   nodes are found in a "not found" column are sorted into the
     *   column that they have the most links into.  This makes the 
     *   resulting picture better, as well as being a best guess where
     *   the node really belongs.
     */
    public void disperseNotFoundColumn(Vector disperseColList) {

	/* make sure nodes are not placed back into NOT FOUND column */
	/* could be avoided if the node knew which column it belonged to */

	Column tmpCol;
	Coordinate tmpCoord;

	// column we are dispersing
	Column disperseCol;	

	// candidate to transfer coord into
	Column targetCol;	

	// the coordinate we are currently moving
	Coordinate thisCoord;	

	// holds the coordinates of the col we're tinkering with
	Coordinate[] coordList;	

	// counts which neighbor column is the best candidate
	int targetScore;	

	// current running total
	int tmpScore;		

	// keep track of which column we're scoring currently
	int lastNeighborColNum; 

	// stores all the neighbors for the current node
	Node[] neighborList;	

	Comparator compare;

	for(int i=0; i < columnList.size(); i++) {

	    tmpCol = (Column)columnList.elementAt(i);
	    
	    // if column is listed as a column to disperse
	    if(tmpCol.getName() != null && 
		    vectContainsString(tmpCol.getName(), disperseColList)) {

		disperseCol = tmpCol;
		coordList = disperseCol.getCoords();
		compare = new CompareCoordsByY();
		PlotPathsUtil.quicksort(coordList, compare);
		
		// try to assign all nodes in disperseCol to another one
		//     loop over list backwards so that children are more 
		//     likely not to be in the disperse column 
		//     (they've already been touched)
		for(int j=0; j < coordList.length; ++j) {

		    thisCoord = coordList[j];
		    targetScore 	= -1;
		    targetCol 	= null;
		    

		    // find which neighboring AS is the most popular
		    if(thisCoord instanceof Node && 
				    ((Node)thisCoord).hasNeighborNodes()) {

			neighborList = ((Node)thisCoord).getNeighborNodes();

			// sort neighbors by column number
			compare = new CompareCoordsByColNum();
			PlotPathsUtil.quicksort(neighborList, compare);

			// find the column we have the most links into

			if(neighborList.length > 0) {
			    lastNeighborColNum = neighborList[0].colNum;
			} else {
			    lastNeighborColNum = -1;
			}

			tmpScore = 0;

			for(int k=0; k < neighborList.length; k++) {

			    tmpCoord = neighborList[k];

			    if(tmpCoord.colNum == lastNeighborColNum ) {
				tmpScore++;
			    }

			    if(tmpCoord.colNum != lastNeighborColNum) {
				if(tmpScore > targetScore && 
					!columnList.elementAt(
					lastNeighborColNum)
					.equals(disperseCol)) {

				    targetCol = (Column)columnList.elementAt(
					lastNeighborColNum);
				    targetScore = tmpScore;

				    tmpScore = 0;
				    lastNeighborColNum = tmpCoord.colNum;
				}
			    }

			    // boundary condition
			    if(k == neighborList.length-1) {
				if(tmpScore > targetScore && 
					!columnList.elementAt(tmpCoord.colNum)
					.equals(disperseCol)) {

				    targetCol = (Column)columnList.elementAt(
					tmpCoord.colNum);
				}
			    }
			}
    
			// move node to new column
			if(targetCol != null) {
			    disperseCol.removeCoord(thisCoord);
			    thisCoord.setColumnName(targetCol.getName());
			    addCoordinateToColumn(targetCol.getName(), 
					thisCoord);
			} else {
			    System.out.println(
				    "[Error]: disperse col could not find "+
				    "a valid target: node \""+
				    thisCoord.getName()+"\"");
			}

		    } else if (thisCoord instanceof Label) {
			System.out.println("Removing label "+
			    thisCoord.getName());
			disperseCol.removeCoord(thisCoord);
			data.removeLabel((Label)thisCoord);
		    } else {
			System.out.println("Coord found with no neighbors in"+
			    " the remove column: node \""+thisCoord.getName()+
			    "\"");
		    }

		}
		
		// if column has no more nodes, delete the column

		if(disperseCol.isEmpty()) {
		    System.out.println("Removing column "+
				    disperseCol.getName());

		    removeColumnAt(i);
		    i--;
		} else {
		    System.out.println("Disperse col "+
				    disperseCol.getName()+" still has nodes");

		    //disperseCol.debug();
		}
	    }
	}
    }

    private boolean vectContainsString(String str, Vector vect) {
	for(int i=0; i < vect.size(); i++) {
	    if(str.equals((String)vect.elementAt(i))) {
		return true;
	    }
	}
	return false;
    }
    
    private void removeColumnAt(int index) {

	Column tmpCol;

	columnList.removeElementAt(index);
	for(int i=index; i < columnList.size(); i++) {
	    tmpCol = (Column)columnList.elementAt(i);
	    tmpCol.setColIdx(i);
	}
    }

    /**
     * Add a line to the encapsulated DisplayObjects object,  
     *   but no column specific stuff is done.  The ColumnSet 
     *   behavior is pretty transparent here.  The real work  
     *   is still done by the DisplayObject.                  
     *
     * @param line Line to add to DisplayObjects and the column / row object
     */
    public void addLine(Line line) {
	//updateNodeConnections(	line, line.getCoord1(), line.getCoord2());

    }

    /**
     * Update connection info but don't actually chage the DisplayObjects data.
     *
     * @param line Line to add to DisplayObjects and the column / row object
    public void updateLine(Line line) {
	updateNodeConnections(	line, 
				line.getCoord1(), 
				line.getCoord2());
    }
    
    private void updateNodeConnections(Line line, Coordinate coord1, 
					    Coordinate coord2) {
    
	// associate nodes with each other and the link between them
	if(coord1 instanceof Node && coord2 instanceof Node) {

	    ((Node)coord1).addNeighborNode((Node)coord2);
	    ((Node)coord2).addNeighborNode((Node)coord1);
	}
    }
     */


    /**
     * Add a label to the encapsulated DisplayObjects object, 
     *   and include the label in it's appropriate column.    
     *
     * @param label The new label
     */
    public void addLabel(Label label) {
	addCoordinateToColumn(label.getColumnName(), label);
    }

    /**
     * Add a label to row and column data structure, but don't touch 
     *   the actual DisplayObjects data.
     *
     * @param label The new label
    public void updateLabel(Label label) {
	addCoordinateToColumn(label.colName, label);
    }
     */

    /**
     * Return the encapsulated DisplayObjects object.
     *   Useful when we're finally ready to draw and
     *   we want to see what we have.  The object
     *   returned is not a copy, its the real thing.
     */
    public DisplayObjects getDisplayObjects() {
	
	return data;
    }

    /**
     * Set our respective coordinates X and Y values based on their row and 
     *     column.
     *
     * @param columnStart The offset for the first column.
     * @param coumnSpacing The spacing between columns.
     * @param rowSpacing The spacing between rows.
     */
    public void setXandYvalues(int columnStart, int columnSpacing, 
				    int rowSpacing) {

	Column tmpCol;

	// update our class vars to reflect that they want these values.
	columnWidth = columnSpacing;
	rowHeight = rowSpacing;
    
	for(int i=0; i < columnList.size(); i++) {
	    tmpCol = (Column)columnList.elementAt(i);
	    tmpCol.setXandYvalues(0, (columnStart+(i*columnSpacing)), 
				    rowSpacing);
	}
    }

    /**
     * Set our respective coordinates X and Y values with default values for
     *   the row and column, starting at the specified offset.
     *
     * @param columnStart The offset of the first column.
     */
    public void setXandYvalues(int columnStart) {
	setXandYvalues(0, columnWidth, rowHeight);
    }

    /**
     * Set our respective coordinates X and Y values with default values for
     *   the row and column, starting a 0.
     */
    public void setXandYvalues() {
	setXandYvalues(0 ,columnWidth, rowHeight);
    }
    
    /** 
     * Add data to the column object and its encapsulted
     *   DisplayObjects object by specifying a DisplayObjects
     *   object.  Values are actually copied out of the
     *   argument and into the encapsulated data object.
     *
     * @param displayData The DisplayObjects object to process
     */
    public void addDisplayObjects(DisplayObjects displayData) {
	
	data = displayData;

	Enumeration nodeEnum = data.getNodeEnumeration();
	while(nodeEnum.hasMoreElements()) {
	    addNode((Node)nodeEnum.nextElement()); 
	}

	Enumeration labelEnum = data.getLabelEnumeration();
	while(labelEnum.hasMoreElements()) {
	    addLabel((Label)labelEnum.nextElement());
	}

	Enumeration lineEnum = data.getLineEnumeration();
	while(lineEnum.hasMoreElements()) {
	    addLine((Line)lineEnum.nextElement());
	}
    }


    /** 
     * Use an already built display object
     *
     * @param displayData The DisplayObjects object to process.
    public void setDisplayObjects(DisplayObjects displayData) {
	
	data = displayData;

	for(int i=0; i < displayData.nodes.size(); i++) {
	    updateNode((Node)displayData.nodes.elementAt(i)); 
	}
	for(int i=0; i < displayData.lines.size(); i++) {
	    updateLine((Line)displayData.lines.elementAt(i));
	}
	for(int i=0; i < displayData.labels.size(); i++) {
	    updateLabel((Label)displayData.labels.elementAt(i));
	}
    }
     */


    /**
     * Add a coordinate to the column that it's column name given.
     *
     * @param colName The column to add the coordinate to.
     * @param coord The coordinate being added.
     */
    private void addCoordinateToColumn(String colName, Coordinate coord) {

	Column tmpCol;
	boolean nodeAdded = false;

	for(int i=0; i < columnList.size(); i++) {

	    tmpCol = (Column)columnList.elementAt(i);

	    if( tmpCol.getName().equals(colName)) {
		tmpCol.addCoordinate(coord);
		coord.setColumnNum(i);
		nodeAdded = true;
	    }

	}

	if(!nodeAdded) {
	    Column newCol = new Column(colName);
	    newCol.addCoordinate(coord);
	    coord.setColumnNum(columnList.size());
	    columnList.addElement(newCol);
	}
    }

    /**
     * Insert hidden nodes into all columns that have overlapping vertical
     *   links on nodes.
     */
    public void insertHiddenNodes() {

	Column tmpCol;
	for(int i=0; i < columnList.size(); i++) {
	    tmpCol = (Column)columnList.elementAt(i);
	    tmpCol.insertHiddenNodes(i, data);
	}
    }

    public void debug() {
	System.out.println("=========== start ColumnSet ========");
	System.out.println("[Debug] Column Set");
	for(int i=0; i < columnList.size(); i++) {
	    ((Column)columnList.elementAt(i)).debug();
	}
	System.out.println("=========== end ColumnSet ==========");
    }

    public void debugColSize() {

	Column tmpCol;
	System.out.println("======================================");
	for(int i=0; i<columnList.size(); i++) {
	    tmpCol = (Column)columnList.elementAt(i);
	    System.out.println("Column "+i+" node  count:"+
			    tmpCol.getNodeCount()+" X coord:"+tmpCol.getXLoc());
	    System.out.println("Column "+i+" label count:"+
			    tmpCol.getLabelCount());
	}
	System.out.println("======================================");
    }

    /*
    private static void quickSortNodesByColNum(DisplayObjects data, 
				int[] sortArray) {

	quickSortNodesByColNum(data, sortArray, 0, sortArray.length-1);
    }

    private static void quickSortNodesByColNum(DisplayObjects data, 
				int[] sortArray, int leftEnd, int rightEnd) {

	if(rightEnd - leftEnd <= 0) {
	    return;
	} else if((rightEnd - leftEnd) == 1) {
	    // short circuit for 2 element case
	    if(((Coordinate)data.nodes.elementAt(
			    sortArray[leftEnd])).colNum > 
			    ((Coordinate)data.nodes.elementAt(
			    sortArray[rightEnd])).colNum) {

		// exchange
		int tmp = sortArray[leftEnd];
		sortArray[leftEnd] = sortArray[rightEnd];
		sortArray[rightEnd] = tmp;
	    }
	    return;
	}
	
	int pivot = leftEnd;
	int newLeftEnd = pivot;			// pick first element as pivot 
	int tmp;
	
	for(int i = leftEnd+1; i <= rightEnd; i++) {
	    if(((Coordinate)data.nodes.elementAt(
			    sortArray[i])).colNum  < 
			    ((Coordinate)data.nodes.elementAt(
			    sortArray[pivot])).colNum) {


		newLeftEnd++;

		if(i != newLeftEnd) {

		    // exchange
		    tmp = sortArray[newLeftEnd];
		    sortArray[newLeftEnd] = sortArray[i];
		    sortArray[i] = tmp;
		}
	    }
	}

	tmp = sortArray[newLeftEnd];
	sortArray[newLeftEnd] = sortArray[pivot];
	sortArray[pivot] = tmp;

	quickSortNodesByColNum(data, sortArray, leftEnd, newLeftEnd-1);
	quickSortNodesByColNum(data, sortArray, newLeftEnd+1, rightEnd);
    }
    */

    ////////////////////////////////////////
    //           Class Variables          //
    ////////////////////////////////////////

    // encapsulated display coords object
    private DisplayObjects data = new DisplayObjects();
    
    // contains Vectors, each one represents an entire column
    private Vector columnList = new Vector();


    private int columnWidth;
    private int rowHeight;
}
