//==========================================================================
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// Permission to use, copy, modify and distribute any part of this Makefile
// for educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above 
// copyright notice, this paragraph and the following paragraphs appear 
// in all copies.
//
// Those desiring to incorporate this into commercial products or use for
// commercial purposes should contact the
//
//      Technology Transfer Office, University of California, San Diego,
//      9500 Gilman Drive, La Jolla, CA 92093-0910
//      Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS PLOTPATH, EVEN 
// IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
//
// THE PLOTPATH, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF 
// CALIFORNIA PLOTPATH NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF 
// ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
// PURPOSE, OR THAT THE USE OF THE SKITTERTRACES WILL NOT INFRINGE ANY 
// PATENT, TRADEMARK OR OTHER RIGHTS.
//
// PlotPaths is developed by John Gallagher Bradley L. Huffaker at the
// University of California, San Diego under the Cooperative Association
// for Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by NSF "Internet Atlas" grant ANI-I-9996248, NSF "CAIDA Core"
// NCR-9711092, and by CAIDA memebers.
//==========================================================================
package org.caida.otter;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Stack;


/**
 *  Represent a directed graph, support simple queries on the graph
 *
 *    The name field of the graph node should/will be unique
 *    because the nodes are kept track of with a hashtable 
 *    keying on name.
 *
 *    Constructors do a shallow copy for new elements.
 * 
 * @author John Gallagher
 */
public class DirectedGraph {
    
    /** Construct an empty graph */
    public DirectedGraph() {
	nodeBasket = new Hashtable();
    }

    /** Add a path to the graph.
     *    Adds a series of nodes that are connected
     *    directionally from first to last.
     *    Ignores redundancies, inserts new data.
     *
     * @param graphNodeList Vector of nodes specifying a path
     *        from the first element of the array to the last
     * @return true if success, false if failue to add all data
     */
     public boolean addPath(GraphNode root, String source) {
	Stack nodeStack = new Stack();
	GraphNode currentNode;
	Vector children;
	    
	// add the source to our list if necessary
	SourceObj srcObj = (SourceObj)sourceHashGet(source);

	if(srcObj == null) {

	    srcObj = new SourceObj();
	    sourceHashPut(source, srcObj);

	}

	sourceHashPut(source, srcObj);
	nodeStack.push(root);

	while(!nodeStack.empty()) {
	    currentNode = (GraphNode)nodeStack.pop();

	    // prep children for next iteration
	    children = currentNode.getChildren(source);

	    if(children != null) {

		for(int i = children.size()-1;i >= 0; i--) {
		    nodeStack.push((GraphNode)children.elementAt(i));
		}

	    }

	    //  validate node (make sure it's in the graph)
	    if(!nodeBasket.containsKey(currentNode.getName())) {

		nodeBasket.put(currentNode.getName(), currentNode);
		// validate children (make sure they're in the graph)
		currentNode.validateChildren(source, this);

	    } else {

		((GraphNode)nodeBasket.get(currentNode.getName())).
			    mergeChildren(source, currentNode);
		((GraphNode)nodeBasket.get(currentNode.getName())).
			    validateChildren(source, this);
	    }


	}
	return true;
    }


    /** 
     * Retrieve all the neighbors of a node
     *
     * @param name Name of the GraphNode to return the children of.
     * @return Vector of nodes that are neighbors, null
     *          if the node does not exist.
     */
    public Vector getChildren(String source, String name) {

	GraphNode target = (GraphNode)nodeBasket.get(name);
	if(target != null) {
	    return target.getChildren(source);
	}
	return null;

    }
    
    /** 
     * Return the count of nodes in the graph 
     *
     * @return The size of the graph.
     */
    public int getGraphSize() {
	return nodeBasket.size(); 
    }
    
    /** 
     * Return array of all nodes 
     * 
     * @return An array of all the nodes in the graph
     */
    public GraphNode[] getAllNodes() {
	GraphNode[] big_list = new GraphNode[nodeBasket.size()];
	
	Enumeration node_iterate = nodeBasket.elements();
		 
	for(int i = 0; i < nodeBasket.size(); i++) {
	    big_list[i] = (GraphNode)node_iterate.nextElement();
	}
	
	return big_list;
    }
    
    /** 
     * Return the depth from source to destination if calcMaxDepth() 
     * has already been called, -1 otherwise
     *
     * @return The maxdepth of the graph.
     */
    public int getMaxDepth(String sourceName) {
	SourceObj srcObj = (SourceObj)sourceHashGet(sourceName);
	int depth;
	if(srcObj != null) {
	    
	    return(srcObj.maxDepth);  
	}
	return(-1);
    }

    /**
     * Query of a given source is oriented foward or backward in the graph
     *
     * @return The direction of the source.
     */
    public boolean getOrientation(String sourceName) {
	SourceObj srcObj = (SourceObj)sourceHashGet(sourceName);
	int depth;
	if(srcObj != null) {
	    return(srcObj.orientation);  
	}
	return(true);
    }


    /** 
     * Read in a Vector of GraphNodes that will serve as sourcee
     *    for the calcMaxDepth algorithm
     *
     * @param sources A Vector of all the sources to use when calculating
     *     the max depth.
     */
    public int calcMaxDepth(Vector sources, String sourceName) { 

	if(sources != null && sources.size() > 0) {
	    GraphNode tmpNode = (GraphNode)sources.elementAt(0);
	    return tmpNode.calcMaxDepth(sources, sourceName);
	}
	return -1;
    }

    /** 
     * Return all the sources that have been stored in the graph
     *
     * @return An array of all the source names.
     */
    public String[] getAllSources() {
	String[] allSources = new String[sourceHashSize()];
	System.out.println("Sourcehash Size: " + sourceHashSize());


	for(int i=0; i<sourceHashSize(); i++) {
	    allSources[i] = (String)fakeHashKey.elementAt(i);
	}
	return allSources;
    }

    /** 
     * Get a GraphNode by its name 
     *
     * @param name The name of the GraphNode to look up.
     * @return The GraphNode object corresponding to the name, otherwise null.
     */
    public GraphNode getNode(String name) {
	return((GraphNode)nodeBasket.get(name));
    }

    public Object sourceHashGet(String getName) {
	for(int i = 0; i < fakeHashKey.size(); i++) {
	    if(((String)fakeHashKey.elementAt(i)).equals(getName)) {
		return fakeHashValue.elementAt(i);
	    }
	}
	return null;
    }

    public Object sourceHashPut(String putName, Object putThing) {
	for(int i = 0; i < fakeHashKey.size(); i++) {
	    if(((String)fakeHashKey.elementAt(i)).equals(putName)) {
		Object replacedObj = fakeHashValue.elementAt(i);
		fakeHashValue.removeElementAt(i);
		fakeHashValue.insertElementAt(putThing, i);
		return replacedObj;
	    }
	}
	fakeHashKey.addElement(putName);
	fakeHashValue.addElement(putThing);

	return null;
    }

    public int sourceHashSize() {
	return fakeHashKey.size();
    }

    //keep a finger on important nodes
    private int max_depth = -1;
    private Vector fakeHashKey = new Vector();
    private Vector fakeHashValue = new Vector();

    //keep a pointer to every node (traversing graph could be ugly)
    private Hashtable nodeBasket;

    private class SourceObj {
	public int maxDepth = -1;
	public boolean orientation = true;   // true = bottom to top.
	
    }
}

