//==========================================================================
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// Permission to use, copy, modify and distribute any part of this Makefile
// for educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above 
// copyright notice, this paragraph and the following paragraphs appear 
// in all copies.
//
// Those desiring to incorporate this into commercial products or use for
// commercial purposes should contact the
//
//      Technology Transfer Office, University of California, San Diego,
//      9500 Gilman Drive, La Jolla, CA 92093-0910
//      Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS PLOTPATH, EVEN 
// IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
//
// THE PLOTPATH, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF 
// CALIFORNIA PLOTPATH NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF 
// ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
// PURPOSE, OR THAT THE USE OF THE SKITTERTRACES WILL NOT INFRINGE ANY 
// PATENT, TRADEMARK OR OTHER RIGHTS.
//
// PlotPaths is developed by John Gallagher Bradley L. Huffaker at the
// University of California, San Diego under the Cooperative Association
// for Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by NSF "Internet Atlas" grant ANI-I-9996248, NSF "CAIDA Core"
// NCR-9711092, and by CAIDA memebers.
//==========================================================================
package org.caida.otter;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Stack;

//
// Rename variables in CalcMaxDepth (as directed by comments)
// Look at calcMaxDepth w/ multiple sources, I don't think the
//    algorithm will appropriately represent those multiple sources
//

/** 
 * A node in a directed graph.
 */
public class GraphNode {

    /** 
     * Make a new node w/ no children 
     */
    public GraphNode(String name) {
	this.name = name;
    }
    
    /** 
     * Make a new node w/ one child 
     */
    public GraphNode(String source, String name, GraphNode new_child) {
	this.name = name;
	addChild(source, new_child);
    }


    /** 
     * Add one parent to a node 
     */ 
    public void setParent(String source, GraphNode parent) {
	if(source == null) {
	    return;
	}
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    srcObj.parent = parent;
	}
    }

    /** 
     * Add one child to a node 
     */
    public void addChild(String source, GraphNode new_child) {
	if(source == null || new_child == null) {
	    return;
	}
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj == null) {
	    srcObj = new SourceObj(this);
	    sourceHash.put(source, srcObj);
	}
	srcObj.children.addElement(new_child);
    }

    /** 
     * Add several children to a node 
     */
    public void addChild(String source, Vector new_children) {
	if(source == null || new_children == null) {
	    return;
	}
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj == null) {
	    srcObj = new SourceObj(this);
	    sourceHash.put(source, srcObj);
	}
	try {
	    for(int i = new_children.size() - 1; i >= 0; i--) {
		srcObj.children.addElement(
			    ((GraphNode)new_children.elementAt(i)));
	    } 
	} catch(ClassCastException e) {
	    throw new IllegalArgumentException("Must supply Vector "+
		    "of GraphNode as neighbor list.");
	}
    }

    /**
     * Add several children to a node 
     */
    public void addChild(String source, GraphNode[] new_children) {
	if(new_children == null || source == null) {
	    return;
	}
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj == null) {
	    srcObj = new SourceObj(this);
	    sourceHash.put(source, srcObj);
	}
	for(int i = new_children.length - 1; i >= 0; i--) {
	    srcObj.children.addElement(new_children[i]);
	}
    }

    /** 
     * Override equals to test name equivalence 
     */
    public boolean equals(Object node) {
	if(node instanceof GraphNode) {
	    if(this.name.equals(((GraphNode)node).name)) {
		return true;
	    } 
	    return false;
	} else {
	    if(this.equals(node)) {
		return true;
	    } 
	    return false;
	} 
    }
    
    /**
     * Get a node parent for a given source if one exists 
     */
    public GraphNode getParent(String source) {
	if(source == null) {
	    return null;
	}
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    return srcObj.parent;
	}
	return null;
    }

    /** 
     * Get a node's children for a given source if they exist 
     */
    public Vector getChildren(String source) {
	if(source == null) {
	    return null;
	}
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    return srcObj.children;
	}
	return null;
    }
    
    /** 
     * Get a node's name 
     */
    public String getName() {
	return name;
    }
    

    /** 
     * Set a node's depth in the graph 
     */
    public void setDepth(String source, int depth) {
	if(source == null) {
	    return;
	}
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj == null) {
	    srcObj = new SourceObj(this);
	    sourceHash.put(source, srcObj);
	}
	srcObj.graph_depth = depth;
    }

    /** 
     * Get a node's depth in the graph 
     */
    public int getDepth(String source) {
	if(source == null) {
	    return(-1);
	}
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    return srcObj.graph_depth;
	}
	return(-1);
    }

    /** 
     * Get all sources that this node is used in 
     */
    public String[] getAllSources() {
	String[] allSources = new String[sourceHash.size()];

	Enumeration sourceIterate = sourceHash.keys();

	for(int i=0; i<sourceHash.size(); i++) {
	    allSources[i] = (String)sourceIterate.nextElement();
	}
	return allSources;
    }
    
    /** 
     * Recursively check the parents to see if any of 
     *    them correspond to the node given 
     */
    public boolean isChildOf(String source, GraphNode node) {
	if(source == null || node == null) {
	    return false;
	}

	SourceObj tmpSrc = (SourceObj)sourceHash.get(source);
	GraphNode current;

	if(tmpSrc != null) {
	    current = tmpSrc.parent;		
	} else {
	    current = null;
	}
		
        while(current != null && current != node) { 
	    tmpSrc = (SourceObj)current.sourceHash.get(source);
	    if(tmpSrc != null) {
		current = tmpSrc.parent;
	    } else {
		current = null;
	    }
	}
    
                if(current == null) {
                        return false;
                }
                return true;
    }


    /** 
     * Calculate depth in graph for each node using the current
     *   node as the only source.
     */
    public int calcMaxDepth() {

	Vector sourceVect = new Vector();
	sourceVect.addElement(this);

	return calcMaxDepth(sourceVect, this.name);
    }   



    /**
     * CalcMaxDepth of the graph from multiple nodes under a single
     *   source category.
     */
    public int calcMaxDepth(Vector sourceList, String sourceName) {

	String source = sourceName;

	GraphNode current = this;
	if(current == null) {
	    return(-1);
	}

	GraphNode child;
	Vector childList;
	Stack nodeStack = new Stack();
	int max_depth = 0;

	// initialize source
	for(int i = 0; i < sourceList.size(); i++) {
	    ((GraphNode)sourceList.elementAt(i)).setDepth(source, 1);
	    nodeStack.push((GraphNode)sourceList.elementAt(i));
	}

	while(!nodeStack.empty()) {
	    current = (GraphNode)nodeStack.pop();

	    if(current != null) {
		if(current.getDepth(source) > max_depth) {
		    max_depth = current.getDepth(source);
		}
		childList = current.getChildren(source);
		for(int i = childList.size() - 1; i >= 0; i--) {

		    child = (GraphNode)childList.elementAt(i);

		    if(child != null) {

			// mark nodes that havn't been visited
			if(child.getDepth(source) == -1) {
			    child.setDepth(source, current.getDepth(source)+1);
			    child.setParent(source, current);
			    nodeStack.push(child);
			}

			// have to look harder if the node has been visited
			else if(child.getDepth(source) <= 
					current.getDepth(source) &&
					!current.isChildOf(source, child)) {

			    child.setDepth(source, current.getDepth(source)+1);
			    child.setParent(source, current);
			    nodeStack.push(child);
			}
		    } else {
			System.out.println("Invalid child at node: "+
				((GraphNode)childList.elementAt(i)).getName());
		    }
		}
	    } else {
		System.out.println("uninitialized children");
	    }
	}
	return max_depth;
    }   

    /** 
     * Merge the children for a given source.  
     */
    public void mergeChildren(String source, GraphNode ref_node) {

	if(source == null || ref_node == null) {
	    return;
	}

	SourceObj updateSrcObj = (SourceObj)sourceHash.get(source);
	SourceObj refSrcObj    = (SourceObj)ref_node.sourceHash.get(source);

	if(refSrcObj == null) {
	    return;
	}

	if(updateSrcObj == null) {
	    updateSrcObj = new SourceObj(this);
	    sourceHash.put(source, updateSrcObj);
	}
	for(int i=refSrcObj.children.size()-1; i >= 0; i--) {

	    if(!updateSrcObj.contains(
			((GraphNode)refSrcObj.children.elementAt(i)).
			getName())) {

		//add the child
		updateSrcObj.children.addElement(
			(GraphNode)refSrcObj.children.elementAt(i));
	    }
	    // sum the weights
	    updateSrcObj.weight += refSrcObj.weight;
	}

    }

    /** 
     * Make sure all the children are valid children for the graph 
     *     meaning that the children are contained in the graph.
     */
    public void validateChildren(String source, DirectedGraph graph) {
	if(source == null || graph == null) {
	    return;
	}
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {

	    for(int i=srcObj.children.size()-1; i >= 0; i--) {

		GraphNode childNode = 
			graph.getNode(((GraphNode)srcObj.children.elementAt(i)).
			getName());

		if(childNode != null && childNode != 
			((GraphNode)srcObj.children.elementAt(i))) {

		    // update child pointer so it points to the correct
		    // memory
		    srcObj.children.removeElementAt(i);
		    srcObj.children.insertElementAt(childNode,i);
		}
	    }
	}
	return;
    }
    
    private String name;
    private Hashtable sourceHash 	= new Hashtable();

    //private boolean hasLatLong 	= false;
    private float latitude 		= 0;
    private float longitude 	= 0;

    //private boolean hasCity 	= false;
    private String city;

    //private boolean hasASNum	= false;
    private int ASNumber 		= -199;

    private class SourceObj {
	public Vector children = new Vector();
	public GraphNode parent = null;
	public float weight = 0;
	public int graph_depth = -1;
	public GraphNode sourceNode = null;
	
	public SourceObj(GraphNode source) {
	    sourceNode = source;
	}
	
	public Vector getChildren() {
	    return children;
	}
	

	/** 
	 * check if any of the children correspond to the name given 
	 */
	public boolean contains(String name) {
	    if(name == null) {
		return false;
	    }
	    for(int i=children.size()-1;i >= 0; i--) {
		if(((GraphNode)children.elementAt(i)).getName().equals(name)) {
		    return true;
		}
	    }
	    return false;
	}

	/** 
	 * Recursively check the parents to see if any of them correspond 
	 * to the node given 
	 */
	public boolean isChildOf(String source, GraphNode node) {
 
	    if(source == null || node == null) {
		return false;
	    }
 
	    GraphNode current = parent;
	    SourceObj tmpSrc;
 
	    while(current != null && current != node) {

		tmpSrc = (SourceObj)current.sourceHash.get(source);

		if(tmpSrc != null) {
		    current = tmpSrc.parent;
		} else {
		    current = null;
		}
	    }
 
	    if(current == null) {
		return false;
	    }
	    return true;
	}                             
    }

//////////////////////////////////
/////// Unused methods ///////////
//////////////////////////////////

    /** 
     * Make a new node w/ one child and weight
     */
    public GraphNode(String source, String name, GraphNode new_child, 
				    float weight) {
	this.name = name;
	addChild(source, new_child);
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    srcObj.weight = weight;
	}
    }

    /** 
     * Make a new node w/ one child and weight
     */
    public GraphNode(String source, String name, float weight) {
	this.name = name;
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj == null) {
	    srcObj = new SourceObj(this);
	    sourceHash.put(source, srcObj);
	}
	srcObj.weight = weight;
    }
    
    /** 
     * Make a new node w/ several children 
     */
    public GraphNode(String source, String name, GraphNode[] new_childList) {
	this.name = name;
	addChild(source, new_childList);
    }

    /** 
     * Make a new node w/ several children and weight
     */
    public GraphNode(String source, String name, GraphNode[] new_childList, 
				    float weight) {
	this.name = name;
	addChild(source, new_childList);
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    srcObj.weight = weight;
	}
    }

    /** 
     * Make a new node w/ several children 
     */
    public GraphNode(String source, String name, Vector new_childList) {
	this.name = name;
	addChild(source, new_childList);
    }
    
    /** 
     * Make a new node w/ several children and weight
     */
    public GraphNode(String source, String name, Vector new_childList, 
				    float weight) {
	this.name = name;
	addChild(source, new_childList);
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    srcObj.weight = weight;
	}
    }
    
    /** 
     * Make a new node w/ no children, w/ Lat Long 
     */
    public GraphNode(String name, float latitude, float longitude) {
	this.name = name;
	this.latitude = latitude;
	this.longitude = longitude;
    }
    
    
    /** 
     * Make a new node w/ one child, w/ Lat Long 
     */
    public GraphNode(String source, String name, GraphNode new_child, 
		float latitude, float longitude) {
	this.name = name;
	this.latitude = latitude;
	this.longitude = longitude;
	addChild(source, new_child);
    }
    
    /** 
     * Make a new node w/ one child, w/ Lat Long and weight
     */
    public GraphNode(String source, String name, GraphNode new_child, 
		float latitude, float longitude, float weight) {
	this.name = name;
	this.latitude = latitude;
	this.longitude = longitude;
	addChild(source, new_child);
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    srcObj.weight = weight;
	}
    }
    
    /** 
     * Make a new node w/ several children, w/ Lat Long 
     */
    public GraphNode(String source, String name, Vector new_childList, 
		float latitude, float longitude) {
	this.name = name;
	this.latitude = latitude;
	this.longitude = longitude;
	addChild(source, new_childList);
    }

    /** 
     * Make a new node w/ several children, w/ Lat Long and weight
     */
    public GraphNode(String source, String name, Vector new_childList, 
		float latitude, float longitude, float weight) {
	this.name = name;
	this.latitude = latitude;
	this.longitude = longitude;
	addChild(source, new_childList);
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    srcObj.weight = weight;
	}
    }

    /** 
     * Make a new node w/ several children, w/ Lat Long 
     */
    public GraphNode(String source, String name, GraphNode[] new_childList, 
		float latitude, float longitude) {
	this.name = name;
	this.latitude = latitude;
	this.longitude = longitude;
	addChild(source, new_childList);
    }

    /** 
     * Make a new node w/ several children, w/ Lat Long and weight
     */
    public GraphNode(String source, String name, GraphNode[] new_childList, 
		float latitude, float longitude, float weight) {
	this.name = name;
	this.latitude = latitude;
	this.longitude = longitude;
	addChild(source, new_childList);
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    srcObj.weight = weight;
	}
    }

    /** 
     * Get a node's latitude 
     */
    public float getLat() {
	return latitude;
    }

    /** 
     * Set a node's latitude 
     */
    public void setLat(float lat) {
	latitude = lat;
    }

    /** 
     * Set a node's AS number 
     */
    public void setASNumber(int as_num) {
	ASNumber = as_num;
    }
    
    /** 
     * Get a node's AS Number 
     */
    public int getASNumber() {
	return ASNumber;
    }

    /** 
     * Set a node's city 
     */
    public void setCity(String city) {
	this.city = city;
    }
    
    /** 
     * Get a node's city 
     */
    public String getCity() {
	return city;
    }

    /**
     * Get a node's longitude 
     */
    public float getLong() {
	return longitude;
    }

    /** 
     * Set a nodes longitude 
     */
    public void setLong(float lon) {
	longitude = lon;
    }

    /**
     * Add to a node's weight for a given source 
     */
    public void incrementWeight(String source, float num) {
	if(source == null) {
	    return;
	}
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    srcObj.weight += num;
	}
    }

    /** 
     * Decrement a node's weight for a given source 
     */
    public void decrementWeight(String source, float num) {
	if(source == null) {
	    return;
	}
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    srcObj.weight -= num;
	}
    }

    /** 
     * Get a node's weight given a source 
     */
    public float getWeight(String source) {
	if(source == null) {
	    return(0);
	}
	SourceObj srcObj = (SourceObj)sourceHash.get(source);
	if(srcObj != null) {
	    return srcObj.weight;
	}
	return(0);
    }
} 
