//==========================================================================
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// Permission to use, copy, modify and distribute any part of this Makefile
// for educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above 
// copyright notice, this paragraph and the following paragraphs appear 
// in all copies.
//
// Those desiring to incorporate this into commercial products or use for
// commercial purposes should contact the
//
//      Technology Transfer Office, University of California, San Diego,
//      9500 Gilman Drive, La Jolla, CA 92093-0910
//      Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS PLOTPATH, EVEN 
// IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
//
// THE PLOTPATH, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF 
// CALIFORNIA PLOTPATH NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF 
// ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
// PURPOSE, OR THAT THE USE OF THE SKITTERTRACES WILL NOT INFRINGE ANY 
// PATENT, TRADEMARK OR OTHER RIGHTS.
//
// PlotPaths is developed by John Gallagher Bradley L. Huffaker at the
// University of California, San Diego under the Cooperative Association
// for Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by NSF "Internet Atlas" grant ANI-I-9996248, NSF "CAIDA Core"
// NCR-9711092, and by CAIDA memebers.
//==========================================================================
package org.caida.otter;

import java.io.*;
import java.util.Enumeration;

/**
 * IOotter - a class that handles IO in Otter format
 */
public class IOotter implements PlotPathsIO {

    /** 
     * Output data in Otter format .
     * 
     * @param data DisplayObjects object to extract node/label/link info from
     * @param outfile Name of file to write to disk.  The file will be 
     *           overwritten if it already exists.
     * 
     */
     public void writeOutput(DisplayObjects data, String outfile, GetOpts CmdOpts) {

	PrintWriter out = null;

	try {  
	    out = new PrintWriter(new BufferedWriter(new FileWriter(outfile)));
	} catch(IOException e) {
	    System.out.println("[Error]: printing output to file: " + outfile);
	    System.out.println(e.toString());
	    out = null;
	}
	if(out == null) {
	    System.out.println(
		    "[Error]: No output stream open - unable to write output");
	    return;
	}

	// print Otter header stuff
	int nodeTotalCount  = 0;
	int labelTotalCount = 0;
	int lineTotalCount = 0;

	Enumeration nodeEnum;
	Enumeration lineEnum;
	Enumeration labelEnum;

	// get line count
	lineTotalCount = data.getLineCount();
	
	// get label count
	labelTotalCount = data.getLabelCount();

	// get (visible) node count
	nodeEnum = data.getNodeEnumeration();
	while(nodeEnum.hasMoreElements()) {
	    if(!((Node)nodeEnum.nextElement()).isHidden()) {
		nodeTotalCount++;
	    }
	}

	if(CmdOpts.get_DISPLAY_LABELS()) {
	    out.println("t "+(nodeTotalCount+labelTotalCount));
	    out.println("T "+lineTotalCount);
	} else {
	   out.println("t "+nodeTotalCount);
	    out.println("T "+lineTotalCount);
	}


	// print source (group) info
	// build array of arrays representing all catergories and attributes
	Object[] categoryList = data.enumerateCategories();
	Object[][] categoryAttributeList = new Object[categoryList.length][];
	for(int i=0; i < categoryList.length; i++) {
	    categoryAttributeList[i] = 
		    data.enumerateAttributes(categoryList[i]);
	}

	if(categoryList.length > 0) {
	    for(int i=0; i < categoryList.length; i++) {
		if(categoryAttributeList[i] == null) {
		    System.out.println("Category "+(String)categoryList[i]+
			    " has no attributes");
		} else {
		    out.println("g "+i+" s "+categoryAttributeList[i].length+
			    " "+(String)categoryList[i]);
		    out.print("f "+i+" ");
		}
		for(int j=0; j < categoryAttributeList[i].length; j++) {
		    out.print((String)categoryAttributeList[i][j]);
		    if(j < categoryAttributeList[i].length-1) {
			out.print("'");
		    }
		}
		out.println();
	    }
	}

    
	// print node data
	nodeEnum = data.getNodeEnumeration();
	while(nodeEnum.hasMoreElements()) {

	    Node tmpNode = (Node)nodeEnum.nextElement();
	    if(!tmpNode.isHidden()) {
		out.print("n "+tmpNode.index+" "+tmpNode.x+" "+tmpNode.y+" "+
						tmpNode.name);
		if(tmpNode.isRoot()) {
		    out.print(" r");
		}
		out.println();
	    } 
	}

	// specify node size (for following label nodes)
	//out.println("s 1");


	if(CmdOpts.get_DISPLAY_LABELS()) {
	    // print label data
	    labelEnum = data.getLabelEnumeration();
	    while(labelEnum.hasMoreElements()) {
		Label tmpLabel = (Label)labelEnum.nextElement();
		out.println("n "+(tmpLabel.index+nodeTotalCount)+
			" "+tmpLabel.x+" "+tmpLabel.y+" "+tmpLabel.getName()+
			" r");
	    }

	    System.out.println("\nNumber of Statistics: "+
				categoryAttributeList.length);
	    System.out.println( "Lines: \t\t"+lineTotalCount+
				"\nNodes: \t\t"+nodeTotalCount+
				"\nLabels: \t"+labelTotalCount+
				"\nHidden Nodes: \t"+
				    (data.getNodeCount()-nodeTotalCount));
	}

	// print line data
	String pathWeight;
	lineEnum = data.getLineEnumeration();
	for(int i = 0; lineEnum.hasMoreElements(); i++) {
	    Line tmpLine = (Line)lineEnum.nextElement();
	    int idx1, idx2;
	    idx1 = tmpLine.getCoord1().index;
	    idx2 = tmpLine.getCoord2().index;
	    out.println("l "+i+" "+idx1+" "+idx2);

	    if(categoryList.length > 0) {
		for(int j=0; j < categoryList.length; j++) {
		    out.print("V "+i+" "+j+" ");
		    for(int k=0; k < categoryAttributeList[j].length; k++) {
			pathWeight = (String)tmpLine.getAttribute(
					(String)categoryList[j],
					(String)categoryAttributeList[j][k]);

			if(pathWeight != null) {
			    out.print(pathWeight);
			} else {
			    out.print("not selected");
			}
			if(k < categoryAttributeList[j].length-1) {
			    out.print("'");
			}
		    }
		    out.println();
		}
	    }
	}
	out.close();
     }

     /**
      * placeholder b/c we don't load otter format data.  (er.. yet)
      */
    public PathSet readInput(String inputFile, String dataFile) {
	throw new UnsupportedOperationException("Method not implemented");
    }
}
