//==========================================================================
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// Permission to use, copy, modify and distribute any part of this Makefile
// for educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above 
// copyright notice, this paragraph and the following paragraphs appear 
// in all copies.
//
// Those desiring to incorporate this into commercial products or use for
// commercial purposes should contact the
//
//      Technology Transfer Office, University of California, San Diego,
//      9500 Gilman Drive, La Jolla, CA 92093-0910
//      Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS PLOTPATH, EVEN 
// IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
//
// THE PLOTPATH, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF 
// CALIFORNIA PLOTPATH NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF 
// ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
// PURPOSE, OR THAT THE USE OF THE SKITTERTRACES WILL NOT INFRINGE ANY 
// PATENT, TRADEMARK OR OTHER RIGHTS.
//
// PlotPaths is developed by John Gallagher Bradley L. Huffaker at the
// University of California, San Diego under the Cooperative Association
// for Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by NSF "Internet Atlas" grant ANI-I-9996248, NSF "CAIDA Core"
// NCR-9711092, and by CAIDA memebers.
//==========================================================================
package org.caida.otter;

import java.io.*;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * IOtext - handle IO relating to text format data.
 */
public class IOtext implements PlotPathsIO {

    /** Build a graph given an input filename
     *
     * @param filename The file to read data from.
     * @return PathSet object containing the loaded paths.
     */
    public PathSet readInput(String inputFile, String dataFile) {

	boolean pathIsComplete = true;
    
	PathSet graph = new PathSet(PlotPaths.UNKNOWN_NODE_NAME);
	int noDataCount = 0;

	try {
	    String inputLine, start, end, token;
	    StringTokenizer toke;
	    Vector pathNames;
	    Vector pathNodes;
	    boolean forwardPath = true;  
	    float pathWeight;
	    BufferedReader input = new BufferedReader(new FileReader(inputFile));
	    inputLine = input.readLine();
	    while(inputLine != null) {
		

		// parse out the graph info
		
		// get start <-> destination
		if(inputLine.length() > 0 && inputLine.charAt(0) != ' ') {
		    //System.out.println(inputLine);
		    toke = new StringTokenizer(inputLine);
		    start = (String)toke.nextElement();
		    toke.nextElement();
		    end = (String)toke.nextElement();
		    end = end.substring(0,end.length() - 1);
		    
		    inputLine = input.readLine();
		    if(inputLine == null) {
			break;
		    }
		    toke = new StringTokenizer(inputLine);
		    token = (String)toke.nextElement();
		
		    while(inputLine != null && !token.equals("Total")) {
			if(inputLine != null && token.equals("Path")) {
			    pathIsComplete = true;

			    // find the path weight
			    // eigth token (already saw one), 
			    for(int k=0;k<7;k++) {
				token = (String)toke.nextElement();
			    }
			    token = token.substring(0,token.length()-1);
			    pathWeight = Float.valueOf(token).floatValue();
			    inputLine = input.readLine();
			    if(inputLine == null) {
				break;
			    }
			    //System.out.println("path: "+inputLine);
			    toke = new StringTokenizer(inputLine);
			    token = (String)toke.nextElement();

			    pathNames = new Vector();
			    pathNodes = new Vector();



			    while(inputLine != null && 
					!token.equals("Total") &&
					!token.equals("Path")) {
				if(forwardPath && 
					!inputLine.trim().equals("no data")) {
				    pathNames.addElement(inputLine.trim());
				} else if(!forwardPath && 
					!inputLine.trim().equals("no data")) {
				    pathNames.addElement(inputLine.trim());
				} else if(inputLine.trim().equals("no data")) {
				    noDataCount++;
				    pathNames.addElement("*");
				    pathIsComplete = false;
				}
				inputLine = input.readLine();
				toke = new StringTokenizer(inputLine);
				token = (String)toke.nextElement();
			    }
			    if(pathNames.size() > 0) {

				// Add condition that path is significant
				if(pathWeight > PlotPaths.MIN_PERCENT_USAGE &&
					(pathIsComplete || 
					!PlotPaths.IGNORE_INCOMPLETE_PATHS)) {

				    graph.addPath(  pathNames, 
						    pathWeight, 
						    forwardPath);
				}
			    }
			} else {
			    inputLine = input.readLine();
			    if(inputLine == null) {
				break;
			    }
			    toke = new StringTokenizer(inputLine);
			    token = (String)toke.nextElement();
			}
		    }

		    /////////////////////////////
		    // PATH FROM TOP or BOTTOM //
		    /////////////////////////////

		    if(PlotPaths.USE_ALL_FORWARD_PATHS) {

			// every path is bottom to top
			forwardPath = true;

		    } else {

			// alternate orienation, every other one
			if(forwardPath) {
			    forwardPath = false;
			} else {
			    forwardPath = true;
			}
		    }

		} else {
		    inputLine = input.readLine();
		}
	    }
	    input.close();
	} catch(FileNotFoundException e) {
	    System.out.println("Unable to open file: "+inputFile);
	    return null;
	} catch(IOException e) {
	    System.out.print(e.toString()); 
	    return null;
	}
     
	System.out.println(noDataCount+" nodes are non-responding.");

	// add on the data from the data file (AS info)
	getNodeInfo(graph, dataFile);

	return graph;
    }

    /** 
     * Add lat and long data (and city name and anything else)
     *  to set of graph data points.
     * 
     * @param graph PathSet object that contains nodes that need to have
     *   lat long data found.
     */
    public void getNodeInfo(PathSet graph, String nodeDataFile) {
	
	String[] nodeList = graph.getAllNodes();
	boolean hasAS;
	    
	try {
	    String inputLine, token;
	    StringTokenizer toke;
	    int i; // loop var
	    BufferedReader input = 
		    new BufferedReader(new FileReader(nodeDataFile));

	    inputLine = input.readLine();
	    while(inputLine != null) {
		toke = new StringTokenizer(inputLine);
		token = (String)toke.nextElement();
		for(i=nodeList.length-1; i >= 0; i--) {
		    if(token.equals(nodeList[i])) {
			token = (String)toke.nextElement();

			if(!token.equals("not_found")) {

			    //grab AS num
			    graph.setASNumber(nodeList[i], 
					Integer.valueOf(token).intValue());
			}

			if(toke.hasMoreTokens()) {

			    // we have the AS name
			    token = (String)toke.nextElement();
			    graph.setASName(nodeList[i], token);
			}
		    }
		}
		inputLine = input.readLine();
	    }
	    input.close();

	} catch(IOException e) {
	    System.out.println("[Error]: Could not open data file " +
			"\""+nodeDataFile+"\"");
	    //System.out.println(e.toString());
	}
    }

    /**
     * placeholder b/c we don't write out plain text output. (er.. yet)
     */
    public void writeOutput(DisplayObjects data, String filename, GetOpts CmdOpts) {
	throw new UnsupportedOperationException("Method not implemented");
    }
}
