//==========================================================================
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// Permission to use, copy, modify and distribute any part of this Makefile
// for educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above 
// copyright notice, this paragraph and the following paragraphs appear 
// in all copies.
//
// Those desiring to incorporate this into commercial products or use for
// commercial purposes should contact the
//
//      Technology Transfer Office, University of California, San Diego,
//      9500 Gilman Drive, La Jolla, CA 92093-0910
//      Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS PLOTPATH, EVEN 
// IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
//
// THE PLOTPATH, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF 
// CALIFORNIA PLOTPATH NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF 
// ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
// PURPOSE, OR THAT THE USE OF THE SKITTERTRACES WILL NOT INFRINGE ANY 
// PATENT, TRADEMARK OR OTHER RIGHTS.
//
// PlotPaths is developed by John Gallagher Bradley L. Huffaker at the
// University of California, San Diego under the Cooperative Association
// for Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by NSF "Internet Atlas" grant ANI-I-9996248, NSF "CAIDA Core"
// NCR-9711092, and by CAIDA memebers.
//==========================================================================
package org.caida.otter;

import java.io.*;
import java.util.*;
import java.lang.Number;
import org.caida.libsea.*;

/**
 *  Class that bridges the LibSea graph format to the data structures
 *  that exist in PlotPaths.
 */
public class LibSeaToPlotPaths
{
    private Vector rejectedPaths                  = new Vector();
    private Hashtable linkHash                    = new Hashtable();
    private Hashtable srcDstHashtable             = new Hashtable();
    private GraphInfo graphInfo                   = new GraphInfo();

    /**
     *  Extracts the path infomation from the LibSea graph file
     *  and adds the proper information into the PathSet data structures
     *  that originally were written for PlotPaths.
     *  POSSIBLE MODIFICATIONS:
     *  - UNKNOWN_NODE_NAME may need to be replaced or done away with.
     *  It is a relic of the original PlotPaths.
     *  @param LibSea_Graph LibSea graph file
     *  @param UNKNOWN_NODE_NAME The string representation of an 
     *  unknown node.
     *  @return A PathSet graph file with the paths given in unique 
     *  LibSea graph file ID values and AS numbers associated with
     *  each ID.
     */
    public PathSet loadPaths(Graph LibSea_Graph, GetOpts CmdOpts)
    {
	PathSet graph = new PathSet(CmdOpts.get_UNKNOWN_NODE_NAME());

	//Extract the Frequency attributes
	AttributeDefinitionIterator FrqMainIterator = LibSea_Graph.getAttributeDefinition("frequency");
	AttributesByAttributeIterator FrqIterator = FrqMainIterator.getPathAttributes();

	Hashtable srcDstTotalFrequency = new Hashtable();
	Vector FrqPaths = new Vector();

	while(!FrqIterator.atEnd())
	    {
		ValueIterator FrqValue = FrqIterator.getAttributeValues();
		Integer tempInt = new Integer(FrqValue.getIntegerValue());
		FrqPaths.addElement(tempInt);
		FrqIterator.advance();
	    }

	// add the paths to the PathSet object
	Vector Sources = CmdOpts.get_SOURCES();	
	Vector allPaths = new Vector();
	Integer source = null, destination = null, intValue = null;
	int rejectCount = 0;
	boolean forwardPath = true, reject = false;

	String srcDst = "";
	
	//Extract the path information from the LibSea graph file.
	for (int i = 0; i < LibSea_Graph.getNumPaths(); i++)
	    {
		PathIterator pathIterator = LibSea_Graph.getPath(i);
		LinkIterator linkIterator = pathIterator.getLinks();
		intValue = new Integer(linkIterator.getSource());
		Vector path = new Vector();
		path.addElement(intValue.toString());
		while(!linkIterator.atEnd())
		    {
			intValue = new Integer(linkIterator.getDestination());
			path.addElement(intValue.toString());

			// add all the links and their ID's to the linkHash
			// need this to write out the statistic attributes
			source = new Integer(linkIterator.getSource());
			destination = intValue;
			srcDst = source.toString() + "," + destination.toString();
			linkHash.put(srcDst, new Integer(linkIterator.getObjectID()));
			linkIterator.advance();
		    }

		allPaths.addElement(path);

		// Use the first source that is encountered if the user did not specify
		// any sources.
		if(Sources.isEmpty() && i == 0)
		    {
			Sources.addElement(new Integer(path.elementAt(0).toString()));
		    }

		source = new Integer(path.elementAt(0).toString());
		destination = new Integer(path.lastElement().toString());

		srcDst = source.toString() + "," + destination.toString();

		// Keep track of the path ID's
		if(srcDstHashtable.containsKey(srcDst))
		    {
			HashSet pathIDHashSet = (HashSet)srcDstHashtable.get(srcDst);
			pathIDHashSet.add(new Integer(pathIterator.getObjectID()));
			srcDstHashtable.put(srcDst, pathIDHashSet);
		    }
		else
		    {
			HashSet pathIDHashSet = new HashSet();
			pathIDHashSet.add(new Integer(pathIterator.getObjectID()));
			srcDstHashtable.put(srcDst, pathIDHashSet);
		    }
		
		// Keep track of the path frequencies
		if(srcDstTotalFrequency.containsKey(srcDst))
		    {
			Integer totalFrequency = (Integer)srcDstTotalFrequency.get(srcDst);
			Integer currentPathFrequency = (Integer)FrqPaths.get(i);
			int tempint = totalFrequency.intValue() + currentPathFrequency.intValue();
			srcDstTotalFrequency.put(srcDst, new Integer(tempint));
		    }
		else
		    {
			srcDstTotalFrequency.put(srcDst,(Integer)FrqPaths.get(i));
		    }
		
	    }
	
	for(int i = 0; i < allPaths.size(); i++)
	    {
		Vector path = new Vector();
		path = (Vector)allPaths.get(i);
		source = new Integer(path.elementAt(0).toString());
		destination = new Integer(path.lastElement().toString());
		srcDst = source.toString() + "," + destination.toString();

		if(Sources.contains(source) && Sources.contains(destination))
		    {
			reject = true;
			rejectCount++;
		    }
		else if(Sources.contains(source))
		    {
			forwardPath = true;
			reject = false;
		    }
		else if(Sources.contains(destination))
		    {
			forwardPath = false;
			reject = false;
		    }
		else
		    {
			reject = true;
			rejectCount++;
		    }
		
		if(!reject)
		    {
			float numerator = ((Integer)(FrqPaths.get(i))).floatValue();
			float denomonator = ((Integer)(srcDstTotalFrequency.get(srcDst))).floatValue();
			float pathWeight = (numerator / denomonator) * 100;
			graph.addPath(path, pathWeight, forwardPath);
		    }
		else
		    {
			rejectedPaths.addElement(new Integer(i));
		    }
	    }
	
	//Check to see if all the paths were rejected
	if(rejectCount == LibSea_Graph.getNumPaths())
	    {
		System.out.println("All paths have been rejected. Make sure that you have ");
		System.out.println("chosen a proper source.");
		System.exit(1);
	    }
	
	System.out.println(rejectCount + ": paths were rejected.");

	return graph;
    }
    

    /**
     *  Sets the column attribute that will be used to reference the columns.
     *  Type Summary:
     *  Boolean = 0;
     *  Integer = 1;
     *  Float = 2;
     *  Double = 3;
     *  String = 4;
     *  Float3 = 5;
     *  Double3 = 6;
     *  Enumeration = 7;
     *  @param LibSea_Graph The LibSea graph object.
     *  @param graph The PathSet graph object.
     *  @param COLUMN_ATTRIBUTE The attribute that the columns will be seperated into.
     *  @return The PathSet graph object with the nodes set to the COLUMN_ATTRIBUTE.
     */
    public PathSet setColumnAttribute(Graph LibSea_Graph, PathSet graph, String COLUMN_ATTRIBUTE)
    {
	int tempNode = 0; 
	int numNodes = LibSea_Graph.getNumNodes();
	GraphInfo graphInfo = new GraphInfo();
	Vector AttributeNodes = new Vector();
	boolean T_or_F;

	//Extract the Column Attribute
	AttributeDefinitionIterator AttributeMainIterator =
	    LibSea_Graph.getAttributeDefinition(COLUMN_ATTRIBUTE);
	AttributesByAttributeIterator AttributeIterator = AttributeMainIterator.getNodeAttributes();
	if(AttributeIterator.atEnd())
	    {
		System.out.println("======================================================");
		System.out.println("You must select an Attribute that contains Complete Node Values");
		graphInfo.printAttributeInfo(LibSea_Graph);		
		System.exit(1);
	    }

	int type = AttributeMainIterator.getType().getType();
	
	//Initialize the array with everything set to false
	if(type == 0)
	    {
		for(int i = 0; i < numNodes; i++)
		    {
			AttributeNodes.addElement("False");
		    }
	    }

	if(type != 0)
	    {
		while(!AttributeIterator.atEnd())
		    {
			tempNode++;
			AttributeIterator.advance();
		    }
	    }

	AttributeIterator.rewind();

	if(type != 0)
	    {
		if(!(tempNode == numNodes))
		    {
			System.out.println("======================================================");
			System.out.println("You must select an Attribute that contains Complete Node Values.");
			System.out.println("Except for Boolean values.  PlotPaths will auto fill False values.");
			graphInfo.printAttributeInfo(LibSea_Graph);
			System.exit(1);
		    }
	    }

	while(!AttributeIterator.atEnd())
	    {
		ValueIterator AttributeValueIterator = AttributeIterator.getAttributeValues();

		switch(type)
		    {
		    case 0:
			{
			    //Handle case where the type is a Boolean
			    T_or_F = AttributeValueIterator.getBooleanValue();
			    if(T_or_F)
				{
				    AttributeNodes.insertElementAt("True", AttributeIterator.getObjectID());
				}
			    break;
			}
		    case 1:
			{
			    //Handle case where the type is an Integer
			    Integer addInt = new Integer(AttributeValueIterator.getIntegerValue());
			    AttributeNodes.addElement(addInt.toString());
			    break;
			}
		    case 4:
			{
			    //Handle case where the type is a String.
			    AttributeNodes.addElement(AttributeValueIterator.getStringValue());
			    break;
			}
		    default:
			{
			    System.out.println("======================================================");
			    System.out.println("The attribute type is not handled.");
			    System.out.println("Please select an attribute with one of the following types:");
			    System.out.println("Boolean, Integer, or String");
			    graphInfo.printAttributeInfo(LibSea_Graph);
			    System.exit(1);
			}
		    }
		AttributeIterator.advance();
	    }

	String[] nodeList = graph.getAllNodes();
	for(int i = 0; i < nodeList.length; i++)
	    {
		Integer nodes = new Integer(nodeList[i]);
		String tempString = (String)AttributeNodes.get(nodes.intValue());
		graph.setAttributeName(nodeList[i], tempString);
	    }
    
	return graph;
    }
    
    /**
     *  One stop shopping for all the attributes that need to be added
     *  in order to create an otter file.
     *  @param LibSea_Graph LibSea graph object that will be modified.
     *  @param Display_Graph Display Object that contains the node info.
     */
    public void addLibSeaAttributes(Graph LibSea_Graph, DisplayObjects Display_Graph)
    {
	addScreenXYAttributes(LibSea_Graph, Display_Graph);
	addRejectAttribute(LibSea_Graph, Display_Graph);
	addNameAttribute(LibSea_Graph, Display_Graph);
	addTotalUsageAttribute(LibSea_Graph, Display_Graph);
	addOutboundUsageAttribute(LibSea_Graph, Display_Graph);
	addPairwiseUsageAttribute(LibSea_Graph, Display_Graph);
	addInboundUsageAttribute(LibSea_Graph, Display_Graph);
	addUsageCountAttribute(LibSea_Graph, Display_Graph);
    }



    /**
     *  Adds two attributes to the LibSea_Graph file for the (x,y) 
     *  screen coordinates.
     *  PRE-Condition:
     *  All proccessing on the Display_Graph must be completed prior
     *  to calling addSreenXYAttributes.
     *  @param LibSea_Graph The LibSea graph object.
     *  @param Display_Graph A DisplayObjects object.
     */
    private void addScreenXYAttributes(Graph LibSea_Graph, DisplayObjects Display_Graph)
    {
	AttributeDefinitionIterator MainIterator1 = LibSea_Graph.getAttributeDefinitions();
	AttributeDefinitionIterator MainIterator2 = LibSea_Graph.getAttributeDefinitions();
	try
	    {
		MainIterator1.addMutableAttributeDefinition("ScreenX", ValueType.INTEGER, 
							    null, null, true, false, false);
		MainIterator2.addMutableAttributeDefinition("ScreenY", ValueType.INTEGER, 
							    null, null, true, false, false);
	    }
	catch(DuplicateObjectException e)
	    {
		String msg = "Duplicate Object: " + e.toString();
		System.err.println(msg);
		System.err.println("If attributes ScreenX and ScreenY exist, then this file");
		System.err.println("was probably already processed with plotpath.");
		graphInfo.printAttributeInfo(LibSea_Graph);
		System.exit(1);
	    }

	int id1 = MainIterator1.getID();
	int id2 = MainIterator2.getID();


	// -999 is being used as a reject code.
	// Set values to -999 for the ScreenX and ScreenY attributes.
	// Nodes that have not been rejected will replace -999 with a proper value.
	for(int i = 0; i < LibSea_Graph.getNumNodes(); i++)
	    {
		LibSea_Graph.setIntegerAttribute(ObjectType.NODE, 
						 i, id1, -999);
		LibSea_Graph.setIntegerAttribute(ObjectType.NODE, 
						 i, id2, -999);
	    }

	
	java.util.Iterator tmpNodeIterator = (java.util.Iterator)Display_Graph.getNodeEnumeration();
	while(tmpNodeIterator.hasNext())
	    {
		Node tmpNode = (Node)tmpNodeIterator.next();
		if(!tmpNode.isHidden()) 
		    {
			LibSea_Graph.setIntegerAttribute(ObjectType.NODE, 
							 tmpNode.nodeNumber, id1, tmpNode.x);
			LibSea_Graph.setIntegerAttribute(ObjectType.NODE, 
							 tmpNode.nodeNumber, id2, tmpNode.y);
		    } 
	    }
    }
    
    /**
     *  Adds the Reject Attribute.  Node and Path will be set to true if it
     *  has been rejected by loadPaths.  
     *  Pre-Condition:  The loadPaths method must have already been called 
     *  on this object so that the Vector can be populated.  Otherwise 
     *  incorrect results may be added to the Reject Attribute.
     *  @param LibSea_Graph The libsea graph file that will be written to.
     *  @param Display_Graph The display object that contains the proper node info.
     */
    private void addRejectAttribute(Graph LibSea_Graph, DisplayObjects Display_Graph)
    {
	AttributeDefinitionIterator MainIterator1 = LibSea_Graph.getAttributeDefinitions();

	//Create the attributes
	try
	    {
		MainIterator1.addMutableAttributeDefinition("rejected", ValueType.BOOLEAN, 
							    null, null, true, false, true);
	    }
	catch(DuplicateObjectException e)
	    {
		String msg = "Duplicate Object: " + e.toString();
		System.err.println(msg);
		System.err.println("If attribute Rejects exists, then this file");
		System.err.println("was probably already processed with plotpath.");
		graphInfo.printAttributeInfo(LibSea_Graph);
		System.exit(1);
	    }

	int id1 = MainIterator1.getID();

	// need to initialize all the values to true so nodes that
	// were used can set value to false.
	for(int i = 0; i < LibSea_Graph.getNumNodes(); i++)
	    {
		LibSea_Graph.setBooleanAttribute(ObjectType.NODE, i, id1, true);
	    }

	java.util.Iterator tmpNodeIterator = (java.util.Iterator)Display_Graph.getNodeEnumeration();
	while(tmpNodeIterator.hasNext())
	    {
		Node tmpNode = (Node)tmpNodeIterator.next();
		if(!tmpNode.isHidden()) 
		    {
			LibSea_Graph.setBooleanAttribute(ObjectType.NODE, 
							 tmpNode.nodeNumber, id1, false);
		    } 
	    }
	
	// Add rejected paths
	if(!rejectedPaths.isEmpty())
	    {
		for(int i = 0; i < rejectedPaths.size(); i++)
		    {
			LibSea_Graph.setBooleanAttribute(ObjectType.PATH, 
							 ((Integer)(rejectedPaths.get(i))).intValue(), id1, true);
		    }
	    }
    }

    /**
     *  Add the name attribute to the graph.
     *  @param LibSea_Graph The graph object that the attribute will be added to.
     *  @param Display_Graph The display object that contains the proper node info.
     */
    private void addNameAttribute(Graph LibSea_Graph, DisplayObjects Display_Graph)
    {
	AttributeDefinitionIterator MainIterator1 = LibSea_Graph.getAttributeDefinitions();

	//Create the attributes
	try
	    {
		MainIterator1.addMutableAttributeDefinition("name", ValueType.STRING, 
							    null, null, true, false, false);
	    }
	catch(DuplicateObjectException e)
	    {
		String msg = "Duplicate Object: " + e.toString();
		System.err.println(msg);
		System.err.println("If attribute name exists, then this file");
		System.err.println("was probably already processed with plotpath.");
		graphInfo.printAttributeInfo(LibSea_Graph);
		System.exit(1);
	    }

	int id1 = MainIterator1.getID();

	// need to initialize all the values so no null values exist.
	// workaround for bug in libsea.
	for(int i = 0; i < LibSea_Graph.getNumNodes(); i++)
	    {
		LibSea_Graph.setStringAttribute(ObjectType.NODE, i, id1, "no name");
	    }

	java.util.Iterator tmpNodeIterator = (java.util.Iterator)Display_Graph.getNodeEnumeration();
	while(tmpNodeIterator.hasNext())
	    {
		Node tmpNode = (Node)tmpNodeIterator.next();
		if(!tmpNode.isHidden()) 
		    {
			LibSea_Graph.setStringAttribute(ObjectType.NODE, 
							 tmpNode.nodeNumber, id1, tmpNode.name);
		    } 
	    }
    }

    /**
     *  Add the total usage attribute.  Used to store the processed data from
     *  plotpath for the total usage of sources and destinations.
     *  @param LibSea_Graph The graph object that the attribute will be added to.
     *  @param Display_Graph The display object that contains the proper node info.
     */
    private void addTotalUsageAttribute(Graph LibSea_Graph, DisplayObjects Display_Graph)
    {
	java.util.Iterator lineEnum;
	Object[] categoryAttributeList = Display_Graph.enumerateAttributes("Total Usage");

	AttributeDefinitionIterator MainIterator1 = LibSea_Graph.getAttributeDefinitions();
	AttributeDefinitionIterator MainIterator2 = LibSea_Graph.getAttributeDefinitions();

	//Create the attributes
	try
	    {
		MainIterator1.addMutableAttributeDefinition("TotalUsageValues", ValueType.STRING, 
							    null, null, false, true, false);
		MainIterator2.addMutableAttributeDefinition("TotalUsageHeaderOrder", ValueType.INTEGER, 
							    null, null, true, false, false);
	    }
	catch(DuplicateObjectException e)
	    {
		String msg = "Duplicate Object: " + e.toString();
		System.err.println(msg);
		System.err.println("If attribute TotalUsage exists, then this file");
		System.err.println("was probably already processed with plotpath.");
		graphInfo.printAttributeInfo(LibSea_Graph);
		System.exit(1);
	    }

	int id1 = MainIterator1.getID();
	int id2 = MainIterator2.getID();

	// initialize attributes
	for(int i = 0; i < LibSea_Graph.getNumLinks(); i++)
	    {
		LibSea_Graph.setStringAttribute(ObjectType.LINK, i, id1, "not Used");
	    }

	for(int i = 0; i < LibSea_Graph.getNumNodes(); i++)
	    {
		LibSea_Graph.setIntegerAttribute(ObjectType.NODE, i, id2, -1);
	    }
	
	// print line data
	String pathWeight;
	String tempStrValues = "", tempStrHeader = "";
	String srcDstHashKey;
	lineEnum = (java.util.Iterator)Display_Graph.getLineEnumeration();
	for(int i = 0; lineEnum.hasNext(); i++) 
	    {
		Line tmpLine = (Line)lineEnum.next();
		Integer idx1, idx2, linkID, nodeValue;
		idx1 = new Integer(tmpLine.getCoord1().nodeNumber);
		idx2 = new Integer(tmpLine.getCoord2().nodeNumber);
		srcDstHashKey = idx1.toString() + "," + idx2.toString();

		tempStrValues = "";

		for(int k = 0; k < categoryAttributeList.length; k++) 
		    {
			pathWeight = (String)tmpLine.getAttribute(
								  "Total Usage",
								  (String)categoryAttributeList[k]);

			nodeValue = new Integer(categoryAttributeList[k].toString());

			LibSea_Graph.setIntegerAttribute(ObjectType.NODE,
							 nodeValue.intValue(), id2, k);

			if(pathWeight != null) 
			    {
				tempStrValues = tempStrValues + pathWeight;
			    } 
			else 
			    {
				tempStrValues = tempStrValues + "not selected";
			    }
			if(k < categoryAttributeList.length-1) 
			    {
				tempStrValues = tempStrValues + "'";
			    }
		    }
		
		if(linkHash.containsKey(srcDstHashKey))
		    {
			linkID = (Integer)(linkHash.get(srcDstHashKey));
			LibSea_Graph.setStringAttribute(ObjectType.LINK, 
							linkID.intValue(), id1, tempStrValues);
		    }
	    }
    }

    /**
     *  Add the Outbound usage attribute.  Used to store the processed data from
     *  plotpath for the outbound usage of sources.
     *  @param LibSea_Graph The graph object that the attribute will be added to.
     *  @param Display_Graph The display object that contains the proper node info.
     */
    private void addOutboundUsageAttribute(Graph LibSea_Graph, DisplayObjects Display_Graph)
    {
	java.util.Iterator lineEnum;
	Object[] categoryAttributeList = Display_Graph.enumerateAttributes("Outbound Usage");

	AttributeDefinitionIterator MainIterator1 = LibSea_Graph.getAttributeDefinitions();
	AttributeDefinitionIterator MainIterator2 = LibSea_Graph.getAttributeDefinitions();

	//Create the attributes
	try
	    {
		MainIterator1.addMutableAttributeDefinition("OutboundUsageValues", ValueType.STRING, 
							    null, null, false, true, false);
		MainIterator2.addMutableAttributeDefinition("OutboundUsageHeaderOrder", ValueType.INTEGER, 
							    null, null, true, false, false);
	    }
	catch(DuplicateObjectException e)
	    {
		String msg = "Duplicate Object: " + e.toString();
		System.err.println(msg);
		System.err.println("If attribute OutboundUsage exists, then this file");
		System.err.println("was probably already processed with plotpath.");
		graphInfo.printAttributeInfo(LibSea_Graph);
		System.exit(1);
	    }

	int id1 = MainIterator1.getID();
	int id2 = MainIterator2.getID();

	// initialize attributes
	for(int i = 0; i < LibSea_Graph.getNumLinks(); i++)
	    {
		LibSea_Graph.setStringAttribute(ObjectType.LINK, i, id1, "not Used");
	    }

	for(int i = 0; i < LibSea_Graph.getNumNodes(); i++)
	    {
		LibSea_Graph.setIntegerAttribute(ObjectType.NODE, i, id2, -1);
	    }
	
	// print line data
	String pathWeight;
	String tempStrValues = "", tempStrHeader = "";
	String srcDstHashKey;
	lineEnum = (java.util.Iterator)Display_Graph.getLineEnumeration();
	for(int i = 0; lineEnum.hasNext(); i++) 
	    {
		Line tmpLine = (Line)lineEnum.next();
		Integer idx1, idx2, linkID, nodeValue;
		idx1 = new Integer(tmpLine.getCoord1().nodeNumber);
		idx2 = new Integer(tmpLine.getCoord2().nodeNumber);
		srcDstHashKey = idx1.toString() + "," + idx2.toString();

		tempStrValues = "";

		for(int k = 0; k < categoryAttributeList.length; k++) 
		    {
			pathWeight = (String)tmpLine.getAttribute(
								  "Outbound Usage",
								  (String)categoryAttributeList[k]);

			nodeValue = new Integer(categoryAttributeList[k].toString());

			LibSea_Graph.setIntegerAttribute(ObjectType.NODE,
							 nodeValue.intValue(), id2, k);

			if(pathWeight != null) 
			    {
				tempStrValues = tempStrValues + pathWeight;
			    } 
			else 
			    {
				tempStrValues = tempStrValues + "not selected";
			    }
			if(k < categoryAttributeList.length-1) 
			    {
				tempStrValues = tempStrValues + "'";
			    }
		    }
		
		if(linkHash.containsKey(srcDstHashKey))
		    {
			linkID = (Integer)(linkHash.get(srcDstHashKey));
			LibSea_Graph.setStringAttribute(ObjectType.LINK, 
							linkID.intValue(), id1, tempStrValues);
		    }
	    }
    }

    /**
     *  Add the pairwise usage attribute.  Used to store the processed data from
     *  plotpath for the pairwise usage between sources and destinations.
     *  @param LibSea_Graph The graph object that the attribute will be added to.
     *  @param Display_Graph The display object that contains the proper node info.
     */
    private void addPairwiseUsageAttribute(Graph LibSea_Graph, DisplayObjects Display_Graph)
    {
	java.util.Iterator lineEnum;
	Object[] categoryAttributeList = Display_Graph.enumerateAttributes("Pairwise Usage");

	AttributeDefinitionIterator MainIterator1 = LibSea_Graph.getAttributeDefinitions();
	AttributeDefinitionIterator MainIterator2 = LibSea_Graph.getAttributeDefinitions();

	//Create the attributes
	try
	    {
		MainIterator1.addMutableAttributeDefinition("PairwiseUsageValues", ValueType.STRING, 
							    null, null, false, true, false);
		MainIterator2.addMutableAttributeDefinition("PairwiseUsageHeaderOrder", ValueType.INTEGER, 
							    null, null, false, false, true);
	    }
	catch(DuplicateObjectException e)
	    {
		String msg = "Duplicate Object: " + e.toString();
		System.err.println(msg);
		System.err.println("If attribute PairwiseUsage exists, then this file");
		System.err.println("was probably already processed with plotpath.");
		graphInfo.printAttributeInfo(LibSea_Graph);
		System.exit(1);
	    }

	int id1 = MainIterator1.getID();
	int id2 = MainIterator2.getID();

	// initialize attributes
	for(int i = 0; i < LibSea_Graph.getNumLinks(); i++)
	    {
		LibSea_Graph.setStringAttribute(ObjectType.LINK, i, id1, "not Used");
	    }

	for(int i = 0; i < LibSea_Graph.getNumPaths(); i++)
	    {
		LibSea_Graph.setIntegerAttribute(ObjectType.PATH, i, id2, -1);
	    }
	
	// print line data
	String pathWeight;
	String tempStrValues = "", tempStrHeader = "";
	String srcDstHashKey;
	lineEnum = (java.util.Iterator)Display_Graph.getLineEnumeration();
	for(int i = 0; lineEnum.hasNext(); i++) 
	    {
		Line tmpLine = (Line)lineEnum.next();
		Integer idx1, idx2, linkID, nodeValue;
		idx1 = new Integer(tmpLine.getCoord1().nodeNumber);
		idx2 = new Integer(tmpLine.getCoord2().nodeNumber);
		srcDstHashKey = idx1.toString() + "," + idx2.toString();

		tempStrValues = "";

		for(int k = 0; k < categoryAttributeList.length; k++) 
		    {
			pathWeight = (String)tmpLine.getAttribute(
								  "Pairwise Usage",
								  (String)categoryAttributeList[k]);

			StringTokenizer srcDstTokens = new StringTokenizer((String)categoryAttributeList[k]);
			String srcDst = srcDstTokens.nextToken();
			srcDstTokens.nextToken();
			srcDst = srcDst + "," + srcDstTokens.nextToken();
			
			if(srcDstHashtable.containsKey(srcDst))
			    {
				HashSet pathIDHashSet = (HashSet)srcDstHashtable.get(srcDst);
				java.util.Iterator pathIDIterator = pathIDHashSet.iterator();
				while(pathIDIterator.hasNext())
				    {
					Integer pathID = (Integer)pathIDIterator.next();
					LibSea_Graph.setIntegerAttribute(ObjectType.PATH,
									 pathID.intValue(), id2, k);
				    }
			    }
			else
			    {
				// The path should have existed.
				System.out.println("A Pairwise attribute has encountered errors.");
			    }

			if(pathWeight != null) 
			    {
				tempStrValues = tempStrValues + pathWeight;
			    } 
			else 
			    {
				tempStrValues = tempStrValues + "not selected";
			    }
			if(k < categoryAttributeList.length-1) 
			    {
				tempStrValues = tempStrValues + "'";
			    }
		    }
		
		if(linkHash.containsKey(srcDstHashKey))
		    {
			linkID = (Integer)(linkHash.get(srcDstHashKey));
			LibSea_Graph.setStringAttribute(ObjectType.LINK, 
							linkID.intValue(), id1, tempStrValues);
		    }
	    }
    }

    /**
     *  Add the Inbound usage attribute.  Used to store the processed data from
     *  plotpath for the inbound usage destinations.
     *  @param LibSea_Graph The graph object that the attribute will be added to.
     *  @param Display_Graph The display object that contains the proper node info.
     */
    private void addInboundUsageAttribute(Graph LibSea_Graph, DisplayObjects Display_Graph)
    {
	java.util.Iterator lineEnum;
	Object[] categoryAttributeList = Display_Graph.enumerateAttributes("Inbound Usage");

	AttributeDefinitionIterator MainIterator1 = LibSea_Graph.getAttributeDefinitions();
	AttributeDefinitionIterator MainIterator2 = LibSea_Graph.getAttributeDefinitions();

	//Create the attributes
	try
	    {
		MainIterator1.addMutableAttributeDefinition("InboundUsageValues", ValueType.STRING, 
							    null, null, false, true, false);
		MainIterator2.addMutableAttributeDefinition("InboundUsageHeaderOrder", ValueType.INTEGER, 
							    null, null, true, false, false);
	    }
	catch(DuplicateObjectException e)
	    {
		String msg = "Duplicate Object: " + e.toString();
		System.err.println(msg);
		System.err.println("If attribute Inbound Usage exists, then this file");
		System.err.println("was probably already processed with plotpath.");
		graphInfo.printAttributeInfo(LibSea_Graph);
		System.exit(1);
	    }

	int id1 = MainIterator1.getID();
	int id2 = MainIterator2.getID();

	// initialize attributes
	for(int i = 0; i < LibSea_Graph.getNumLinks(); i++)
	    {
		LibSea_Graph.setStringAttribute(ObjectType.LINK, i, id1, "not Used");
	    }

	for(int i = 0; i < LibSea_Graph.getNumNodes(); i++)
	    {
		LibSea_Graph.setIntegerAttribute(ObjectType.NODE, i, id2, -1);
	    }
	
	// print line data
	String pathWeight;
	String tempStrValues = "", tempStrHeader = "";
	String srcDstHashKey;
	lineEnum = (java.util.Iterator)Display_Graph.getLineEnumeration();
	for(int i = 0; lineEnum.hasNext(); i++) 
	    {
		Line tmpLine = (Line)lineEnum.next();
		Integer idx1, idx2, linkID, nodeValue;
		idx1 = new Integer(tmpLine.getCoord1().nodeNumber);
		idx2 = new Integer(tmpLine.getCoord2().nodeNumber);
		srcDstHashKey = idx1.toString() + "," + idx2.toString();

		tempStrValues = "";

		for(int k = 0; k < categoryAttributeList.length; k++) 
		    {
			pathWeight = (String)tmpLine.getAttribute(
								  "Inbound Usage",
								  (String)categoryAttributeList[k]);

			nodeValue = new Integer(categoryAttributeList[k].toString());
			LibSea_Graph.setIntegerAttribute(ObjectType.NODE,
							 nodeValue.intValue(), id2, k);

			if(pathWeight != null) 
			    {
				tempStrValues = tempStrValues + pathWeight;
			    } 
			else 
			    {
				tempStrValues = tempStrValues + "not selected";
			    }
			if(k < categoryAttributeList.length-1) 
			    {
				tempStrValues = tempStrValues + "'";
			    }
		    }
		
		if(linkHash.containsKey(srcDstHashKey))
		    {
			linkID = (Integer)(linkHash.get(srcDstHashKey));
			LibSea_Graph.setStringAttribute(ObjectType.LINK, 
							linkID.intValue(), id1, tempStrValues);
		    }
	    }
    }

    /**
     *  Add the Usage Count attribute.  Used to store the processed data from
     *  plotpath for the total usage count of links for a particular path.
     *  @param LibSea_Graph The graph object that the attribute will be added to.
     *  @param Display_Graph The display object that contains the proper node info.
     */
    private void addUsageCountAttribute(Graph LibSea_Graph, DisplayObjects Display_Graph)
    {
	java.util.Iterator lineEnum;
	Object[] categoryAttributeList = Display_Graph.enumerateAttributes("Usage Count");

	AttributeDefinitionIterator MainIterator1 = LibSea_Graph.getAttributeDefinitions();

	//Create the attributes
	try
	    {
		MainIterator1.addMutableAttributeDefinition("UsageCountValues", ValueType.STRING, 
							    null, null, false, true, false);
	    }
	catch(DuplicateObjectException e)
	    {
		String msg = "Duplicate Object: " + e.toString();
		System.err.println(msg);
		System.err.println("If attribute Usage Count exists, then this file");
		System.err.println("was probably already processed with plotpath.");
		graphInfo.printAttributeInfo(LibSea_Graph);
		System.exit(1);
	    }

	int id1 = MainIterator1.getID();

	// initialize attributes
	for(int i = 0; i < LibSea_Graph.getNumLinks(); i++)
	    {
		LibSea_Graph.setStringAttribute(ObjectType.LINK, i, id1, "not Used");
	    }

	// print line data
	String pathWeight;
	String tempStrValues = "", tempStrHeader = "";
	String srcDstHashKey;
	lineEnum = (java.util.Iterator)Display_Graph.getLineEnumeration();
	for(int i = 0; lineEnum.hasNext(); i++) 
	    {
		Line tmpLine = (Line)lineEnum.next();
		Integer idx1, idx2, linkID, nodeValue;
		idx1 = new Integer(tmpLine.getCoord1().nodeNumber);
		idx2 = new Integer(tmpLine.getCoord2().nodeNumber);
		srcDstHashKey = idx1.toString() + "," + idx2.toString();

		tempStrValues = "";

		for(int k = 0; k < categoryAttributeList.length; k++) 
		    {
			pathWeight = (String)tmpLine.getAttribute(
								  "Usage Count",
								  (String)categoryAttributeList[k]);

			if(pathWeight != null) 
			    {
				tempStrValues = tempStrValues + pathWeight;
			    } 
			else 
			    {
				tempStrValues = tempStrValues + "not selected";
			    }
			if(k < categoryAttributeList.length-1) 
			    {
				tempStrValues = tempStrValues + "'";
			    }
		    }
		
		if(linkHash.containsKey(srcDstHashKey))
		    {
			linkID = (Integer)(linkHash.get(srcDstHashKey));
			LibSea_Graph.setStringAttribute(ObjectType.LINK, 
							linkID.intValue(), id1, tempStrValues);
		    }
	    }
    }

    /**
     *  Adds the attributes to the graph file that will associate nodes to 
     *  a label.  Still in beta since how to handle labels has not been
     *  fully determined.
     *  PRE-Condition:
     *  All proccessing on the Display_Graph must be completed prior
     *  to calling addColumnLabels.
     *  @param LibSea_Graph LibSea graph file that attributes will be added to.
     *  @param Display_Graph Object that contains the node <-> label associations.
     */
    private void addColumnLabels(Graph LibSea_Graph, DisplayObjects Display_Graph)
    {
	GraphInfo graphInfo = new GraphInfo();

	AttributeDefinitionIterator MainIterator1 = LibSea_Graph.getAttributeDefinitions();

	//Create the attributes
	try
	    {
		MainIterator1.addMutableAttributeDefinition("ColNum", ValueType.INTEGER, 
							    null, null, true, false, false);
	    }
	catch(DuplicateObjectException e)
	    {
		String msg = "Duplicate Object: " + e.toString();
		System.err.println(msg);
		System.err.println("If attributes ColAssoc, ColX, ColY, and ColName exist, then this file");
		System.err.println("was probably already processed with plotpath.");
		graphInfo.printAttributeInfo(LibSea_Graph);
		System.exit(1);
	    }

	int id1 = MainIterator1.getID();

	java.util.Iterator tmpNodeIterator = (java.util.Iterator)Display_Graph.getNodeEnumeration();
	while(tmpNodeIterator.hasNext())
	    {
		Node tmpNode = (Node)tmpNodeIterator.next();
		if(!tmpNode.isHidden()) 
		    {
			LibSea_Graph.setIntegerAttribute(ObjectType.NODE, 
							 tmpNode.nodeNumber, id1, tmpNode.colNum);
		    } 
	    }
	
    }
}
