//==========================================================================
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// Permission to use, copy, modify and distribute any part of this Makefile
// for educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above 
// copyright notice, this paragraph and the following paragraphs appear 
// in all copies.
//
// Those desiring to incorporate this into commercial products or use for
// commercial purposes should contact the
//
//      Technology Transfer Office, University of California, San Diego,
//      9500 Gilman Drive, La Jolla, CA 92093-0910
//      Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS PLOTPATH, EVEN 
// IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
//
// THE PLOTPATH, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF 
// CALIFORNIA PLOTPATH NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF 
// ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
// PURPOSE, OR THAT THE USE OF THE SKITTERTRACES WILL NOT INFRINGE ANY 
// PATENT, TRADEMARK OR OTHER RIGHTS.
//
// PlotPaths is developed by John Gallagher Bradley L. Huffaker at the
// University of California, San Diego under the Cooperative Association
// for Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by NSF "Internet Atlas" grant ANI-I-9996248, NSF "CAIDA Core"
// NCR-9711092, and by CAIDA memebers.
//==========================================================================
package org.caida.otter;

import java.util.Vector;

/**
 * Nodes may be connected by lines.  This class otherwise serves as a
 *    container for all the information needed to draw the node.
 */
public class Node extends Coordinate {

    /**
     * Make a new node
     * 
     * @param index The index of the new node.
     */
    public Node(int index) {
	super(index);
    }

    /**
     * Make a new node.
     *
     * @param index The index of the new node.
     * @param col The column number of the new node.
     * @param row The row number of the new node.
     * @param name The name of the new node.
     * @param colName The column name of the new node.
     */
    public Node(int index, int col, int row, String name, String colName) {
	super(index, col, row, name, colName);
    }

    /**
     * Make a new node.
     * John Tusson Tue Aug 14 16:26:41 PDT 2001
     * @param index The index of the new node.
     * @param col The column number of the new node.
     * @param row The row number of the new node.
     * @param name The name of the new node.
     * @param colName The column name of the new node.
     * @param nodeNumber
     */
    public Node(int index, int col, int row, String name, String colName, int nodeNumber) {
	super(index, col, row, name, colName, nodeNumber);
    }



    /**
     * Add a node neighbor by its index (either parent or child)
     *
     * @param idx The index of the node that will be our neighbor.
     */
    public void addNeighborNode(Node neighbor) {

	// avoid duplicates
	for(int i=0; i < neighborNodes.size(); i++) {
	    if(neighborNodes.elementAt(i).equals(neighbor)) {
		return;
	    }
	}

	neighborNodes.addElement(neighbor);
    }

    /**
     * Get neighbor node list.
     *
     * @return array of our neighbors.
     */
    public Node[] getNeighborNodes() {
	Node[] ret_val = new Node[neighborNodes.size()];
    
	for(int i=0; i < neighborNodes.size(); i++) {
	    ret_val[i] = (Node)neighborNodes.elementAt(i);
	}
	return ret_val;
    }

    /**
     * See if the node has any neighbors at all.
     *
     * @return true if we have neighbors, false otherwise.
     */
    public boolean hasNeighborNodes() {
	if(neighborNodes.size() >= 0) {
	    return true;
	} else {
	    return false;
	}
    }

    /**
     * Flip root flag on node.
     *
     * @param newIsRoot true if we want this to be a root, false otherwise.
     */
    public void setIsRoot(boolean newIsRoot) {
	isRoot = newIsRoot;
    }

    /**
     * Test isRoot flag.
     *
     * @return true if this node is a root, false otherwise.
     */
    public boolean isRoot() {
	return isRoot;
    }

    /**
     * Set isHidden flag.
     *
     * @param isHidden true if we want this node to be hidden, false if otherwise.
     */
    public void setIsHidden(boolean isHidden) {
	this.isHidden = isHidden;
    }

    /** 
     * Test isHidden flag
     *
     * @return true if this node is hidden, flase otherwise.
     */
    public boolean isHidden() {
	return isHidden;
    }

    /**
     * Add a category to the set.
     */
    public Object addCategory(Object category) {
	return attribSet.addCategory(category);
    }

    /**
     * Add an attribute to inside a category
     */
    public Object addAttribute(Object category, Object attribute, 
			    Object value) {
	return attribSet.addAttribute(category, attribute, value);
    }

    /**
     * Get an attribute from inside a category
     */
    public Object getAttribute(Object category, Object attribute) {
	return attribSet.getAttribute(category, attribute);
    }

    /**
     * Remove an attribute from the set of categories.
     */
    public Object deleteCategory(Object category) {
	return attribSet.deleteCategory(category);
    }

    /**
     * Remove an attribute from inside a category
     */
    public Object deleteAttribute(Object category, Object attribute) {
	return attribSet.deleteAttribute(category, attribute);
    }

    /**
     * Replace an attribute inside a category
     */
    public Object changeAttribute(Object category, Object attribute, 
			    Object value) {

	return attribSet.changeAttribute(category, attribute, value); 
    }
    
    /**
     * Get a list of objects that represent all of the categories
     *    contained in this node's category set.
     */
    public Object[] enumerateCategories() {
	return attribSet.enumerateCategories();
    }

    /** 
     * Get a list of attributes in a given category
     */
    public Object[] enumerateAttributes(Object category) {
	return attribSet.enumerateAttributes(category);
    }

    private AttributeSet attribSet = new AttributeSet();
    private Vector neighborNodes = new Vector();
    private boolean isRoot = false;
    private boolean isHidden = false;
}		 
