//==========================================================================
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// Permission to use, copy, modify and distribute any part of this Makefile
// for educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above 
// copyright notice, this paragraph and the following paragraphs appear 
// in all copies.
//
// Those desiring to incorporate this into commercial products or use for
// commercial purposes should contact the
//
//      Technology Transfer Office, University of California, San Diego,
//      9500 Gilman Drive, La Jolla, CA 92093-0910
//      Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS PLOTPATH, EVEN 
// IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
//
// THE PLOTPATH, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF 
// CALIFORNIA PLOTPATH NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF 
// ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
// PURPOSE, OR THAT THE USE OF THE SKITTERTRACES WILL NOT INFRINGE ANY 
// PATENT, TRADEMARK OR OTHER RIGHTS.
//
// PlotPaths is developed by John Gallagher Bradley L. Huffaker at the
// University of California, San Diego under the Cooperative Association
// for Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by NSF "Internet Atlas" grant ANI-I-9996248, NSF "CAIDA Core"
// NCR-9711092, and by CAIDA memebers.
//==========================================================================
package org.caida.otter;

import java.util.Vector;

/*
 * TODO make a Path an individual object so we don't have to maintain
 *  these parallel data structures.
 *
 */

/**
 * Store a collection of Paths and their associated statistics
 */
public class PathSet {

    /**
     * Register the name of nodes that are "not found"
     */
    PathSet(String unknownNodeName) {
	this.unknownNodeName = unknownNodeName;
    }

    /** 
     * Add a path to the path list (no calculcations yet)
     *
     * @param newPath Vector of Strings, each one representing a single
     *          node in the path.
     * @param weight A path weight associated with the path.
     * @param orientation The orientation of the new path.
     */
    public void addPath(Vector newPath, float weight, boolean orientation) {
	if(newPath == null) {
	    throw new IllegalArgumentException("new path was null");
	}

	// done seperately so if it fails, no updates have been made to 
	//    class vars
	for(int i = newPath.size()-1; i >= 0; i--) {
	    if(!(newPath.elementAt(i) instanceof String)) {
		throw new IllegalArgumentException(
				"Element in path found not to be a String");
	    }
	}

	pathOrientList.addElement(new Boolean(orientation));

	//outputPath(newPath, "Before Edit");
	checkForUnknownNodes(newPath, orientation);
	//outputPath(newPath, "After Edit");


	pathList.addElement(newPath);
	pathWeightList.addElement(new Float(weight));

	for(int i = newPath.size()-1; i >= 0; i--) {
	    updateNodeIndex((String)newPath.elementAt(i), pathList.size()-1, i);
	}
    }

    /**
     * Output the contents of a path to the screen
     */
    private void outputPath(Vector path, String description) {
	System.out.println(description);
	System.out.println("-------------------------");
	for(int i=0; i < path.size(); i++) {
	    System.out.println("Path Data:"+(String)path.elementAt(i));
	}
	System.out.println();
    }

    /** 
     * Update the master list of node locations when a new node
     * is stored in the class.  For each new path the master list
     * keeps a pointer to each path that each node is located in.
     * Find that node and then add the pointer to the path.  If that
     * node doesn't exist yet, add a new one in the master list.
     *
     * @param nodeName The node name that is in the new path
     * @param pathIndex The index of the path in the path list
     * @param elemIndex The index of the node within the new path
     */
    private void updateNodeIndex(String nodeName, int pathIndex, 
				    int elemIndex) {

	boolean alreadyInList = false;

	// linear search for now...
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {

		// add new indexes to list of old indexes
		Vector pointerList = (Vector)nodeCoordList.elementAt(i);
		pointerList.addElement(new Coordinate(pathIndex, elemIndex));
		alreadyInList = true;
		break;
	    }
	}
	if(!alreadyInList) {

	    // add node name 
	    nodeNameList.addElement(nodeName);

	    // add uninitialized graph depth
	    nodeMaxDepthList.addElement(new Integer(-1));
	    nodeLatList.addElement(new Float(0));
	    nodeLongList.addElement(new Float(0));
	    nodeCityList.addElement(new String(""));
	    nodeASNameList.addElement(new String(""));
	    nodeASNumberList.addElement(new Integer("-199"));
	    //Added by John Tusson Tue Aug 14 14:38:06 PDT 2001
	    nodeAttributeNameList.addElement(new String(""));

	    // add node coordinates
	    Vector tmpVect = new Vector();
	    tmpVect.addElement(new Coordinate(pathIndex, elemIndex));
	    nodeCoordList.addElement(tmpVect);
	}
    }


    /**
     * Go through all paths and assign appropriate names to unknownNodeName 
     *  nodes.  This is necessary to build an appropriate number of new 
     *  links for the respective unknown nodes (many nodes are compressed 
     *  across paths)
     *
     *  Note it is invalid to have an unknown node at either the start or end 
     *  of a path.
     *
     * @param path Vector of nodes representing the new path being checked
     *          for "Unknown" nodes.
     * @param orientation The orientation of the path supplied.
     */
    private void checkForUnknownNodes(Vector path, boolean orientation) {

	if(((String)path.elementAt(0)).equals(unknownNodeName) || 
		((String)path.elementAt(path.size()-1)).equals(
							unknownNodeName)) {
	    throw new IllegalArgumentException(
				"First or last node must not be unknown ("+
				unknownNodeName+")");
	}

	int unknownGroupStartIdx = -1;
	int unknownGroupEndIdx   = -1;

	boolean inUnknownGroup = false;
	// b/c first and last node must not be unknown
	for(int i=1; i < path.size()-1; i++) {
	    if(((String)path.elementAt(i)).equals(unknownNodeName) && 
			inUnknownGroup == false &&
			i == path.size()-2) {

		unknownGroupStartIdx = i;
		unknownGroupEndIdx = i;
		renameUnknownNodes( path, 
				    unknownGroupStartIdx, 
				    unknownGroupEndIdx, 
				    orientation);

	    } else if(((String)path.elementAt(i)).equals(unknownNodeName) && 
			inUnknownGroup == false) {

		// start an unknown group
		inUnknownGroup = true;
		unknownGroupStartIdx = i;

	    } else if(!((String)path.elementAt(i)).equals(unknownNodeName) && 
			inUnknownGroup == true) {

		// end an unknown group
		inUnknownGroup = false;
		unknownGroupEndIdx = i-1;
		renameUnknownNodes( path, 
				    unknownGroupStartIdx, 
				    unknownGroupEndIdx, 
				    orientation);

	    } else if(((String)path.elementAt(i)).equals(unknownNodeName) && 
			inUnknownGroup == true && 
			i == path.size()-2) {

		// boundary condition
		inUnknownGroup = false;
		unknownGroupEndIdx = i;
		renameUnknownNodes( path,
				    unknownGroupStartIdx, 
				    unknownGroupEndIdx, 
				    orientation);
	    }
	}
    }

    /**
     * Actually assign names to an unknown group that has been singled out.
     *    the indexes represent the first and last unknown - names are extracted
     *    from the nodes just before and just after the unknown group.
     *
     * btw, an unknown group is a series of unknown nodes bordered by known 
     *   nodes. Inside the group, all nodes are assumed to be unknown.
     *
     * @param path The path being changed.
     * @param unknownGroupStartIdx The beginning of the group of unknown nodes.
     * @param unknownGroupEndIdx The end of the group of unknown nodes.
     */
    private void renameUnknownNodes(Vector path, int unknownGroupStartIdx, 
			int unknownGroupEndIdx, boolean orientation) {

	String beforeNode = (String)path.elementAt(unknownGroupStartIdx-1);
	String afterNode  = (String)path.elementAt(unknownGroupEndIdx+1);
	String unknownName;
	
	int fowardDist = 1;
	int backwardDist = unknownGroupEndIdx - unknownGroupStartIdx + 1;
	for(int i=unknownGroupStartIdx; i <= unknownGroupEndIdx; i++) {


	    // build unknown name
	    unknownName = "("+fowardDist+") "+beforeNode+
			    " ~~ ("+backwardDist+") "+afterNode;

	    if(orientation) {
		unknownName = "("+fowardDist+") "+beforeNode+ 
			    " ~~ ("+backwardDist+") "+afterNode;
	    } else {
		unknownName = "("+fowardDist+") "+afterNode+ 
			    " ~~ ("+backwardDist+") "+beforeNode;
	    }

	    path.removeElementAt(i);
	    path.insertElementAt(("(unknown) "+unknownName), i);
	    
	    System.out.println("unknownName : " + unknownName);
	    fowardDist++;
	    backwardDist--;


	}
    }

    /** 
     * Calculate the max depth over the entire graph.  All nodes should be
     *   reachable from at least one of the listed sources via either an
     *   inbound or outbound path.
     *
     * The sources used in the max depth calculation are the beginning of a 
     *   foward path or the end of a backward path.
     */
    public int calcMaxDepth() {
	Vector sourceNodes = new Vector();
	Vector currentPath;
	DirectedGraph graph = new DirectedGraph();

	// add all the paths (reversing ones w/ false orientation)
	for(int i = pathList.size()-1; i >= 0; i--) {
	    currentPath = (Vector)pathList.elementAt(i);

	    // build connected graph nodes
	    if(((Boolean)pathOrientList.elementAt(i)).booleanValue() == true && 
		currentPath.size() > 0) {

		//System.out.println("adding foward path:");

		String nodeName = (String)currentPath.elementAt(
			    currentPath.size()-1);

		GraphNode prevNode = new GraphNode(nodeName);  

		//System.out.println("--"+prevNode.getName());

		for(int j = currentPath.size()-2; j >= 0; j--) {
		    nodeName = (String)currentPath.elementAt(j);
		    prevNode = new GraphNode(SOURCE, nodeName, prevNode); 

		    //System.out.println("--"+prevNode.getName());
		}
		graph.addPath(prevNode, SOURCE);
		sourceNodes.addElement(prevNode);

	    } else if(currentPath.size() > 0) {

		//System.out.println("adding backward path:");

		String nodeName = (String)currentPath.elementAt(0);
		GraphNode prevNode = new GraphNode(nodeName);  

		//System.out.println("--"+prevNode.getName());

		for(int j = 1; j < currentPath.size(); j++) {
		    nodeName = (String)currentPath.elementAt(j);
		    prevNode = new GraphNode(SOURCE, nodeName, prevNode); 

		    //System.out.println("--"+prevNode.getName());
		}
		graph.addPath(prevNode, SOURCE);
		sourceNodes.addElement(prevNode);
	    }
	}

	// see what actually made it into the graph
	GraphNode[] nodes = graph.getAllNodes();


	// weed out duplicates
	PlotPathsUtil.removeDuplicatesFromVect(sourceNodes);

	for(int i=0; i < sourceNodes.size(); i++) {
	    System.out.println("Sources :  "+
			((GraphNode)sourceNodes.elementAt(i)).getName());
	}


	// call calcMaxDepth on all sources
	maxDepth = graph.calcMaxDepth(sourceNodes, SOURCE);


	// save maxDepth value
	GraphNode node;
	for(int i = nodeNameList.size()-1; i >= 0; i--) {

	    node = graph.getNode((String)nodeNameList.elementAt(i));
	    
	    // update depth field
	    nodeMaxDepthList.remove(i);

	    if(node != null) {
		nodeMaxDepthList.add(i, new Integer(node.getDepth(SOURCE)));
	    } else {
		System.out.println("Name-- "+(String)nodeNameList.elementAt(i)+
			" not found in graph");
	    }
	}
	    

	System.out.println("PathSet Max Depth: "+maxDepth);
	return maxDepth;
    }

    /** 
     * Find the depth of a node in the graph
     *
     * @param nodeName The node to look up the max depth of
     * @return The depth of the nodes requested, else -1.
     */
    public int getDepth(String nodeName) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		return ((Integer)nodeMaxDepthList.elementAt(i)).intValue();
	    }
	}
	return -1;
    }

    /**
     * Originally meant for backwards compatibility, its convenient to figure
     * out the names of all the paths that were oriented in a certain direction.
     * Here we hard code that 'false' values for orientation are used.
     *
     * Then since all paths are stored source to destination, we excise 
     *    the last element
     * 
     * @return Vector of all the destination node names in the graph.
     */
    public Vector getAllDestinations() {

	Vector destinations = new Vector();

	// find all the foward oriented paths and grab first node
	for(int i = pathList.size()-1; i >= 0; i--) {

	    Vector tmpVect = (Vector)pathList.elementAt(i);
	    if(tmpVect.size() > 0) {
		destinations.addElement(tmpVect.elementAt(tmpVect.size()-1));
	    }
	}


	// weed out duplicates
	PlotPathsUtil.removeDuplicatesFromVect(destinations);

	return destinations;
    }

    /**
     * Originally meant for backwards compatibility, its convenient to figure
     * out the names of all the paths that were oriented in a certain direction.
     * Here we hard code that 'true' values for orientation are used.
     *
     * Then since all paths are stored source to destination, we excise 
     *    the first element
     *
     * @return Vector of all the source node names in the graph.
     */
    public Vector getAllSources() {

	Vector sources = new Vector();

	// find all the foward oriented paths and grab first node
	for(int i = pathList.size()-1; i >= 0; i--) {

	    Vector tmpVect = (Vector)pathList.elementAt(i);
	    if(tmpVect.size() > 0) {
		sources.addElement(tmpVect.elementAt(0));
	    }
	}


	// weed out duplicates
	PlotPathsUtil.removeDuplicatesFromVect(sources);

	return sources;
    }


    /** 
     * Get all nodes that are at the top of the graph as indicated
     *   by that path's orientation.
     *
     * @return Vector of all the nodes appearing at the top of the graph.
     *     (involves looking at foward and backward paths)
     */
    public Vector getAllTop() {
	Vector sources = new Vector();

	// find all the foward oriented paths and grab first node
	for(int i = pathList.size()-1; i >= 0; i--) {

	    if(((Boolean)pathOrientList.elementAt(i)).booleanValue()) {
		Vector tmpVect = (Vector)pathList.elementAt(i);
		if(tmpVect.size() > 0) {
		    sources.addElement(tmpVect.elementAt(0));
		}
	    } else {
		Vector tmpVect = (Vector)pathList.elementAt(i);
		if(tmpVect.size() > 0) {
		    sources.addElement(tmpVect.elementAt(tmpVect.size()-1));
		}
	    }
	}


	// weed out duplicates
	PlotPathsUtil.removeDuplicatesFromVect(sources);
	
	return sources;
    }
    
    /** 
     * Get all nodes that are at the bottom of the graph as indicated
     *   by that path's orientation.
     *
     * @return Vector of all the nodes appearing at the bottom of the graph.
     *     (involves looking at foward and backward paths)
     */
    public Vector getAllBottom() {
	Vector sources = new Vector();

	// find all the foward oriented paths and grab first node
	for(int i = pathList.size()-1; i >= 0; i--) {

	    if(!((Boolean)pathOrientList.elementAt(i)).booleanValue()) {
		Vector tmpVect = (Vector)pathList.elementAt(i);
		if(tmpVect.size() > 0) {
		    sources.addElement(tmpVect.elementAt(0));
		}
	    } else {
		Vector tmpVect = (Vector)pathList.elementAt(i);
		if(tmpVect.size() > 0) {
		    sources.addElement(tmpVect.elementAt(tmpVect.size()-1));
		}
	    }
	}


	// weed out duplicates
	PlotPathsUtil.removeDuplicatesFromVect(sources);

	return sources;
    }
    
    /**
     * look up all paths the link is in, sum path weights (according to source)
     *
     * @param fromNodeName The "from" node in the pairwise link being looked at.
     * @param toNodeName The "to" node in the pairwise link being looked at.
     * @param source The name of the source node in the path.  This would be the
     *    first node in the path we want to look at.
     * @param destination The name of the destination node in the path.  This 
     *    would be the last node in the path we want to look at.
     * @return The weight of a link between the two given node names for a
     *    specific source and destination pair.
     */
    public float getPairLinkWeight(String fromNodeName, String toNodeName, 
			    String source, String destination) { 

	float resultWeight = 0;
	float totalPathCount = 0;

	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(((String)nodeNameList.elementAt(i)).equals(fromNodeName)) {
		Vector coordList = (Vector)nodeCoordList.elementAt(i);
		int pathIndex;
		int elemIndex;
		for(int j = coordList.size()-1; j >= 0; j--) {
		    pathIndex = ((Coordinate)coordList.elementAt(j)).pathIndex;
		    elemIndex = ((Coordinate)coordList.elementAt(j)).elemIndex;
		    if(((Vector)pathList.elementAt(pathIndex)).elementAt(
			    0).equals(source) &&

			    ((Vector)pathList.elementAt(pathIndex)).elementAt(
				((Vector)pathList.elementAt(pathIndex)).size()-1
				).equals(destination) &&

			    elemIndex + 1 < ((Vector)pathList.elementAt(
				pathIndex)).size() &&

			    toNodeName.equals(
				(String)((Vector)pathList.elementAt(
				pathIndex)).elementAt(elemIndex+1)
			    )) {

			resultWeight += ((Float)pathWeightList.elementAt(
					    pathIndex)).floatValue();
			totalPathCount++;
		    }
		}
	    }
	}
	return resultWeight;
    }

    /**
     * Get inbound and outbound in a single measurement
     *
     * @param fromNodeName The "from" node in the link we want to look at.
     * @param toNodeName The "to" node in the link we want to look at.
     * @param source_dest The name of the node that we're treating as our source
     *     and destination.  We're just looking for all paths involving this
     *     node as either of the two.
     * @return The total usage inbound and outbound of a link.
     */
    public float getSourceAndDestLinkWeight(String fromNodeName, 
			String toNodeName, String source_dest) {

	float inboundResultWeight = 0;
	float outboundResultWeight = 0;
	float inboundPathCount = 0;
	float outboundPathCount = 0;

	Vector oppositeSource = new Vector();
	Vector oppositeDest = new Vector();

	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(((String)nodeNameList.elementAt(i)).equals(fromNodeName)) {
		Vector coordList = (Vector)nodeCoordList.elementAt(i);
		int pathIndex;
		int elemIndex;
		for(int j = coordList.size()-1; j >= 0; j--) {
		    pathIndex = ((Coordinate)coordList.elementAt(j)).pathIndex;
		    elemIndex = ((Coordinate)coordList.elementAt(j)).elemIndex;
		    if((((Vector)pathList.elementAt(pathIndex)).elementAt(
				((Vector)pathList.elementAt(pathIndex)).size()-1
				).equals(source_dest)) &&

			    elemIndex + 1 < ((Vector)pathList.elementAt(
				pathIndex)).size() &&

			    toNodeName.equals((String)
				((Vector)pathList.elementAt(
				pathIndex)).elementAt(elemIndex+1)
			    )) {

			inboundResultWeight += ((Float)pathWeightList.elementAt(
				pathIndex)).floatValue();
			inboundPathCount++;

			String opposite = 
				(String)((Vector)pathList.elementAt(
				pathIndex)).elementAt( 0);
				
			if(!vectContainsString(oppositeSource, opposite)) {
			    oppositeSource.addElement(opposite);
			}
		    }
		    if((((Vector)pathList.elementAt(pathIndex)).elementAt(
				0).equals(source_dest) ) &&

			    elemIndex + 1 < ((Vector)pathList.elementAt(
				pathIndex)).size() &&

			    toNodeName.equals((String)((Vector)pathList.
				elementAt(pathIndex)).elementAt(elemIndex+1)
			    )) {

			outboundResultWeight += ((Float)pathWeightList.
				elementAt(pathIndex)).floatValue();
			outboundPathCount++;

			String opposite = 
				(String)((Vector)pathList.elementAt(
				pathIndex)).elementAt(
				((Vector)pathList.elementAt(
				pathIndex)).size()-1);
				
			if(!vectContainsString(oppositeDest, opposite)) {
			    oppositeDest.addElement(opposite);
			}
		    }
		}
	    }
	}
	if(inboundPathCount != 0 && outboundPathCount != 0) {
	    /*
	    System.out.println(
		    "Warning, found link used in inbound and outbound path");
	    System.out.println(
		    "   link: "+fromNodeName+" -> "+toNodeName);
	    */
	}

	if(inboundPathCount != 0) {
	    inboundResultWeight = inboundResultWeight / oppositeSource.size();
	}
	if(outboundPathCount != 0) {
	    outboundResultWeight = outboundResultWeight / oppositeDest.size();
	}

	//System.out.println("--------------------");
	//System.out.println("Inbound: "+inboundResultWeight);
	//System.out.println("Outpdfs: "+outboundResultWeight);

	if(inboundResultWeight > outboundResultWeight) {
	    //System.out.println("Result: "+inboundResultWeight);
	    return (inboundResultWeight);
	} else {
	    //System.out.println("Result: "+outboundResultWeight);
	    return (outboundResultWeight);
	}
    }

    /**
     * look up all paths the link is in, sum path weights (according to 
     *   destination)
     *
     * @param fromNodeName The "from" node in the link we want to look at.
     * @param toNodeName The "to" node in the link we want to look at.
     * @param destination The destination of the paths we want to look at.
     * @return The weight of the selected link over all paths going to 
     *     that destination.
     */
    public float getDestLinkWeight(String fromNodeName, String toNodeName, 
			    String destination) { 

	float resultWeight = 0;
	float totalPathCount = 0;

	Vector oppositeSource = new Vector();

	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(((String)nodeNameList.elementAt(i)).equals(fromNodeName)) {
		Vector coordList = (Vector)nodeCoordList.elementAt(i);
		int pathIndex;
		int elemIndex;
		for(int j = coordList.size()-1; j >= 0; j--) {
		    pathIndex = ((Coordinate)coordList.elementAt(j)).pathIndex;
		    elemIndex = ((Coordinate)coordList.elementAt(j)).elemIndex;
		    if(((Vector)pathList.elementAt(pathIndex)).elementAt(
				((Vector)pathList.elementAt(pathIndex)).size()-1
				).equals(destination) &&

			    elemIndex + 1 < ((Vector)pathList.elementAt(
				pathIndex)).size() &&

			    toNodeName.equals((String) 
				((Vector)pathList.elementAt(pathIndex)).
				elementAt(elemIndex+1)
			    )) {

			resultWeight += ((Float)pathWeightList.elementAt(
				pathIndex)).floatValue();
			totalPathCount++;

			String opposite = 
				(String)((Vector)pathList.elementAt(
				pathIndex)).elementAt(0);

			if(!vectContainsString(oppositeSource, opposite)) {
			    oppositeSource.addElement(opposite);
			}
		    }
		}
	    }
	}
	int totalDestPairs = countTotalDestPairs(destination);

	return (resultWeight / totalDestPairs);
    }

    /**
     * look up all paths the link is in, sum path weights (according to source)
     *
     * @param fromNodeName The "from" node in the selected link
     * @param toNodeName The "to" node in the selected link
     * @param source The source that we want to select all paths involving.
     * @return The weight of the selected link over all paths involving the
     *   specified source.
     */
    public float getSourceLinkWeight(String fromNodeName, String toNodeName, 
				String source) { 

	float resultWeight = 0;
	float totalPathCount = 0;

	Vector oppositeDest = new Vector();

	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(((String)nodeNameList.elementAt(i)).equals(fromNodeName)) {
		Vector coordList = (Vector)nodeCoordList.elementAt(i);
		int pathIndex;
		int elemIndex;
		for(int j = coordList.size()-1; j >= 0; j--) {
		    pathIndex = ((Coordinate)coordList.elementAt(j)).pathIndex;
		    elemIndex = ((Coordinate)coordList.elementAt(j)).elemIndex;
		    if(((Vector)pathList.elementAt(pathIndex)).elementAt(
				0).equals(source) &&

			    elemIndex + 1 < ((Vector)pathList.elementAt(
				pathIndex)).size() &&

			    toNodeName.equals((String) 
				((Vector)pathList.elementAt(pathIndex)).
				elementAt(elemIndex+1)
			    )) {
			resultWeight += ((Float)pathWeightList.elementAt(
				pathIndex)).floatValue();
			totalPathCount++;

			String opposite = 
				(String)((Vector)pathList.elementAt(
				pathIndex)).elementAt(
				((Vector)pathList.elementAt(
				pathIndex)).size()-1);

			if(!vectContainsString(oppositeDest, opposite)) {
			    oppositeDest.addElement(opposite);
			}
		    }
		}
	    }
	}
	int totalSourcePairs = countTotalSourcePairs(source);

	return (resultWeight / totalSourcePairs);
    }


    /**
     * Count the total number of source-destination pairs that
     *    include a specific source.
     *
     * @param source The source to compare to
     * @return The count of source-destination pairs.
     */
    private int countTotalSourcePairs(String source) {
	Vector thisPath;
	Vector destList = new Vector();
	String tmpDest;
	for(int i=0; i < pathList.size(); i++) {
	    thisPath = (Vector)pathList.elementAt(i);
	    if(thisPath.size() > 0) {
		if(source.equals((String)thisPath.elementAt(0))) {
		    tmpDest = (String)thisPath.elementAt(thisPath.size()-1);
		    if(!vectContainsString(destList, tmpDest)) {
			destList.addElement(tmpDest);
		    }
		}
	    }
	}
	return destList.size();
    }

    /**
     * Count the total number of source-destination pairs that
     *    include a specific destination.
     *
     * @param destination The destination to compare to
     * @return The count of source-destination pairs.
     */
    private int countTotalDestPairs(String destination) {
	Vector thisPath;
	Vector srcList = new Vector();
	String tmpSrc;
	for(int i=0; i < pathList.size(); i++) {
	    thisPath = (Vector)pathList.elementAt(i);
	    if(thisPath.size() > 0) {
		if(destination.equals((String)thisPath.elementAt(
				thisPath.size()-1))) {
		    tmpSrc = (String)thisPath.elementAt(0);
		    if(!vectContainsString(srcList, tmpSrc)) {
			srcList.addElement(tmpSrc);
		    }
		}
	    }
	}
	return srcList.size();
    }

    /**
     * look up all paths the node is in, sum path weights
     *
     * @param nodeName The node we're looking up
     * @param source The source we're interested in that will supply the paths 
     *     we extract data from.
     * @return The weight of the node in all paths involving the specified 
     *     source.
     */
    public float getNodeWeight(String nodeName, String source) { 

	float resultWeight = 0;
	float totalPathCount = 0;

	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(((String)nodeNameList.elementAt(i)).equals(nodeName)) {
		Vector coordList = (Vector)nodeCoordList.elementAt(i);
		int pathIndex;
		for(int j = coordList.size()-1; j >= 0; j--) {
		    pathIndex = ((Coordinate)coordList.elementAt(j)).pathIndex;
		    if(((Vector)pathList.elementAt(pathIndex)).elementAt(
			    0).equals(source)) {
			resultWeight += ((Float)pathWeightList.elementAt(
			    pathIndex)).floatValue();
			totalPathCount++;
		    }
		}
	    }
	}
	return resultWeight;
    }

    /**
     * look up node in list, return non-redundant list of sources
     * that it appears in.
     *
     * @param fromNodeName The "from" node of the specified link
     * @param toNodeName The "to" node of the specified link
     * @return A Vector of all the destinations of the paths 
     *     that a link is involved in.
     */
    public Vector getLinkDestinations(String fromNodeName, String toNodeName) {
	Vector resultList = new Vector();

	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(((String)nodeNameList.elementAt(i)).equals(fromNodeName)) {
		Vector coordList = (Vector)nodeCoordList.elementAt(i);
		Vector currentPath;
		int pathIndex;
		int elemIndex;
		for(int j = coordList.size()-1; j >= 0; j--) {
		    pathIndex = ((Coordinate)coordList.elementAt(j)).pathIndex;
		    elemIndex = ((Coordinate)coordList.elementAt(j)).elemIndex;
		    currentPath = (Vector)pathList.elementAt(pathIndex);
		    if(elemIndex + 1 < currentPath.size() && 
			    toNodeName.equals((String)currentPath.elementAt(
			    elemIndex+1))) {
			resultList.addElement(currentPath.elementAt(
			    currentPath.size()-1));
		    }
		}
	    }
	}

	// weed out duplicates
	PlotPathsUtil.removeDuplicatesFromVect(resultList);

	return resultList;
    }
    public Vector getLinkSources(String fromNodeName, String toNodeName) {
	Vector resultList = new Vector();

	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(((String)nodeNameList.elementAt(i)).equals(fromNodeName)) {
		Vector coordList = (Vector)nodeCoordList.elementAt(i);
		Vector currentPath;
		int pathIndex;
		int elemIndex;
		for(int j = coordList.size()-1; j >= 0; j--) {
		    pathIndex = ((Coordinate)coordList.elementAt(j)).pathIndex;
		    elemIndex = ((Coordinate)coordList.elementAt(j)).elemIndex;
		    currentPath = (Vector)pathList.elementAt(pathIndex);
		    if(elemIndex + 1 < currentPath.size() && 
			    toNodeName.equals((String)currentPath.elementAt(
			    elemIndex+1))) {
			resultList.addElement(currentPath.elementAt(0));
		    }
		}
	    }
	}

	// weed out duplicates
	PlotPathsUtil.removeDuplicatesFromVect(resultList);

	return resultList;
    }
    /**
     * look up node in list, return non-redundant list of sources
     * that it appears in.
     *
     * @param nodeName The name of the specified node
     * @return A Vector of all the sources of paths that the specified node 
     *     is found in.
     */
    public Vector getNodeSources(String nodeName) {
	Vector resultList = new Vector();

	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(((String)nodeNameList.elementAt(i)).equals(nodeName)) {
		Vector coordList = (Vector)nodeCoordList.elementAt(i);
		Vector currentPath;
		int pathIndex;
		for(int j = coordList.size()-1; j >= 0; j--) {
		    pathIndex = ((Coordinate)coordList.elementAt(j)).pathIndex;
		    currentPath = (Vector)pathList.elementAt(pathIndex);
		    resultList.addElement(currentPath.elementAt(0));
		}
	    }
	}

	// weed out duplicates
	PlotPathsUtil.removeDuplicatesFromVect(resultList);

	return resultList;
    }

    /** 
     * Count all the times that a node is referenced in any path and 
     *     return that count
     *
     * @param nodeName The node to find in the path list
     * @return The number of different paths the specified node appears in.
     */
    public int countPathsContainingNode(String nodeName) {

	boolean alreadyInList;
	Vector pathCounter = new Vector(); // used to count unique paths
	String source, destination;
	String pathName;

	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(((String)nodeNameList.elementAt(i)).equals(nodeName)) {
		Vector coordList = (Vector)nodeCoordList.elementAt(i);
		Vector currentPath;
		int pathIndex;
		for(int j = coordList.size()-1; j >= 0; j--) {
		    pathIndex = ((Coordinate)coordList.elementAt(j)).pathIndex;
		    currentPath = (Vector)pathList.elementAt(pathIndex);

		    source = (String)currentPath.elementAt(0);
		    destination = (String)currentPath.elementAt(
			    currentPath.size()-1);
		    pathName = (source+" - "+destination);

		    alreadyInList = false;
		    for(int k =0; k < pathCounter.size(); k++) {
			if(pathName.equals(pathCounter.elementAt(k))) {
			    alreadyInList = true;
			    break;
			}
		    }
		    if(alreadyInList == false) {
			pathCounter.addElement(pathName);
		    }
		}
	    }
	}
	return pathCounter.size();
    }

    /** 
     * Count all the times that a node is referenced in any path and 
     *    return that count.
     *    A single source-destination pair may count only once.
     *
     * @param nodeName1 The "from" node in the link.
     * @param nodeName2 The "to" node in the link.
     * @return The number of unique paths containing that link.
     */
    public int countUniquePathsContainingLink(String nodeName1, 
			String nodeName2) {

	boolean alreadyInList;
	Vector pathCounter = new Vector(); // used to count unique paths
	String source, destination;
	String pathName, destNodeName;

	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(((String)nodeNameList.elementAt(i)).equals(nodeName1)) {
		Vector coordList = (Vector)nodeCoordList.elementAt(i);
		Vector currentPath;
		int pathIndex;
		for(int j = coordList.size()-1; j >= 0; j--) {
		    pathIndex = ((Coordinate)coordList.elementAt(j)).pathIndex;
		    currentPath = (Vector)pathList.elementAt(pathIndex);

		    // test to see if this link connects to the 
		    //    destination node given
		    if(((Coordinate)coordList.elementAt(j)).elemIndex+1 < 
				currentPath.size()) {

			destNodeName = (String)currentPath.elementAt(
				((Coordinate)coordList.elementAt(j)).
				elemIndex+1);

			if(destNodeName.equals(nodeName2)) {
    
			    source = (String)currentPath.elementAt(0);
			    destination = (String)currentPath.elementAt(
				    currentPath.size()-1);
			    pathName = (source+" - "+destination);
	
			    alreadyInList = false;
			    for(int k =0; k < pathCounter.size(); k++) {
				if(pathName.equals(pathCounter.elementAt(k))) {
				    alreadyInList = true;
				    break;
				}
			    }
			    if(alreadyInList == false) {
				pathCounter.addElement(pathName);
			    }
			}
		    }
		}
	    }
	}
	return pathCounter.size();
    }

    /** 
     * count all the times that a node is referenced in any path and 
     *     return that count.
     *
     * @param nodeName1 The "from" node in the specified link.
     * @param nodeName2 The "to" node in the specified link.
     * @return The total count of paths containing the specified link.
     */
    public int countTotalPathsContainingLink(String nodeName1, 
					    String nodeName2) {

	int pathCounter = 0;
	String destNodeName;

	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(((String)nodeNameList.elementAt(i)).equals(nodeName1)) {
		Vector coordList = (Vector)nodeCoordList.elementAt(i);
		Vector currentPath;
		int pathIndex;
		for(int j = coordList.size()-1; j >= 0; j--) {
		    pathIndex = ((Coordinate)coordList.elementAt(j)).pathIndex;
		    currentPath = (Vector)pathList.elementAt(pathIndex);

		    // test to see if this link connects to the 
		    //    destination node given
		    if( ((Coordinate)coordList.elementAt(j)).elemIndex+1 < 
			    currentPath.size()) {

			destNodeName = (String)currentPath.elementAt(
			    ((Coordinate)coordList.elementAt(j)).elemIndex+1);

			if(destNodeName.equals(nodeName2)) {
			    pathCounter++;
			}
		    }
		}
	    }
	}
	return pathCounter;
    }

    /**
     * Return an array that is a complete list of all nodes in the graph
     * without duplicates.
     *
     * @return An array of node names consisting of all the nodes in the graph.
     */
    public String[] getAllNodes() {
	String[] allNodeNames = new String[nodeNameList.size()];
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    allNodeNames[i] = (String)nodeNameList.elementAt(i);
	}
	return allNodeNames;
    }

    /** 
     * Store Latitude associated with a node
     *
     * @param nodeName The node to set the latitude for.
     * @param lat The latitude to set that node to.
     */
    public void setLat(String nodeName, float lat) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		nodeLatList.removeElementAt(i);
		nodeLatList.insertElementAt(new Float(lat), i);
		break;
	    }
	}
    }

    /** 
     * Store Longitude associated with a node
     *
     * @param nodeName The node to set the longitude for.
     * @param longitude The longitude to set that node to.
     */
    public void setLong(String nodeName, float longitude) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		nodeLongList.removeElementAt(i);
		nodeLongList.insertElementAt(new Float(longitude), i);
		break;
	    }
	}
    }

    /** 
     * Store City associated with a node
     *
     * @param nodeName The node to set the city name for.
     * @param city_name The city name to assign to that node.
     */
    public void setCity(String nodeName, String city_name) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		nodeCityList.removeElementAt(i);
		nodeCityList.insertElementAt(city_name, i);
		break;
	    }
	}
    }

    /** 
     * Store AS name associated with a node
     *
     * @param nodeName The node to set the AS name for.
     * @param ASName The AS name to assign to that node.
     */
    public void setASName(String nodeName, String ASName) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		nodeASNameList.removeElementAt(i);
		nodeASNameList.insertElementAt(ASName, i);
		break;
	    }
	}
    }

    /** 
     * Store AS Number associated with a node
     *
     * @param nodeName The node to set the AS for.
     * @param as_num The AS to assign to that node.
     */
    public void setASNumber(String nodeName, int as_num) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		nodeASNumberList.removeElementAt(i);
		nodeASNumberList.insertElementAt(new Integer(as_num), i);
		break;
	    }
	}
    }


    /////////////////////////////////////////////////////////////////////////////
    // Added by John Tusson Tue Aug 14 14:16:12 PDT 2001                       //
    /////////////////////////////////////////////////////////////////////////////
    /** 
     * Store Attribute name that will be used to sort columns
     *
     * @param nodeName The node to set the Attribute name for.
     * @param AttributeName The Attribute name to assign to that node.
     */
    public void setAttributeName(String nodeName, String AttributeName) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		nodeAttributeNameList.removeElementAt(i);
		nodeAttributeNameList.insertElementAt(AttributeName, i);
		break;
	    }
	}
    }

    /////////////////////////////////////////////////////////////////////////////
    // Added by John Tusson Tue Aug 14 14:24:23 PDT 2001                       //
    /////////////////////////////////////////////////////////////////////////////
    /** 
     * retrieve Attribute Name data 
     *
     * @param nodeName The name of the node to retrieve the Attribute name for
     * @return String of the Attribute name, else null
     */
    public String getAttributeName(String nodeName) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		return (String)nodeAttributeNameList.elementAt(i);
	    }
	}
	return null;
    }


    /** 
     * retrieve lat data
     *
     * @param nodeName The name of the node to retrieve latitude info for.
     * @return the Latitude of the specified node, else 0.
     */
    public float getLat(String nodeName) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		return ((Float)nodeLatList.elementAt(i)).floatValue();
	    }
	}
	return 0;
    }

    /** 
     * retrieve long data
     *
     * @param nodeName The name of the node to retrieve longitude info for.
     * @return the Longitude of the specified node, else 0.
     */
    public float getLong(String nodeName) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		return ((Float)nodeLongList.elementAt(i)).floatValue();
	    }
	}
	return 0;
    }

    /** 
     * retrieve AS data
     *
     * @param nodeName The name of the node to retrieve the AS num for.
     * @return The AS number for the specified node, else -1.
     */
    public int getASNumber(String nodeName) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		return ((Integer)nodeASNumberList.elementAt(i)).intValue();
	    }
	}
	return -1;
    }

    /** 
     * retrieve city data
     *
     * @param nodeName The name of the node to retrieve the city name for.
     * @return String of the city name, else null
     */
    public String getCity(String nodeName) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		return (String)nodeCityList.elementAt(i);
	    }
	}
	return null;
    }

    /** 
     * retrieve AS Name data 
     *
     * @param nodeName The name of the node to retrieve the city name for.
     * @return String of the AS name, else null
     */
    public String getASName(String nodeName) {
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		return (String)nodeASNameList.elementAt(i);
	    }
	}
	return null;
    }

    /**
     * return the children of a node from a specific source
     *
     * @param nodeName The node to look for children for.
     * @param source The source of paths we will look for children in.
     * @return A Vector of the children of a node given a specified source.
     */
    public Vector getChildren(String nodeName, String source) {

	Vector children = new Vector();
	Vector coordinateList = null; 
	Coordinate coordinate = null;
	Vector path = null;

	// look up Vector of coordinates
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		coordinateList = (Vector)nodeCoordList.elementAt(i);
	    }
	}
	

	// drop out if we don't find the node in the graph
	if(coordinateList == null) {
	    return children; 	// empty Vector
	}


	// build list of children as we go
	for(int i = coordinateList.size()-1; i >= 0; i--) {
	    coordinate = (Coordinate)coordinateList.elementAt(i);

	    // if it isn't the last element, add the following element to the
	    //     list of children (we'll weed the list later).
	    path = (Vector)pathList.elementAt(coordinate.pathIndex);
	    if(coordinate.elemIndex < path.size()-1 && 
			source.equals((String)path.elementAt(0))) {

		// add the child
		// ps. its a string
		children.addElement(path.elementAt(coordinate.elemIndex + 1)); 

	    }
	}

	// weed out duplicates
	PlotPathsUtil.removeDuplicatesFromVect(children);

	return children;
    }


    /**
     * return the children of a node from across all paths
     *
     * @param nodeName The node we want to retrieve children for.
     * @return Vector of all children for that node across all sources.
     */
    public Vector getChildren(String nodeName) {

	Vector children = new Vector();
	Vector coordinateList = null; 
	Coordinate coordinate = null;
	Vector path = null;

	// look up Vector of coordinates
	for(int i = nodeNameList.size()-1; i >= 0; i--) {
	    if(nodeName.equals((String)nodeNameList.elementAt(i))) {
		coordinateList = (Vector)nodeCoordList.elementAt(i);
	    }
	}
	

	// drop out if we don't find the node in the graph
	if(coordinateList == null) {
	    return children; 	// empty Vector
	}


	// build list of children as we go
	for(int i = coordinateList.size()-1; i >= 0; i--) {
	    coordinate = (Coordinate)coordinateList.elementAt(i);

	    // if it isn't the last element, add the following element to the
	    //     list of children (we'll weed the list later).
	    path = (Vector)pathList.elementAt(coordinate.pathIndex);
	    if(coordinate.elemIndex < path.size()-1) {

		// add the child
		// ps. its a string
		children.addElement(path.elementAt(coordinate.elemIndex + 1)); 

	    }
	}

	// weed out duplicates
	PlotPathsUtil.removeDuplicatesFromVect(children);

	return children;
    }

    /**
     * Test if a vector of Strings contains a particlar string
     */
    private boolean vectContainsString(Vector vect, String str) {
	for(int i=0; i < vect.size(); i++) {
	    if(str.equals((String)vect.elementAt(i))) {
		return true;
	    }
	}
	return false;
    }

    
    /** 
     * Return the max depth of the graph.
     *   calcMaxDepth() must be called for this to give
     *   meaningful results.
     *
     * @return The max depth of the last run of calcMaxDepth()
     */
    public int getMaxDepth() {
	return maxDepth;
    }

    /**
     * Get the neighbors of a node (their string names that is)
     *
     * @param nodeName node to find neighbors of
     * @return vector of node names (strings)
     */
    public Vector getNodeNeighbors(String nodeName) {
	Vector resultList = new Vector();

	for(int i = 0; i < nodeNameList.size(); ++i) {
	    if(((String)nodeNameList.elementAt(i)).equals(nodeName)) {
		Vector coordList = (Vector)nodeCoordList.elementAt(i);
		Vector currentPath;
		int pathIndex;
		int elemIndex;
		for(int j = 0; j < coordList.size(); ++j) {
		    pathIndex = ((Coordinate)coordList.elementAt(j)).pathIndex;
		    elemIndex = ((Coordinate)coordList.elementAt(j)).elemIndex;
		    currentPath = (Vector)pathList.elementAt(pathIndex);
		    if(elemIndex - 1 >= 0) {
			resultList.addElement(
				currentPath.elementAt(elemIndex-1));
		    }
		    if(elemIndex + 1 < currentPath.size()) {
			resultList.addElement(
				currentPath.elementAt(elemIndex+1));
		    }
		}
	    }
	}

	// weed out duplicates
	PlotPathsUtil.removeDuplicatesFromVect(resultList);

	return resultList;
    }

    // Each element is a Vector that contains a path which
    //      is a list of node names.  Path indexes are parallel
    //      with the indexes of weightList and orientationList.
    private Vector pathList = new Vector();

    // List of path weights that is parallel to pathList
    private Vector pathWeightList = new Vector();

    // List of path orientations that is parallel to pathList
    private Vector pathOrientList = new Vector();




    // Each element is a node name, which must be unique 
    // Index into the Vector corresponds to nodeCoordList
    private Vector nodeNameList   = new Vector();

    // Each element is a list of pointers to every occurence of the node 
    //      in the pathList. 
    private Vector nodeCoordList = new Vector();

    // Each element is a Integer that corresponds in index to the node
    //      listed in nodeNameList.
    private Vector nodeMaxDepthList = new Vector();

    // Each element is a Float whose index corresponds to the index of 
    //      the node in nodeNameList
    private Vector nodeLatList = new Vector();

    // Each element is a Float whose index corresponds to the index of 
    //      the node in nodeNameList
    private Vector nodeLongList = new Vector();

    // Each element is a String whose index corresponds to the index of 
    //      the node in nodeNameList
    private Vector nodeCityList = new Vector();
    
    // Each element is an Integer whose index corresponds to the index of 
    //      the node in nodeNameList
    private Vector nodeASNumberList = new Vector();

    // Each element is an Integer whose index corresponds to the index of 
    //      the node in nodeNameList
    private Vector nodeASNameList = new Vector();

    // Each element is an Integer whose index corresponds to the index of 
    //      the node in nodeNameList(Added by John Tusson Tue Aug 14 14:19:41 PDT 2001)
    private Vector nodeAttributeNameList = new Vector();

    /**
     * Save coordinate pairs to specify location of a node in the list
     * of lists of paths
     */
    private class Coordinate {
	public Coordinate(int pathIndex, int elemIndex) {
	    this.pathIndex = pathIndex;
	    this.elemIndex = elemIndex;
	}
	public int pathIndex;
	public int elemIndex;
    }

    // we try to support multiple sources, but for now just use one.
    private static String SOURCE = "main source";
    private int maxDepth;
    private String unknownNodeName;
}
