//==========================================================================
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// Permission to use, copy, modify and distribute any part of this Makefile
// for educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above 
// copyright notice, this paragraph and the following paragraphs appear 
// in all copies.
//
// Those desiring to incorporate this into commercial products or use for
// commercial purposes should contact the
//
//      Technology Transfer Office, University of California, San Diego,
//      9500 Gilman Drive, La Jolla, CA 92093-0910
//      Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS PLOTPATH, EVEN 
// IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
//
// THE PLOTPATH, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF 
// CALIFORNIA PLOTPATH NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF 
// ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
// PURPOSE, OR THAT THE USE OF THE SKITTERTRACES WILL NOT INFRINGE ANY 
// PATENT, TRADEMARK OR OTHER RIGHTS.
//
// PlotPaths is developed by John Gallagher Bradley L. Huffaker at the
// University of California, San Diego under the Cooperative Association
// for Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by NSF "Internet Atlas" grant ANI-I-9996248, NSF "CAIDA Core"
// NCR-9711092, and by CAIDA memebers.
//==========================================================================
package org.caida.otter;

import java.util.Vector;
import java.util.StringTokenizer;
import org.caida.libsea.*;

public class PlotPaths {
    
    public static void main(String[] args)
    {
	// handles the output into otter format
	// kept around for debugging purposes and
	// until LibSeaToOtter is completed.
	PlotPathsIO outputHandler  = new IOotter();
	
	// Process the command line arguments
	GetOpts CmdOpts = new GetOpts(args);

	// Check to see that the command line argument processing
	// completed succesfully.
	if(!CmdOpts.get_Status())
	    {
		System.exit(1);
	    }

	System.out.println("========== Loading " + CmdOpts.get_INPUT_FILE() + " =============");

	GraphIO graphIO = new GraphIO();
	Graph LibSea_Graph = graphIO.loadGraph(CmdOpts.get_INPUT_FILE());

	//Problem loading file
	if(LibSea_Graph == null)
	    {
		System.exit(1);
	    }

	System.out.println("========== Ordering Graph =============");

	//Print out attribute info for the libsea graph file.
	if(CmdOpts.get_INFO_FLAG())
	    {
		GraphInfo GInfo = new GraphInfo();
		if(CmdOpts.get_LIST_ATTRIBUTES())
		    {
			GInfo.printAttributeInfo(LibSea_Graph);
		    }
		if(CmdOpts.get_LIST_SOURCES())
		    {
			GInfo.printAvailableSources(LibSea_Graph);
		    }
		System.exit(0);
	    }

	LibSeaToPlotPaths libToPlotPaths = new LibSeaToPlotPaths();

	PathSet graph = libToPlotPaths.loadPaths(LibSea_Graph, CmdOpts);

	graph = libToPlotPaths.setColumnAttribute(LibSea_Graph, graph,
						 CmdOpts.get_COLUMN_ATTRIBUTE());

	// calculate max depth 
	graph.calcMaxDepth();

	// our container for the objects that will be drawn
	DisplayObjects displayData;


	displayData = buildDisplayObjects2(graph, CmdOpts);

	displayData.addColumnNameToNodeName();

	///////////////////////////////////////////
	// ** Use ColumnSet to organize nodes ** //
	///////////////////////////////////////////


	// ---- create Column object ----
	ColumnSet colSet = new ColumnSet(CmdOpts.get_COLUMN_WIDTH(), 
					 CmdOpts.get_ROW_SPACING(), displayData);


	// ----- sort away not found AS numbers ------
	System.out.println("Placing nodes without an AS...");
	colSet.disperseNotFoundColumn(CmdOpts.get_REMOVE_COL_NAMES());

	// ------ decide on the column ordering ----------
	System.out.println("Sorting columns by link overlap...");
	colSet.orderColumnsByBoundingBoxes();

	// ---- space out columns by the node count in biggest row -----
	System.out.println("Spacing columns by Node count");
	colSet.spaceColumnsByNodeCount(CmdOpts.get_NODE_SPACING(), 
				       CmdOpts.get_COLUMN_WIDTH());

	// -- add hidden nodes to try and avoid links overlapping with nodes --
	System.out.println("Adding hidden nodes");
	colSet.insertHiddenNodes();

	// ------- spread nodes apart that overlap --------
	System.out.println("Spreading overlapping nodes by parent mean");
	colSet.spaceNodesByParentMean(CmdOpts.get_NODE_SPACING(), 
				      CmdOpts.get_MAX_CHILD_SEARCH_DEPTH());

	// -------- get the display data  -----------
	System.out.println("Preparing to print output");
	displayData = colSet.getDisplayObjects();



	////////////////////////////////////////////
	//  Do some final processing on our image //
	////////////////////////////////////////////

	// invert the y image 
	//displayData.invertImage(false, true);

	// stagger the labels so names don't overlap, are more readable
	displayData.staggerLabels(CmdOpts.get_LABEL_TIERS(), CmdOpts.get_LABEL_SPACING());

	// normalize the size of the image
	displayData.normalizeSize(CmdOpts.get_SCREEN_X(), CmdOpts.get_SCREEN_Y());

	/////////////////////////////////////////////////////////////////////////////////////////
	// Kept around for debugging purposes until I can get the perl scripts? to 
	// accomplish the same task.  John Tusson

	// -------- print otter format output ---------
	// Use the LibSeaToOtter conversion program instead.  At least for now.
	//System.out.println("Printing output to file \""+  CmdOpts.get_OTTER_OUTFILE() + "\"");
	//outputHandler.writeOutput(displayData, CmdOpts.get_OTTER_OUTFILE(), CmdOpts); 
	 
	// add all the elements to the libsea graph file that are
	// required for otter..
	libToPlotPaths.addLibSeaAttributes(LibSea_Graph, displayData);
	
	// write the LibSea graph object into a file.
	// graph.libsea is being used temporarily.
	// LIBSEA_OUTFILE will be used in the future
	System.out.println("Printing output to graph.libsea.");
	graphIO.writeGraph(CmdOpts.get_LIBSEA_OUTFILE(), LibSea_Graph, true);

	System.out.println("======= Processing Complete ==========");
    }

    private static DisplayObjects buildDisplayObjects2(PathSet graph, GetOpts CmdOpts) {

	DisplayObjects data = new DisplayObjects();
	String[] nodeList 	= graph.getAllNodes();
	Vector nodeNeighbors;
	Vector children;
	String neighborName;
	String columnName;
	String neighborColumnName;
	String newNodeName;
	String lineName;
	Node tmpNode;
	Node tmpNodeNeighbor;
	Label tmpLabel;
	Line tmpLine;
	int rowNum;
	boolean isRoot;

	//-- these are all the statistics that we will offer --//
	data.addCategory("Pairwise Usage");
	data.addCategory("Outbound Usage");
	data.addCategory("Inbound Usage");
	data.addCategory("Total Usage");
	data.addCategory("Usage Count");

	Vector allSources 	= graph.getAllSources();
	Vector allDestinations 	= graph.getAllDestinations();


	for(int i=0; i < nodeList.length; ++i) {


	    // insert the node if it hasn't already been
	    // set tmpNode to the node nodeList[i] names

	    if(!data.containsNode(nodeList[i])) {

		columnName = graph.getAttributeName(nodeList[i]);
		rowNum = graph.getDepth(nodeList[i]);
		newNodeName = nodeList[i];

		tmpNode = new Node(i, -1, rowNum, newNodeName, columnName, 
				   new Integer(nodeList[i]).intValue());

		// is our new node a "root"
		isRoot = false;
		for(int z=0;z < allSources.size(); z++) {
		    if(newNodeName.equals(
			    (String)allSources.elementAt(z))) {
			isRoot = true;
			break;
		    }
		}
		if(isRoot == false) {
		    for(int z=0;z < allDestinations.size(); z++) {
			if(newNodeName.equals(
				(String)allDestinations.elementAt(z))) {
			    isRoot = true;
			    break;
			}
		    }
		}

		if(isRoot) {
		    tmpNode.setIsRoot(true);
		}

		data.addNode(tmpNode);

	    } else {
		tmpNode = data.getNode(nodeList[i]);
	    }

	    // make sure all the neighbors have been inserted that 
	    // already exist

	    children = graph.getChildren(nodeList[i]);
	    nodeNeighbors = graph.getNodeNeighbors(nodeList[i]);


	    for(int j=0; j < nodeNeighbors.size(); ++j) {
		neighborName = (String)nodeNeighbors.elementAt(j);
		if(data.containsNode(neighborName)) {
		    tmpNodeNeighbor = data.getNode(neighborName);

		    // add them to each other's neighbor list
		    tmpNode.addNeighborNode(tmpNodeNeighbor);
		    tmpNodeNeighbor.addNeighborNode(tmpNode);

		    // add the line if it isn't already there.
		    boolean isChild = false;
		    for(int k=0; k < children.size(); ++k) {
			if(((String)children.elementAt(k)).equals(
				neighborName)) {
			    isChild = true;
			    break;
			}
		    }
		    if(isChild) {
			lineName = tmpNode.getName() + " -> " +
			    tmpNodeNeighbor.getName();
			tmpLine = new Line(tmpNode, tmpNodeNeighbor, lineName);
			data.addLine(tmpLine);
			addLineAttributes(  graph, 
					    tmpNode,
					    tmpNodeNeighbor, 
					    tmpLine, 
					    data,
					    CmdOpts);
		    } else {
			lineName = tmpNodeNeighbor.getName() + " -> " + 
			    tmpNode.getName();
			tmpLine = new Line(tmpNodeNeighbor, tmpNode, lineName);
			data.addLine(tmpLine);
			addLineAttributes(  graph, 
					    tmpNodeNeighbor, 
					    tmpNode, 
					    tmpLine, 
					    data,
					    CmdOpts);
		    }
		}
	    }
	}
	return data;
    }

    private static void addLineAttributes(PathSet graph, Node fromNode, 
			Node toNode, Line line, DisplayObjects data, GetOpts CmdOpts) {

	String toNodeName = toNode.getName();
	String fromNodeName = fromNode.getName();

	// new code to add hash entries in link path info
	Vector linkSources = 
		graph.getLinkSources(fromNode.getName(), toNode.getName());
	Vector linkDestinations = 
		graph.getLinkDestinations(fromNode.getName(), toNode.getName());

	float pathWeight;

	// seperate source/destination/pairwise for readability.
	// we could try to everything in one pass, but this
	// isn't the bottleneck if we want to optimize.

	// insert outbound data
	for(int n=0; n < linkSources.size(); n++) {
	    pathWeight = graph.getSourceLinkWeight(
			    fromNodeName, 
			    toNodeName,
			    (String)linkSources.elementAt(n));

	    if(pathWeight > 0) {

		data.addAttribute(
			    "Outbound Usage",
			    linkSources.elementAt(n),
			    new Boolean(true));

		line.addAttribute(
			    "Outbound Usage",
			    linkSources.elementAt(n),
			    getPathWeightBinString(pathWeight, CmdOpts));
	    }
	}

	// insert inbound data
	for(int n=0; n < linkDestinations.size(); n++) {
	    pathWeight = graph.getDestLinkWeight(
			    fromNodeName, 
			    toNodeName,
			    (String)linkDestinations.elementAt(n));

	    if(pathWeight > 0) {

		data.addAttribute(
			    "Inbound Usage",
			    linkDestinations.elementAt(n),
			    new Boolean(true));

		line.addAttribute(
			    "Inbound Usage",
			    linkDestinations.elementAt(n),
			    getPathWeightBinString(pathWeight, CmdOpts));
	    }
	}

	// insert Inbound and Outbound (combined stats)
	for(int n=0; n < linkDestinations.size(); n++) {
	    pathWeight = graph.getSourceAndDestLinkWeight(
			    fromNodeName, 
			    toNodeName,
			    (String)linkDestinations.elementAt(n));
	    
	    if(pathWeight > 0) {

		data.addAttribute(
			    "Total Usage",
			    linkDestinations.elementAt(n),
			    new Boolean(true));

		line.addAttribute(
			    "Total Usage",
			    linkDestinations.elementAt(n),
			    getPathWeightBinString(pathWeight, CmdOpts));
	    }
	}
	for(int n=0; n < linkSources.size(); n++) {
	    pathWeight = graph.getSourceAndDestLinkWeight(
			    fromNodeName, 
			    toNodeName,
			    (String)linkSources.elementAt(n));

	    if(pathWeight > 0) {

		data.addAttribute(
			    "Total Usage",
			    linkSources.elementAt(n),
			    new Boolean(true));

		line.addAttribute(
			    "Total Usage",
			    linkSources.elementAt(n),
			    getPathWeightBinString(pathWeight, CmdOpts));
	    }
	}

	// insert pairwise data
	for(int n=0; n < linkSources.size(); n++) {
	    for(int m=0; m < linkDestinations.size(); m++) {

		pathWeight = graph.getPairLinkWeight(
			    fromNodeName, toNodeName, 
			    (String)linkSources.elementAt(n),
			    (String)linkDestinations.elementAt(m));


		if(pathWeight > 0) {

		    String attributeName = linkSources.elementAt(n)+ " -> "+
			    linkDestinations.elementAt(m);

		    data.addAttribute(
			    "Pairwise Usage", 
			    attributeName,
			    new Boolean(true));

		    line.addAttribute(
			    "Pairwise Usage",
			    attributeName,
			    getPathWeightBinString(pathWeight, CmdOpts));
		}
	    }
	}


	// insert number of unique times seen in a unique path
	data.addAttribute( "Usage Count", "Unique Paths", new Boolean(true));
	line.addAttribute( "Usage Count", "Unique Paths", 
	    ""+graph.countUniquePathsContainingLink(fromNodeName, toNodeName));


	// insert number of total times seen in a unique path
	data.addAttribute( "Usage Count", "Total Paths", new Boolean(true));
	line.addAttribute( "Usage Count", "Total Paths", 
	    ""+graph.countTotalPathsContainingLink(fromNodeName, toNodeName));
    }

    /**
     * Convert a float value for a path weight into a string name for the
     *   bin that best represents it.
     */
    private static String getPathWeightBinString(float pathWeight, GetOpts CmdOpts) 
    {
       String result = "";

	//
	// give ourselves a little leeway on rounding errors
	//

	if(pathWeight < 100.1 && pathWeight > 99.9) {
	    result =  "100%";
	} else if(pathWeight < 0.1 && pathWeight >= 0) {
	    result =  "not selected";
	} else if(pathWeight < 0) {
	    result = "Less than 0%";
	} else if(pathWeight > 100) {
	    result =  "More than 100%";
	} else {

	    int bucketSize = (int)((float)100 / (float)CmdOpts.get_PATH_WEIGHT_BIN_COUNT());

	    // use integer division
	    int lowerBound = (int)pathWeight / bucketSize;
	    int upperBound = lowerBound + 1; 

	    result += (lowerBound*bucketSize)+"%-";
	    result += (upperBound*bucketSize)+"%";
	}

	return result;
    }
}
