//==========================================================================
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// Permission to use, copy, modify and distribute any part of this Makefile
// for educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above 
// copyright notice, this paragraph and the following paragraphs appear 
// in all copies.
//
// Those desiring to incorporate this into commercial products or use for
// commercial purposes should contact the
//
//      Technology Transfer Office, University of California, San Diego,
//      9500 Gilman Drive, La Jolla, CA 92093-0910
//      Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS PLOTPATH, EVEN 
// IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
//
// THE PLOTPATH, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF 
// CALIFORNIA PLOTPATH NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF 
// ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
// PURPOSE, OR THAT THE USE OF THE SKITTERTRACES WILL NOT INFRINGE ANY 
// PATENT, TRADEMARK OR OTHER RIGHTS.
//
// PlotPaths is developed by John Gallagher Bradley L. Huffaker at the
// University of California, San Diego under the Cooperative Association
// for Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by NSF "Internet Atlas" grant ANI-I-9996248, NSF "CAIDA Core"
// NCR-9711092, and by CAIDA memebers.
//==========================================================================
package org.caida.otter;

import java.util.Vector;

public class Row {

    /**
     * Create an empty row
     */
    public Row() {
	rowYVal = 0;
    }

    /**
     * Create an empty row with this y val
     */
    public Row(int y_val) {
	rowYVal = y_val;
    }

    /**
     * Create a row and initialize it with this coordinate 
     *
     * @param newCoord The first coordinate to populate the row.
     */
    public Row(Coordinate newCoord) {
	rowYVal = newCoord.y;
	addCoordinate(newCoord);
    }

    /**
     * Add a coordinate to the row
     *
     * @param newCoord The coordinate to add to this row.
     */
    public void addCoordinate(Coordinate newCoord) {
	coordVect.addElement(newCoord);
	newCoord.y = rowYVal;
    }

    /** 
     * Set the y value of a row.  All coordinates in the row are
     *    also immediately assigned this y value.
     *
     * @param y_val The Y value of this row, which will be stored as
     *    a class variable.
     */
    public void setRow(int y_val) {
	rowYVal = y_val;

	for(int i=0; i < coordVect.size(); i++) {
	    ((Coordinate)coordVect.elementAt(i)).y = y_val;
	}
    }

    /**
     * Necessary because rows are subserviant to Columns.
     *    All nodes are set to this X value, but the x value is
     *    not stored, this is the problem of the column that owns
     *    this row.
     *
     * @param x_val The x value to assign to all the coordinates in the row.
     */
    public void setRowX(int x_val) {
	for(int i=0; i < coordVect.size(); i++) {
	    ((Coordinate)coordVect.elementAt(i)).x = x_val;
	}
    }

    /**
     * Shift a row's y value by a constant.  Update all coordinates in the
     *    row, as well as update the class variable.
     *
     * @param y_shift The amount to shift the y value of the row.
     */
    public void shiftRow(int y_shift) {

	rowYVal += y_shift;

	for(int i=0; i < coordVect.size(); i++) {
	    ((Coordinate)coordVect.elementAt(i)).y += y_shift;
	}
    }

    /**
     * Necessary because rows are subserviant to Columns.  Update all 
     *    coordinates.
     *
     * @param x_shift The amount to shift all coordinates in the column.
     */
    public void shiftRowX(int x_shift) {
	for(int i=0; i < coordVect.size(); i++) {
	    ((Coordinate)coordVect.elementAt(i)).x += x_shift;
	}
    }

    /**
     * Get a row's y location.
     *
     * @return The y value of the row.
     */
    public int getYLoc() {
	return rowYVal;
    }

    /**
     * Get a vector of all the coordinates contained by the row.
     *
     * @return The class Vector of all coordinates in the row.  Changes
     *   made to this vector are changes made to this row.
     */
    public Vector getCoordVect() {
	return coordVect;
    }

    /**
     * Get a vector of the nodes in the row.
     *
     * @return A Vector of all the nodes in the row.  This Vector is a copy
     *   of the class variable, but coordinates to the coordinates are permanent
     */
    public Vector getNodeVect() {

	Vector nodeVect = new Vector();
	
	for(int i=0; i < coordVect.size(); i++) {
	    if(coordVect.elementAt(i) instanceof Node) {
		nodeVect.addElement(coordVect.elementAt(i));
	    }
	}
	return nodeVect;
    }

    /**
     * Get a count of the number of nodes in the row.
     *
     * @return The number of nodes in the row.
     */
    public int getNodeCount() {
	int count = 0;
	
	for(int i=0; i < coordVect.size(); i++) {
	    if(coordVect.elementAt(i) instanceof Node) {
		count++;
	    }
	}
	return count;
    }

    /** 
     * Get a count of the number of labels in the row.
     *
     * @return The number of labels in the row.
     */
    public int getLabelCount() {
	int count = 0;
	
	for(int i=0; i < coordVect.size(); i++) {
	    if(coordVect.elementAt(i) instanceof Label) {
		count++;
	    }
	}
	return count;
    }

    /**
     * Get a count of all the coordinate in the row.
     *
     * @return The number of coordinates in the row.
     */
    public int getCoordCount() {
	return coordVect.size();
    }

    /**
     * Test a row if it contains the specified coordinate.
     *
     * @param coord The coordinate to test if it is in the row.
     * @return true if the coordinate is in the column, false otherwise.
     */
    public boolean containsCoord(Coordinate coord) {
	for(int i=0; i < coordVect.size(); i++) {
	    if(((Coordinate)coordVect.elementAt(i)).equals(coord)) {
		return true;
	    }
	}
	return false;
    }

    /**
     * Test a row if it contains the specified coordinate.
     *
     * @param name The name of a coordinate to test if its in the row.
     * @return true if a coordinate by the given name was found in the row, 
     *	    false otherwise.
     */
    public boolean containsCoord(String name) {
	for(int i=0; i < coordVect.size(); i++) {
	    if(((Coordinate)coordVect.elementAt(i)).equals(name)) {
		return true;
	    }
	}
	return false;
    }

    /**
     * Remove the specified coordinate
     *
     * @param name The name of the coordinate to remove.
     * @return true if the coordinate was removed, false otherwise.
     */
    public boolean removeCoord(String name) {
	for(int i=0; i < coordVect.size(); i++) {
	    if(((Coordinate)coordVect.elementAt(i)).equals(name)) {
		coordVect.removeElementAt(i);
		return true;
	    }
	}
	return false;
    }

    /**
     * Remove the specified coordinate
     *
     * @param coord The coordinate to remove.
     * @return true if the coordinate was removed, false otherwise.
     */
    public boolean removeCoord(Coordinate coord) {
	for(int i=0; i < coordVect.size(); i++) {
	    if(((Coordinate)coordVect.elementAt(i)).equals(coord)) {
		coordVect.removeElementAt(i);
		return true;
	    }
	}
	return false;
    }

    /**
     * Explicitly sync all the x's and y's of the coordinates in the row
     *
     * @param xVal The x value to set all coordinates to.
     * @param yVal The y value to set all coordinates to.  This value is
     *    also stored in the row's class variable, since it should know its
     *    own y value.
     */
    public void setXandYvalues(int xVal, int yVal) {

	Coordinate tmpCoord;
	rowYVal = yVal;

	for(int i=0; i < coordVect.size(); i++) {
	    tmpCoord = (Coordinate)coordVect.elementAt(i);
	    tmpCoord.x = xVal;
	    tmpCoord.y = yVal;
	}
    }

    /**
     * Test if the row is empty
     *
     * @return true if there are no coordinates in the row, false otherwise.
     */
    public boolean isEmpty() {

	if(coordVect.size() == 0) {
	    return true;
	} else {
	    return false;
	}
    }

    /**
     * Set a all the nodes to the specified column index
     * 
     * @param index The new index of this column.  It is just written
     *   straight to the nodes so we can figure out who they belong to later.
     */
    public void setColIdx(int index) {

	Coordinate tmpCoord;

	for(int i=0; i < coordVect.size(); i++) {
	    tmpCoord = (Coordinate)coordVect.elementAt(i);
	    tmpCoord.colNum = index;
	}
    }

    /**
     * Sort a vector of coordinate by their parent (and child) mean.
     *   This method will request a sort of the nodes by the children
     *   if the parents result in a tie.  After the child sort returns, 
     *   the keys are presumed in sorted order and the nodes are spaced
     *   accordingly.
     * 
     * @param minNodeSpacing The spacing added between nodes when
     *         they are spaced across the row.
     */
    public void sortNodesByParentMean(int minNodeSpacing, 
			int maxChildSearchDepth) {

	// keep it off the stack when we recurse
	this.maxChildSearchDepth = maxChildSearchDepth;

	sortGroupByParents( coordVect, 
			    minNodeSpacing);
    }

    /**
     * Sort a vector of coordinate by their parent (and child) mean.
     *   This method will request a sort of the nodes by the children
     *   if the parents result in a tie.  After the child sort returns, 
     *   the keys are presumed in sorted order and the nodes are spaced
     *   accordingly.
     * 
     * @param row The row we're sorting by parent mean.
     * @param minNodeSpacing The spacing added between nodes when
     *         they are spaced across the row.
     */
    private void sortGroupByParents(Vector row, int minNodeSpacing) {


	if(row.size() <= 1) {
	    return;
	}
	
	// We'll sort the indexes, then write the correct order and
	//    X coords at the very end.
	int[] coordKeys          = new int[row.size()];

	// initialize coord key list
	for(int i=0; i < coordKeys.length; i++) {
	    coordKeys[i] = i;
	}

	// First sort by parents, worry about ties later.
	int[] parentXMeanArray = new int[row.size()];

	// find the parent means.
	for(int i=0; i < parentXMeanArray.length; i++) {
	    parentXMeanArray[i] = getParentXMean(
		    (Coordinate)row.elementAt(coordKeys[i]));
	}

	// sort nodes by parent means 
	PlotPathsUtil.quicksortByForeignKey(coordKeys, parentXMeanArray);

	// for each tie in the parentMean step, submit to childMean step
	int lastVal		= 0;
	int startTieGroup	= 0;
	int endTieGroup		= 0;
	boolean inTieGroup 	= false;
	if(parentXMeanArray.length > 0) {
	    lastVal = parentXMeanArray[0];
	}
	for(int i=1; i < parentXMeanArray.length; i++) {
	    if(!inTieGroup && parentXMeanArray[i] == lastVal) {
		inTieGroup = true;
		startTieGroup = i-1;
	    }
	    if(inTieGroup && parentXMeanArray[i] != lastVal) {
		endTieGroup = i-1;
		sortNodeGroupByChildrenXMean(	row, 
				coordKeys, 
				startTieGroup, 
				endTieGroup);
		inTieGroup = false;
	    }
	    if(inTieGroup && i == parentXMeanArray.length-1) {
		endTieGroup = i;
		sortNodeGroupByChildrenXMean(	row, 
				coordKeys, 
				startTieGroup, 
				endTieGroup);
		inTieGroup = false;
	    }

	    lastVal = parentXMeanArray[i];
	}

	// actually move the x coords around
	float temp;
	int nodeGroupX = ((Coordinate)row.elementAt(0)).x;
	for(int i = 0; i < coordKeys.length; i++) {

	    temp =  (float)nodeGroupX + minNodeSpacing * 
		    ((float)i - ((float)coordKeys.length-1)/2);

	    ((Coordinate)row.elementAt(coordKeys[i])).x  = (int)temp;
	}

	// sort the vector such that the least x val comes first, 
	//	-> and the biggest comes last.
    }

    /** 
     * Sort a row by its parent (and child) mean.  Start off the recursive
     *   version of this method at a depth of 1.
     */
    private void sortNodeGroupByChildrenXMean(Vector row, int[] coordKeys, 
			int startIdx, int endIdx) {

	// start off with a search depth of 1
	sortNodeGroupByChildrenXMean(	row, 
					coordKeys, 
					startIdx, 
					endIdx, 
					1);

    }

    /**
     * Sort a portion of a vector by its children.  Will proceed recursively.
     *
     * @param row The row we're stepping through recursively.
     * @param coordKeys indexes into the row Vector, but ordered to reflect
     *    the current sorting (we sort keys, not the actual nodes)
     * @param startIdx The start index of the current iteration into the
     *    coordKeys array.
     * @param endIdx The end index of the current iteration into the coordKeys
     *    array.
     * @param depth The current depth we're iterating over for child means.
     */
    private void sortNodeGroupByChildrenXMean(Vector row, int[] coordKeys, 
			int startIdx, int endIdx, int depth) {

	//System.out.println("Depth: "+depth);

	if(startIdx >= endIdx || 
		(depth > maxChildSearchDepth && maxChildSearchDepth != -1)) {
	    //System.out.println("returning");
	    return;
	}

	// so length is consistent for quicksort
	int[] childXMeanArray = new int[coordKeys.length];

	for(int i=startIdx; i <= endIdx; i++) {
	    try {
		childXMeanArray[i] = getChildXMean(
			(Coordinate)row.elementAt(coordKeys[i]), 
			depth);
	    } catch (IllegalArgumentException e) {
		//System.out.println("Exceeded maximum depth in child sort");
		// set it to something arbritrary that will break the tie.
		childXMeanArray[i] = Integer.MIN_VALUE;
	    }
	    //System.out.println("Mean "+i+" - "+childXMeanArray[i]);
	}

	// sort nodes by child means 
	PlotPathsUtil.quicksortByForeignKey( coordKeys, 
					    childXMeanArray,
					    startIdx,
					    endIdx);

	// for each tie in the childMean step, submit to childMean at next depth
	int lastVal		= 0;
	int startTieGroup	= 0;
	int endTieGroup		= 0;
	boolean inTieGroup 	= false;
	if(childXMeanArray.length > 0) {
	    lastVal = childXMeanArray[startIdx];
	}
	for(int i=startIdx+1; i <= endIdx; i++) {
	    if(!inTieGroup && childXMeanArray[i] == lastVal) {
		inTieGroup = true;
		startTieGroup = i-1;
	    }
	    if(inTieGroup && childXMeanArray[i] != lastVal) {
		endTieGroup = i-1;
		if(childXMeanArray[endTieGroup] != Integer.MIN_VALUE) {
		    sortNodeGroupByChildrenXMean(	
				    row, 
				    coordKeys, 
				    startTieGroup, 
				    endTieGroup, 
				    depth+1);
		}
		inTieGroup = false;
	    }
	    if(inTieGroup && i == endIdx) {
		endTieGroup = i;
		if(childXMeanArray[endTieGroup] != Integer.MIN_VALUE) {
		    sortNodeGroupByChildrenXMean(	
				    row, 
				    coordKeys, 
				    startTieGroup, 
				    endTieGroup, 
				    depth+1);
		}
		inTieGroup = false;
	    }

	    lastVal = childXMeanArray[i];
	}


	// contents of coordKeys have been modified on sorted order
    }

    /**
     * Calculate the mean of a coordinate's parent's X coord.
     *
     * @param coord The coordinate to find the parent mean of.
     * @return The mean of the coodinates immediate parents. else the 
     *   coordinate's x value.
     */
    private int getParentXMean(Coordinate coord) {
	if(!(coord instanceof Node)) {
	    return coord.x;
	}
	
	int sum 	= 0;
	int parentCount = 0;
	int mean;
	Node[] neighborNodes;

	// grab the neighbors
	neighborNodes = ((Node)coord).getNeighborNodes();


	// only include parents (greater Y coord)
	Coordinate tmpCoord;
	for(int i=0; i < neighborNodes.length; i++) {
	    tmpCoord = neighborNodes[i];
	    if(tmpCoord.y < coord.y) {
		sum += tmpCoord.x;
		parentCount++;
	    }
	}

	if(parentCount == 0) {
	    return coord.x;
	}

	// calculate the average
	mean = sum / parentCount;
	
	return mean;
    }

    /**
     * Calculate the mean of a coordinate's children's X coordinates at the
     *   given depth, where the depth is the number of links followed before
     *   the X coordinate is extracted.
     *
     * @param coord The coordinate to get the child mean for.
     * @param depth The number of children to follow before calculating the 
     *	    mean.
     * @return The mean of the children at the specified depth from the 
     *   given coodinate. else the given coordinate's x value.
     */
    private int getChildXMean(Coordinate coord, int depth) {

	//System.out.println("Depth: "+depth);
	if(!(coord instanceof Node)) {
	    return coord.x;
	}

	// grab the neighbors at depth by iterating
	Node[] neighborList;

	// neighbors that we're iterating over
	Vector thisList = new Vector();	
	// newly grabbed neighbors -> become thisList
	Vector nextList = new Vector();	

	Node tmpNode;
	Node tmpChildNode;
	thisList.addElement(coord);
	while(depth > 0) {
	    // get children neighbors iteratively
	    for(int i=0; i < thisList.size(); i++) {
		tmpNode = (Node)thisList.elementAt(i);
		neighborList = tmpNode.getNeighborNodes();
		for(int j=0; j < neighborList.length; j++) {
		    tmpChildNode = neighborList[j];
		    if(tmpChildNode.y > tmpNode.y) {
			nextList.addElement(tmpChildNode);
		    }
		}
	    }


	    // swap vectors
	    thisList = nextList;
	    nextList = new Vector();

	    if(thisList.size() == 0) {
		throw new IllegalArgumentException(
			    "Depth specified was too large");
	    }
	    depth--;
	}

	int sum 	= 0;
	int parentCount = 0;
	int mean;

	if(thisList.size() == 0) {
	    return coord.x;
	}

	for(int i=0; i < thisList.size(); i++) {
	    sum += ((Coordinate)thisList.elementAt(i)).x;
	    parentCount++;
	}

	// calculate the average
	mean = sum / parentCount;
	
	return mean;
    }

    public void debug() {
	System.out.println("========= begin Row ===============");
	System.out.println("[Debug] Row");
	System.out.println("rowYVal: "+rowYVal);
	for(int i=0; i < coordVect.size(); i++) {
	    ((Coordinate)coordVect.elementAt(i)).debug();
	}
	System.out.println("========= end Row =================");
    }

    private int rowYVal;
    private Vector coordVect = new Vector();
    private int maxChildSearchDepth = -1;
}
